#!/usr/bin/env python
"""
$Id: Autonomic.py,v 1.11 2005/08/24 02:41:18 rayg Exp $
Copyright 2005, University of Wisconsin Space Science & Engineering Center

    This file is part of the AERI ingest software.

    The AERI ingest software is free software; you can redistribute it and/or 
    modify it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA



"""

import logging,imp,os

import Config

from process import process

log = logging.getLogger("Autonomic")

class Autonomic:
    """Module which proxies the autonomic functions and tasks of the instrument, 
    principally related to mirror safing.
    """

    def __init__(self):
        env = { 'MIRROR_SCD': Config.MIRROR_SCD, 'AEMIRSAF_CFG': Config.AEMIRSAF_CFG, 'CHANNEL_MAP': Config.CHANNEL_MAP }
        self.mirsaf = process(Config.AEMIRSAF_EXE, 
                                environ = env,
                                output_stream=Config.logStream("aemirsaf.out"),
                                output_prefix='aemirsaf')
        
        log.info("reading hatch daemon configuration file containing COM_PORT and TCP_PORT settings")
        cfgmod = imp.load_source( 'hatch_cfg', Config.HATCH_CFG, file(Config.HATCH_CFG,'U') )
        self.hatch_tcp = cfgmod.TCP_PORT
        self.hatch_com = cfgmod.COM_PORT

        env = { 'HATCH_PORT': self.hatch_com,
                'PORT': '%d' % self.hatch_tcp 
                } 
        self.hatch = process( '/usr/bin/python', arguments = [ Config.HATCH_EXE ], environ = env, output_stream = Config.logStream('hatchd.out'), output_prefix='hatchd' )
        
    def setHatch(self,state=True):
        """Change the hatch-enable state on the hatch daemon
        """
        from socket import socket
        s = socket()
        log.debug('connecting to tcp port %d' % self.hatch_tcp)
        s.connect( ('localhost',self.hatch_tcp) )
        s.send( state and 'enable\n' or 'disable\n' )
        s.close()
    
    def startHatch(self):
        #env = os.environ.copy()
        #mypid = os.spawnvpe( os.P_NOWAIT, 'python', [ Config.HATCH_EXE ], env )

        #mypid = os.fork()
        #if not mypid: 
        #    for k,v in env.items:
        #        os.environ[k] = v
        #    os.execvp( 'python', [ Config.HATCH_EXE ] )
        #self.hatch = process( 'hatchd', pid = mypid )
        self.hatch.start()
        
        return self.hatch
        
    def startMirrorSafing(self):
        """bootstrap mirror safing daemon; return task representing it
        """
        self.mirsaf.start()
        return self.mirsaf
        
    def start(self):
        self.mirsaf.start()
        self.hatch.start()
        return self
        
    def preheatMirrorMotor(self):
        """preheat the mirror motor. return when the mirror is warm enough to rotate.
        """
        env = { 'MIRROR_SCD': Config.MIRROR_BEG, 'AEMIRSAF_CFG': Config.MOTHEAT_CFG, 'CHANNEL_MAP': Config.CHANNEL_MAP }
        p = process(Config.AEMIRSAF_EXE, 
                    environ = env,
                    output_stream=Config.logStream("preheat.out"),
                    output_prefix='aemirsaf')
        log.info("launching preheat")
        p()
        
    def kill(self):
        self.mirsaf.kill()
        self.hatch.kill()

def test_simple(args):
    """Do a simple mirror safing test.
    Force mirror to ready state.
    Start up housekeeping subsystem. 
    Start up mirror safing.
    Wait for user input.
    Close down safing and housekeeping.
    """
    from Mirror import Mirror
    from Housekeeping import Housekeeping
    mir = Mirror()
    log.info("forcing mirror to ready state")
    mir.forceReady()
    log.info("housekeeping startup")
    hk = Housekeeping()
    hskTasks = hk.start()
    aut = Autonomic()
    mirsafTasks = aut.startMirrorSafing()
    cmd = raw_input( 'hit enter to stop program: ' )
    mirsafTasks.kill()
    hskTasks.kill()
    
def test_preheat(args):
    print "FIXME -- Not implemented"

def test_hatch(args):
    """Quick toggling test for hatch.
    """
    A = Autonomic()
    print "starting hatch daemon"
    task = A.startHatch()
    from time import sleep
    s = True
    for i in range(0,8):
        sleep(2)
        s = not s
        print "setting hatch state %s" % s
        A.setHatch(s)
    print "stopping hatch daemon"
    task.kill()

def verifyInstall(args):
    """verify that the subsystem has the pieces it needs to operate
    """
    all = [ Config.HATCH_EXE, Config.HATCH_CFG, Config.HATCHON_EXE, Config.AEMIRSAF_CFG, Config.CHANNEL_MAP, Config.AEMIRSAF_EXE ]
    for each in all:
        print "checking for %s" % each
        assert( os.path.exists( each ) )


if __name__=='__main__':
    doc = """
python Autonomic.py preheat -- run AEMIRSAF in preheat mode, along with housekeeping
python Autonomic.py simple -- simple mirror safing test, includes housekeeping startup/shutdown
python Autonomic.py hatch -- cycle hatch DTR signal 4 times
python Autonomic.py verify_install -- check that files needed are installed
"""
    from sys import argv
    logging.basicConfig()
    log.setLevel( logging.DEBUG )
    tests = {   'simple': test_simple,
                'preheat': test_preheat,
                'hatch': test_hatch,
                'verify_install': verifyInstall
                }
                
    if len(argv) > 1 and argv[1] in tests.keys():
        tests[argv[1]](argv[2:])
    else:
        print doc
    
