#!/usr/bin/env python
"""
$Id: Config.py,v 1.70.2.8 2008/09/28 20:28:46 rayg Exp $
Copyright 2005, University of Wisconsin Space Science & Engineering Center

    This file is part of the AERI ingest software.

    The AERI ingest software is free software; you can redistribute it and/or 
    modify it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA



"""

import os,logging,re,sys,time

import win32

log = logging.getLogger('Config')

def envor(envname,default):
    return os.environ.get( envname, default )

# original base information, private, do not use outside this module
_CONFIG_DIR = envor('_CONFIG_DIR','\\config')
_WORK_DIR = envor('_WORK_DIR','\\temp')
_FTP_DIR = envor('_FTP_DIR','\\FTP')
_SCI_DIR = envor('_SCI_DIR','\\')
_OUTGOING_DIR = envor('_OUTGOING_DIR','\\FTP\\OUTGOING')
_AERI_RUN_DIR = envor('_AERI_RUN_DIR','\\AERI_RUN')
_AERI_DIR = envor('_AERI_DIR','\\AERI')

# all errors and criticals are appended to this log
ERROR_LOG=envor('ERROR_LOG', '\\aeri_error.log')

# by default, current day is used; override with env variable YYMMDD
YYMMDD=time.strftime('%y%m%d')

def configDir():
    """ return the location of the standard config files
    #FIXME: not yet configurable
    """
    return _CONFIG_DIR

def workDir():
    """Return the configured work directory for acquiring data.
    #FIXME: this should be configurable
    """
    return _WORK_DIR

def outputDir(yymmdd = None):
    """Return the directory where science output (RNC, SUM, PAR files) is placed.
       requires a 'day' parameter, string YYMMDD.
       If run without yymmdd, it'll return the enclosing directory where the output dirs reside.
    """
    if not yymmdd: return _FTP_DIR
    return win32.path_join( _FTP_DIR, 'AE' + yymmdd )

def scienceWorkDir(yymmdd = None):
    """Return the directory where science intermediate files (CXS, CVS, others) is placed.
       requires a 'day' parameter, string YYMMDD.
       If run without yymmdd, it'll return the enclosing directory where the work dirs reside.
    """
    if not yymmdd: return _SCI_DIR
    return win32.path_join( _SCI_DIR, 'AE' + yymmdd )

def calDir(yymmdd = None):
    """Return the calibration directory where intermediate files only required for processing
       are stored.
       if run without yymmdd, it will return scienceWorkDir().
    """
    if not yymmdd: return _SCI_DIR
    return win32.path_join( scienceWorkDir(yymmdd), 'cal' )

def aeriRunBomemExe( exename ):
    """return a callable/startable executable in BOMEM; do not include .exe in name of program
    """
    return '\\bomem\\%s.exe' % exename

def aeriRunEaeriExe( exename ):
    """return a callable/startable executable in EAERI software directory; do not include .exe in name of program
    """
    return '\\E-AERI\\%s.exe' % exename
    
def eaeriConfigPath(filename):
    """configuration file for data ingest of the E-AERI"""
    return '\\E-AERI\\%s' % filename

def aeriRunExe( exename,suffix='exe' ):
    """return a callable/startable executable in AERI_RUN; do not include .exe in name of program
    """
    return '\\aeri_run\\%s.%s' % (exename,suffix)

def aeriRunHresdExe( exename ):
    """return a callable/startable executable in HRESD; do not include .exe in name of program     
    """
    return '\\aeri_run\\hresd\\%s.exe' % exename

def aeriRunWdtExe( exename ):
    """return a callable/startable executable in WDT; do not included .exe in name of program
    """
    return '\\aeri_run\\wdt\\%s.exe' % exename

def aeriOs2Exe( exename ):
    """return a callable/startable executable in OS2EXE; do not include .exe in name of program
    """
    return '\\aeri\\os2exe\\%s.exe' % exename

def configPath(name):
    """ return an instrument configuration file pathname, preferably verifying that it exists
    """
    return win32.path_join( configDir(), name )
    
def outputPath(name):
    """ return an instrument working file pathname, preferably verifying that it exists
    """
    return win32.path_join( outputDir(YYMMDD), name )

def calPath(name):
    """Return a pathname for a given file in the Config.calDir()
    """
    return win32.path_join( calDir(YYMMDD), name )
    
def aeriRunPath(name):
    return win32.path_join( "\\aeri_run", name )

def workPath(name):
    """ return an instrument working file pathname, preferably verifying that it exists
    """
    return win32.path_join( workDir(), name )

def scienceWorkPath(name):
    """ return an instrument science work directory pathname, preferably verifying that it exists
    """
    return win32.path_join( scienceWorkDir(YYMMDD), name )

def logStream(filename):
    """return an appendable text file which we can send log data to
    """
    logdir = calDir(YYMMDD)
    if not os.path.exists( logdir ):
        log.warning("sending log file %s to ingest work directory - presumably a test" % filename)
        logdir = workDir()
    return file( win32.path_join( logdir, filename ), 'at' )

def runningCfgPath( file ):
    """Return a pathname for 'file' in the location where the running 
    config files are located
    """
    #FIXME this will need to change later when all Config.<filename> vars have changed
    # to just refer to the filename without a path
    return calPath( win32.path_split( file )[1] )

# standard configuration files which are oft-accessed
OP_SET = configPath("op.set")
MIRROR_BEG = configPath("mirror.beg")
MIRPRGM_MOT = configPath("mirprgm.mot")

MIRROR_SCD = workPath("mirror.scd")
SEQ_LAB = workPath("seq.lab")
DISKDAY = workPath("current.day")

AERIHK_CFG = configPath('aerihk.cfg')
AERI_POS = workPath('aeri.pos')
CHANNEL_MAP = configPath('channel.map')
INST_ID = configPath('inst.id')
MIR_POS = workPath('mir.pos')
AEMIRSAF_CFG = configPath('aemirsaf.cfg')

AERIGPS_CFG = configPath('aerigps.cfg')

HATCH_CFG = configPath('hatch.cfg')
HATCH_STA = workPath('hatch.sta')

BBCTRL_TMP = configPath('bbctrl.tmp')
# CTL_SWTS = configPath('ctl_swts')

# housekeeping configuration files
ABCVALS_NUM = configPath('abcvals.num')
# HRES_CFG = configPath('hres.cfg')
MANDB = configPath('mandb')
MOTHEAT_CFG = configPath('motheat.cfg')
RBRIDGE_RES = configPath('rbridge.res')

# science configuration files
RWASPC_SIP = configPath('rwaspc.sip')
AERICALV_SIP = configPath('aericalv.sip')
AERIEMIS_ASC = configPath('aeriemis.asc')
AERISUM_SIP = configPath('aerisum.sip')
FFOVCMR_SI1 = configPath('ffovcmr.si1')
FFOVCMR_SI2 = configPath('ffovcmr.si2')
NLAPP_SIP = configPath('nlapp.sip')
NUMSAMPS_AVG = configPath('numsamps.avg')
ZFLI_SI1 = configPath('zfli.si1')
ZFLI_SI2 = configPath('zfli.si2')

# disk and process management
DISK_MIN_CFG = configPath('disk_min.cfg')
OUTGOING_LIS = configPath("outgoing.lis")
# STARTAPS_LST = configPath("startaps.lst")

# used for restarts, instructs science software to skip sequences
#FIXME this could be done cleaner
AERI_CTL = scienceWorkPath('AERI.CTL')

# JADA config parameters
JAVA_EXE = '\\windows\\system32\\java.exe'
JAVA_VERSION = '1.4.2_07'
JADA_APP_ROOT='\\AERI_RUN\\jada'
JADA_LIB = '%s\\lib' % JADA_APP_ROOT
JADA_CLASSPATH='.;%s\\JSAP-2.0a.jar;%s\\xmlParserAPIs.jar;%s\\dmv.jar;%s\\xercesImpl.jar;%s\\visad.jar;%s\\log4j.jar;jada.jar' %((JADA_LIB,)*6)
JADA_LOG_FILE = 'jada.log'
JADA_LOG_CFG = 'logger.xml.template'
JADA_INSTS_FILE = 'instruments.cfg'

# Ingest Messaging config
MSG_PORT = '2727'
MSG_HOST = 'localhost' # meaning localhost
MSG_MAXBUFF = 8192 # max message seg size
MSG_MAX_CONN = 1 # max connections to ingest msg server
MSG_SVR = '\\aeri_run\\MessageServer.py'

# AERIQC text-mode blinkenlights
AERIQC_EXE = aeriRunExe('aeriqc')
AERIQC_LOG = '%s.QC' %YYMMDD

# major (run for extended periods or daemonic) slaved executables 
# major slaved executables 
ACQ96_EXE = aeriRunBomemExe( "acq96nt" ) #program which drives the Bomem interferometer and collects ifgs
ACQ96_LOG = "ACQ96NT.LOG" #log file written by ACQ96NT to its working directory

# EAERI subsystem executables and settings
# EAERI_CFG = configPath('eaeri.cfg') 
ACQEAERI_EXE = aeriRunEaeriExe( "acqeaeri" )
EAERI_BASE_DIR = "\\E-AERI"

# these are used to start up the main application which collects data
EAERI_APP_ROOT = EAERI_BASE_DIR
EAERI_JRE = "\\E-AERI\\jre\\bin\\javaw.exe"
EAERI_PARMS = [ '-Xms256m', '-Xmx512m' ]
EAERI_JAR = "\\E-AERI\\FTSW_EAERI.jar"
EAERI_MAIN = "com.bomem.aries.console.Main"

# these config files get stored in the PAR file
EAERI_AGILENT_CONFIG = eaeriConfigPath('agilent.config')
EAERI_FTSW_CONFIG = eaeriConfigPath('FTSW_Eaeri.Config')

# are we an E-AERI?
IS_EAERI = os.path.exists( ACQEAERI_EXE )
# if so, then use their JRE to run Jada
if IS_EAERI:
    JAVA_EXE = EAERI_JRE
    AERI_POS = configPath('aeri.pos')  # GPS will come in from housekeeping system when added; \config\aeri.pos as fallback

HRESD_EXE = aeriRunHresdExe( 'hresd' ) #daemon which serves the housekeeping data as a shared mem block
HRESDD_EXE = aeriRunHresdExe( 'hresdd' ) #interfaces with the hardware drivers to provide data to hresd
AERIHK_EXE = aeriRunHresdExe( 'aerihk' ) #buffers housekeeping data in memory and writes *HK.SPC files

# minor (immediate return) slaved executables
# mirror routines
AEREQMIR_EXE = aeriRunExe( "aereqmir" ) #requests that if the mirror can be freed from safe, free it
AEMIRPRG_EXE = aeriRunExe( "aemirprg" ) #uploads the mirror control program from a local file
AERIMIRP_EXE = aeriRunExe( "aerimirp" ) #requests mirror position, or other command be sent to mirror ctrlr
AEMIRCMD_EXE = aeriRunExe( "aemircmd" ) #daemon which serves mirror state as shared memory
AEMIRGO_EXE = aeriRunExe( "aemirgo" ) # not to be confused with retired AERIMIRC
AERIGPS_EXE = aeriRunExe( "aerigps" ) # writes gps info to aeri.pos according to AERIGPS_CFG

# housekeeping routines
HRESVIEW_EXE = aeriRunHresdExe( 'hresview' ) #queries the shared memory with housekeeping, displays it
AERISIHK_EXE = aeriRunHresdExe( 'aerisihk' ) #causes AERIHK to exit
AESCNMRK_EXE = aeriRunHresdExe( 'aescnmrk' ) #makes AERIHK mark the start of an interferogram set acquire
AEWRTHK_EXE = aeriRunHresdExe( 'aewrthk' ) #causes AERIHK to write its buffered HK to a file

# autonomic routines
AEMIRSAF_EXE = aeriRunExe( "aemirsaf" ) # monitors housekeeping and safes mirror if conditions stray
HATCH_EXE = aeriRunExe( "hatchd","py" ) # provides DTR enable signal to hatch system
HATCHON_EXE = aeriRunExe( "hatchon" ) # C++ executable which holds open a com port, since Cygwin python won't open

# science routines
RWASPC_EXE = aeriOs2Exe( 'rwaspc' ) #Read/Write AERI SPC files
NLAPP_EXE = aeriOs2Exe( 'nlapp' ) #Nonlinearity Correction
AERICALV_EXE = aeriOs2Exe( 'aericalv' ) #AERI Calibration 
SCNDIRAV_EXE = aeriOs2Exe( 'scndirav' ) #Average Scan Directions
FFOVCMR_EXE = aeriOs2Exe( 'ffovcmr' ) #Finite Field of View Correction
ZFLI_EXE = aeriOs2Exe( 'zfli' ) #Zero Fill Linear Interpolation 
AERISUM_EXE = aeriOs2Exe( 'aerisum' ) # AERI Summary file creation

# display source configuration files
AESITTER_SCR = configPath('AESITTER.SCR')
RADIANCE_SCR = configPath('RADIANCE.SCR')
SUMARY_SCR = configPath('SUMARY.SCR')

# files which go into the PAR file with their expected path
AERI_PAR_FILES = [   runningCfgPath(INST_ID),        # runningCfgPath(STARTAPS_LST),
                runningCfgPath(AERIHK_CFG),     runningCfgPath(MANDB),
                runningCfgPath(RBRIDGE_RES),    runningCfgPath(NUMSAMPS_AVG),
                runningCfgPath(ABCVALS_NUM),    runningCfgPath(MOTHEAT_CFG),
                runningCfgPath(AERI_POS),       runningCfgPath(MIRROR_BEG),
                runningCfgPath(OP_SET),         runningCfgPath(AERIEMIS_ASC),
                runningCfgPath(RWASPC_SIP),     runningCfgPath(NLAPP_SIP),
                runningCfgPath(AERICALV_SIP),   runningCfgPath(FFOVCMR_SI1),
                runningCfgPath(FFOVCMR_SI2),    runningCfgPath(ZFLI_SI1),
                runningCfgPath(ZFLI_SI2),       runningCfgPath(DISK_MIN_CFG),
                runningCfgPath(AERISUM_SIP),    runningCfgPath(CHANNEL_MAP),
                runningCfgPath(AERIGPS_CFG),    runningCfgPath(BBCTRL_TMP),
                runningCfgPath(OUTGOING_LIS),   runningCfgPath(AEMIRSAF_CFG),
                SUMARY_SCR,
                AESITTER_SCR,
                RADIANCE_SCR
            ]

EAERI_PAR_FILES = [   runningCfgPath(INST_ID),        
                EAERI_FTSW_CONFIG, EAERI_AGILENT_CONFIG,
                runningCfgPath(AERI_POS),       runningCfgPath(MIRROR_BEG),
                runningCfgPath(OP_SET),         runningCfgPath(AERIEMIS_ASC),
                runningCfgPath(RWASPC_SIP),     runningCfgPath(NLAPP_SIP),
                runningCfgPath(AERICALV_SIP),   runningCfgPath(FFOVCMR_SI1),
                runningCfgPath(FFOVCMR_SI2),    runningCfgPath(ZFLI_SI1),
                runningCfgPath(ZFLI_SI2),       runningCfgPath(DISK_MIN_CFG),
                runningCfgPath(AERISUM_SIP),    runningCfgPath(CHANNEL_MAP),
                runningCfgPath(OUTGOING_LIS),   
                SUMARY_SCR,
                AESITTER_SCR,
                RADIANCE_SCR
            ]

if IS_EAERI:
    PAR_FILES = EAERI_PAR_FILES
else:
    PAR_FILES = AERI_PAR_FILES

# system executables
_df_exe = { 'cygwin': '/usr/bin/df', 'darwin': '/bin/df', 'linux2': '/bin/df' }
_du_exe = { 'cygwin': '/usr/bin/du', 'darwin': '/usr/bin/du', 'linux2': '/usr/bin/du' }

DF_EXE = _df_exe[ sys.platform ]
DU_EXE = _du_exe[ sys.platform ]
BANNER_EXE = '/usr/bin/banner'

#Location of WDT port access executable
TESTPORT_EXE = aeriRunWdtExe( 'testport' )

# the number of bytes of disk space required in order to start system up
# Tidy uses this prior to other work directories being created
INITIAL_SPACE=4*1024**3

# Amount of delay for the Watchdog timer in minutes
# Actual time is half a minute less then this
DELAY = 10

re_inst = re.compile( r'\s*(\w+)\s*' )

def instrumentName():
    """Return the name of the instrument, as placed in \config\inst.id
    """
    return re_inst.findall( file(INST_ID).read() )[0]


# allow environment variables to override anything in the local dictionary
for k in os.environ.keys():
    G = globals()
    if G.has_key(k):
        ov = G[k]
        nv = os.environ[k]
        G[k] = nv
        print "INFO: environment overriding system configuration %s as %s (originally %s)" % (k,nv,ov)


if __name__=='__main__':

    from pprint import pprint
    pprint(globals())


