#!/usr/bin/env python
# encoding: utf-8
"""
Eaeri.py

Created by Keoni on 2008-02-04.
Copyright (c) 2008 University of Wisconsin SSEC. All rights reserved.

    This file is part of the AERI ingest software.

    The AERI ingest software is free software; you can redistribute it and/or 
    modify it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
"""

import logging,os,win32,shutil,imp,time, string, re, chmap
import Config
from process import process
from txt2hkspc import txt2hkspc, read_txt, extract_scalars
from mirscd import mirror_schedule

LOG = logging.getLogger( "Eaeri" )

class EAERI_Failure(EnvironmentError): 
    """exception thrown when acqeaeri returns a nonzero result code"""
    pass


# EAERI SUBSYSTEM OBJECT CLASSES
# These are implementations of the hardware proxy object interfaces done with ACQEAERI.
# Ideally these are drop-in replacements that the control logic can use.

class EaeriHousekeeping:
    "An adapter class implementing the Housekeeping interface using EAERI software."
    
    def __init__(self, ear):
        self._eaeri = ear
        # the current file that we're recording housekeeping to
        self._current_hk_fn = None
        # executable we can use
        self.acq = ear.acq
        self.LOG = logging.getLogger("EaeriHousekeeping")
                        
    def start(self): 
        """Bootstrap the housekeeping collection software.
        """                
        # this is already taken care of by Eaeri start()
        # FIXME: consider verifying that the housekeeping is being recorded and verify a record
        return self
        
    def peek(self):
        """Read a record from the high-resolution housekeeping to the screen
        """
        # How to find out the filename currently being written?
        #LOG.warning("EaeriHousekeeping.peek() not implemented")
        print "waiting 10 seconds for housekeeping data.. please wait."
        time.sleep(10)
        # cm = '/cygdrive/c/config/channel.map'
        cm = win32.unix_path(Config.CHANNEL_MAP)
        fn = win32.windows_path(Config.calPath(time.strftime('%Y%m%d_%H%M%S.hk',time.gmtime())))
        self.acq('-k%s' % fn)

        txt = file(fn).read()
        hires = read_txt(txt)
        scalars = extract_scalars(hires)
   
        # get list of channels in channel.map file
        channels = chmap.read( cm )

        for ch in channels:
           try:
              ea_val = scalars[ch[1]]
              print "channel %s: %5.4f:\t %s" % ( ch[0],ea_val,ch[1] )
           except:
              ea_val = "NA"
              print "channel %s: %s:\t\t %s" % ( ch[0],ea_val,ch[1] )      
        return self
       
    def syncAndMarkStartOfSet(self, ifgTask=None):
        """ Synchronize the start of housekeeping collection to the start of interferogram acquisition.
        Start AERIHK recording the data for this acquisition.
        Return no sooner than the start of the interferogram acquire.
        FIXME: Currently does not use pipe synchronization to ACQ96 to find the start-of-scan.
        """
        # send housekeeping between scans down the drain for now
        drainfile = win32.windows_path(Config.calPath(time.strftime('%Y%m%d_%H%M%S.hk',time.gmtime())))
        LOG.debug('start-of-scan, drained housekeeping to ' + drainfile)
        try:
            self.acq('-k%s' % drainfile)
        except EAERI_Failure:
            LOG.info('housekeeping drain: none collected, continuing')
            pass
        return self
      
    def writeHousekeeping(self, pathname):
        """ Write the buffered housekeeping residing in AERIHK buffers to the pathname provided.
        FIXME: the .SET file gets read based on this name. is this a problem?
        raises EnvironmentError if any problem collecting housekeeping occurred
        """
        #hkfile = win32.windows_path(os.path.abspath(pathname))
        hkfile = pathname
        LOG.debug("writeHousekeeping to %s" % hkfile)
        self.acq('-k%s' % hkfile)
        # FIXME: other checks for errors?
        txt2hkspc('%sHK.SPC' % hkfile, [ hkfile ], set_filename='%s.SET' % hkfile, 
	           mirror_scd_filename='%s' % Config.MIRROR_SCD, max_vector_length=64)
        dirname, filename = win32.path_split(pathname)
        assert( os.path.exists(win32.unix_path(pathname)) )

    def kill(self):
        LOG.debug("EaeriHousekeeping kill")

class EaeriGps:
    """ adapter class for GPS positioning
    """
    def __init__(self, ear):
        self._eaeri = ear

    def writeAeriPos(self):
        pass
        # LOG.warning("EaeriGps.writeAeriPos not implemented!")

class EaeriMirror:
    """An adapter class which uses ACQEAERI to command the mirror.
    """
    def __init__(self, ear):
        self._eaeri = ear
        # the current file that we're recording housekeeping to
        self._current_hk_fn = None
        # executable we can use
        self.acq = ear.acq
        self.LOG = logging.getLogger("EaeriMirror")
                        
    def start(self):
        """Start up mirror daemon - this shared memory server holds the state of the mirror system
        """
        return self # as a task, effectively
        
    def kill(self):
        pass        
        
    def isReady(self):
        """return boolean, whether or not the mirror is available to rotate
        """
        self.LOG.warning("isReady not implemented")
        # raise NotImplementedError
        return True 

    def getState(self):
        """ return MIR_* constant indicating the state of the mirror
        """
        self.LOG.warning("getState not implemented")
        
    def programMirrorController(self):
        self.LOG.warning("programMirrorController not implemented")
           
    def rotateToPosition(self,mirror_position,schedule_pathname=None):
        """rotate the mirror go a given scene using the default mirror schedule file in the workdir
        """
        # open mirror schedule and read it into memory
        mp = mirror_schedule()
        if schedule_pathname is None: schedule_pathname = Config.MIRROR_SCD
        mp.read( file(schedule_pathname, 'rt') )
        # get the mirror program elements for the indicated view name
        elements = [ x for x in mp.views if x.name[0].upper() == mirror_position[0].upper() ]
        # make sure we got at least one
        assert(len(elements)>0)
        # get the angle for that named position
        mirror_angle = elements[0].angle
        
        # send command to go to angle
        self.acq('-M%f' % mirror_angle)        
        self.LOG.info("rotating to position %s" % mirror_angle)
        #FIXME: determine whether it actually made it there, or stalled out

    def requestFreeFromSafing(self):
        """ Signal the mirror system that now is a good time to release the mirror from safing.
        In the case that the mirror isn't safed, no-op.
        If the mirror cannot be released from safe, no-op.
        """
        self.LOG.warning('requestFreeFromSafing not implemented')
    
    def forceReady(self):
        """Force the mirror to a homed and ready state.
        """
        self.LOG.warning("forceReady not implemented")
        
    def recordPosition(self,pos_pathname):
        """ write a mirror position file to the given pathname.
        """
        if os.path.exists( pos_pathname ):
            os.unlink(pos_pathname)
        # request current position
        
        # FIXME - how to do an encoder position request and translate to angle?
        self.LOG.debug("encoder position readout not implemented")

        # return the encoder position, or None if no information was obtained
        if not os.path.exists( pos_pathname ):            
            self.LOG.warning("no position information from EAERI mirror")
            return None
        try:        
            return int( file(pos_pathname,'rt').read()[1:] )
        except:
            LOG.warning("no position report from mirror")
            return None

class EaeriAutonomic:
    """autonomic functions - - principally hatch, preheating, safing
    Eaeri handles this on its own
    """
    def __init__(self, ear):
        self._eaeri = ear
        self.LOG = logging.getLogger("EaeriAutonomic")

    def setHatch(self,state=True):
        """Change the hatch-enable state on the hatch daemon
        """
        if state: state = '1'
        else: state = '0'
        self._eaeri.acq('-k',state)
    
    def startHatch(self):
        self.LOG.warning("start Hatch not implemented")        
        
    def startMirrorSafing(self):
        """bootstrap mirror safing daemon; return task representing it
        """
        self.LOG.warning('mirror safing not implemented')
        self.mirsaf = None
        return self.mirsaf
        
    def start(self):
        # self.mirsaf.start()
        # self.hatch.start()
        return self
        
    def preheatMirrorMotor(self):
        """preheat the mirror motor. return when the mirror is warm enough to rotate.
        """
        self.LOG.info("preheatMirrorMotor not implemented")
        pass
        
    def kill(self):
        self.LOG.debug('EaeriMirror does not have any daemons to kill')
        pass

class eaeri_acquire_task:
    def __init__(self,proc,fn_log,flog,hax):
        self._proc = proc
        self._fnlog = fn_log
        self._flog = flog
        self._hax = hax
        
    def wait(self):
        self._proc.wait()
        # Move Files to Temp directory
        self._hax()
        # read the log file and cat it to our log
        try:
            self._flog.write( file(self._fnlog,'rt').read() )
        except:
            LOG.warning("unable to transcribe ACQ96NT.LOG to log file")    
        try:
            os.unlink(self._fnlog)
        except:
            LOG.warning("unable to remove old ACQ96NT.LOG")
            
    def kill(self):
        self._proc.kill()


class EaeriInterferometer:
    def __init__(self,ear):
        self._eaeri = ear
        self._acqlog = Config.logStream('acq96.log')

    def startIngestSpectra(self,set_pathname):
        """Fork a process to ingest the a given .SET file pathname, returning a task proxy.
        Generates SPC data in the same directory as the SET file.
        """
        dirname, filename = win32.path_split( set_pathname )
        stem,exten = os.path.splitext( filename )
        import acqeaeri_hax, subprocess
        hax = acqeaeri_hax.acqeaeri_file_hax( Config.ACQEAERI_EXE, win32.unix_path(dirname) )
        acq96 = self._eaeri.acq_start('M', stem, work_dir=dirname)
        # from Interferometer import acquire_task # FUTURE: clean this up
        return eaeri_acquire_task(acq96,win32.path_join(dirname,Config.ACQ96_LOG),self._acqlog, hax)



class acqeaeri:
    """ wraps an instance of the ACQEAERI bridge application as a process-like object
    raise an EAERI_Failure in the case that ACQEAERI returns a nonzero error code
    """
    
    def __init__(self,**kwargs):
        self._acq = X = process(Config.ACQEAERI_EXE, # work_dir = Config.EAERI_BASE_DIR,
                         output_stream=Config.logStream('acqeaeri.out'),
                         output_prefix='acqeaeri', **kwargs )
        
        self.__repr__ = X.__repr__
        self.start = X.start
        self.kill = X.kill
            
    def check(self):
        running, rc = self._acq.check()
        if not running and rc != 0: 
            raise EAERI_Failure("ACQEAERI failure, rc = %s" % rc)
        return running, rc

    def wait(self):
        """wait for a start()ed executable to exit
           returns result-code
        """
        pid,rc = self._acq.wait()
        if rc != 0:
            raise EAERI_Failure("ACQEAERI pid %d failure, rc = %s" % (pid,rc))
        return (pid,rc)
        
    def __call__(self,*args,**kwargs):
        pid,rc = self._acq(*args, **kwargs)
        if rc != 0:
            raise EAERI_Failure("ACQEAERI pid %d failure, rc = %s" % (pid,rc))
        return (pid,rc)
        
    def getpid(self):
        return self._acq.pid


class Eaeri:
    """ Wrapper class for daemons representing the Eaeri services for housekeeping et al
    This object is also a factory for subsystem objects (housekeeping, mirror, etc) with identical interfaces
    It also coordinates any race conditions or locks that need to happen in order to prevent clobberings
    """
    def __init__(self, port = None, simulate = False):
        pass
        
    def acq(self,*args,**kwargs):
        """ invoke ACQEAERI with default parameters and provided additional parameters
        """
        #  '-r%d' % self._cfg.port,
        X = acqeaeri()
        return X(*args, **kwargs)

    def start(self):
        """ bootstrap system daemons needed for EAERI hardware
        """        
        # we boot the FTSW software in order to get it to clean state;
        # mirror movement commands can, for instance, cause an exception if the 
        # FTSW is still ingesting from a previously killed ACQEAERI command
        LOG.info('restarting FTSW_Eaeri software')
        self.acq('-j')  
        # command mirror to zero position to ensure that we can talk to the instrument rack
        LOG.info('mirror positioning test')
        self.acq('-M0') 
        LOG.debug("FTSW_Eaeri restarted")
        # return a kill()-able corresponding to what was started
        return self
    
    def acq_start(self,*args,**kwargs):
        """ invoke ACQEAERI with default parameters and provided additional parameters
        note that acqeaeri waits for the startup of FTSW_Eaeri.jar to complete before proceeding
        """
        # '-r%d' % self._cfg.port,
        X = acqeaeri()
        X.start( *args, **kwargs)
        return X
        
    def kill(self):
        # no correpsonding command for shutdown in acqeaeri
        pass
    
    def subsys_interferometer(self):
        return EaeriInterferometer(self)
                    
    def subsys_housekeeping(self):
        return EaeriHousekeeping(self)
        
    def subsys_mirror(self):
        return EaeriMirror(self)

    def subsys_autonomic(self):
        return EaeriAutonomic(self)
    
    def subsys_gps(self):
        return EaeriGps(self)

def test_hk_simple(args = None):
    from Housekeeping import test_simple
    return test_simple(args, Eaeri(simulate=True).subsys_housekeeping)

def test_hk_record(args = None):
    from Housekeeping import test_record
    return test_record(args, Eaeri(simulate=True).subsys_housekeeping)

# simulate=True since we expect the EAERI software is manually started when testing
def test_interfer(args = None):
    from Interferometer import test_interfer as mytest
    mytest(args, Eaeri(simulate=True).subsys_interferometer)


def test_cycles(args = None):
    from Ingest import test_cycles as mycycles
    ear = Eaeri(simulate=True)
    mycycles( args, Hfactory = ear.subsys_housekeeping, Ifactory = ear.subsys_interferometer, Mfactory = ear.subsys_mirror )

if __name__=='__main__':
    logging.basicConfig()
    LOG.setLevel(logging.DEBUG)
    test_interfer()
    test_hk_simple()
    test_hk_record()
    test_cycles(['1'])
    
