#!/usr/bin/env python
"""
$Id: Housekeeping.py,v 1.23.2.2 2008/03/24 18:51:12 rayg Exp $
Copyright 2005, University of Wisconsin Space Science & Engineering Center

    This file is part of the AERI ingest software.

    The AERI ingest software is free software; you can redistribute it and/or 
    modify it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA



"""

import time, logging
from exceptions import *
import Config
from process import process

log = logging.getLogger( "Housekeeping" )

class HousekeepingFailure(EnvironmentError):
    """Exception thrown when housekeeping system falls over or locks up.
    """
    pass

class Housekeeping:
    """Wrapper class holding functions related to collection of housekeeping data
    """
    
    def __init__(self):
    
        self._day = Config.YYMMDD
        # all of these are needed by AERIHK
        env = { 'MIRROR_SCD': Config.MIRROR_SCD, 
                'MIRROR_BEG': Config.MIRROR_BEG,
                'AERI_POS': Config.AERI_POS,
                'INST_ID': Config.INST_ID,
                'MIR_POS': Config.MIR_POS,
                'WORK_DIR': Config.workDir(),
                'CHANNEL_MAP': Config.CHANNEL_MAP,
                'AERIHK_CFG': Config.AERIHK_CFG
                }
                
        self.hresd = process(Config.HRESD_EXE,
                             output_stream=Config.logStream('hresd.out'),
                             output_prefix='hresd' )
        self.hresdd = process(Config.HRESDD_EXE,
                              output_stream=Config.logStream('hresdd.out'),
                              output_prefix='hresdd' )
        self.aerihk = process(Config.AERIHK_EXE,
                              environ = env,
                              output_stream=Config.logStream('aerihk.out'),
                              output_prefix='aerihk' )
        # FIXME: This could generate a pretty big file; one chunk of text every 30 sec
        self.liveness = process(Config.HRESVIEW_EXE,  
                                output_stream=file(Config.workPath('hkcheck.out'),'wt'))

    def start(self): 
        """Bootstrap the housekeeping collection software.
        """
        log.info("starting hresd")
        self.hresd.start()
        time.sleep(2)
        log.info("starting hresdd")
        self.hresdd.start()
        time.sleep(2)
        # sync up and check for housekeeping
        log.info("checking housekeeping")
        self.peek()
        # 20050831 -- we've seen issues with NaNs creeping into restart data.
        # FIXME: Make sure that we've got initial data.
        self.peek()
        log.info("starting aerihk")
        self.aerihk.start()

        log.info("ready")
                
        # return ourself as a tasks object. FIXME: consider whether this is appropriate or not
        # for the moment it's very practical since these tasks are by nature singleton 
        return self
        
    def peek(self):
        """Read a record from the high-resolution housekeeping to the screen
        """
        #FIXME: either update hresview to output pythonic or XML format, or make it an extension
        q = process( Config.HRESVIEW_EXE, environ = { 'CHANNEL_MAP': Config.CHANNEL_MAP } )
        q()
      
    def syncAndMarkStartOfSet(self, ifgTask=None):
        """ Synchronize the start of housekeeping collection to the start of interferogram acquisition.
        Start AERIHK recording the data for this acquisition.
        Return no sooner than the start of the interferogram acquire.
        FIXME: Currently does not use pipe synchronization to ACQ96 to find the start-of-scan.
        """
        q = process(Config.AESCNMRK_EXE)
        q()
        
        # start a background job to check the liveness of the housekeeping subsystem
        # FIXME: we shouldn't need this
        self.liveness.start()
        return self
      
    def writeHousekeeping(self, pathname):
        """ Write the buffered housekeeping residing in AERIHK buffers to the pathname provided.
        FIXME: the .SET file gets read based on this name. is this a problem?
        raises EnvironmentError if any problem collecting housekeeping occurred
        """
        pid,rc = self.liveness.wait() # should succeed if at least one HK set has been recorded
        if rc!=0:
            log.critical( "housekeeping subsystem failure! rc=%d" % rc )
            raise HousekeepingFailure("incomplete/timeout response from housekeeping query")
        q = process( Config.AEWRTHK_EXE )
        q( pathname )

    def kill(self):
        """Kill the tasks we have running.
        """
        # graceful shutdown request
        log.info("signalling aerihk to exit")
        sihk = process( Config.AERISIHK_EXE )
        sihk()
        
        time.sleep(1)
        log.info("killing hresd")
        self.hresd.kill()
        log.info("killing hresdd")
        self.hresdd.kill()
        log.info("killing aerihk")
        try: 
            self.aerihk.kill()
        except OSError, err:
            log.info("aerihk already terminated (%s)" % `err`)
            pass
            

        
def test_simple(args, Hfactory = Housekeeping):
    """ Start up the housekeeping subsystem, take a few samples, shut it down.
    """
    H = Hfactory()
    tasks = H.start()
    H.peek()
    H.peek()
    tasks.kill()

def test_record(args, Hfactory = Housekeeping):
    """ Start up the housekeeping system, record 30 seconds worth of data, write it to an SPC file
    """
    H = Hfactory()
    tasks = H.start()
    H.syncAndMarkStartOfSet()
    print "collecting 30 seconds of housekeeping data.. please wait."
    time.sleep(30)
    fn = Config.workPath( time.strftime('%Y%m%d_%H%M%S_') )
    print "writing file " + fn
    H.writeHousekeeping( fn )
    tasks.kill() #currently equivalent to H.kill


if __name__=='__main__':
    doc = """
python Housekeeping.py simple - Start up the housekeeping subsystem, 
                                take a few samples, shut it down.
python Housekeeping.py record - Start up the housekeeping system, record 
                                30 seconds worth of data, write it to an SPC 
                                file in %s
""" %Config.workDir()

    from sys import argv
    logging.basicConfig()
    tests = { 'simple': test_simple,
              'record': test_record
              }
              
    if len(argv) > 1 and argv[1] in tests.keys():
        tests[argv[1]](argv[2:])
    else:
        print doc
