#!/usr/bin/env python
"""
$Id: Ingest.py,v 1.58.2.7 2008/09/26 04:18:01 rayg Exp $
Copyright 2005, University of Wisconsin Space Science & Engineering Center

    This file is part of the AERI ingest software.

    The AERI ingest software is free software; you can redistribute it and/or 
    modify it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA



"""

import logging,sys,signal,time,os

from process import process

#-------------------------------------
# Major subsystems. 

# system configuration - paths and templates
import Config
import win32

# sequencer, creates and maintains the file sets which will be acquired
from Sequencer import Sequencer
# interactive real-time command set
#FIXME from Command import Command
# disk drive space care and feeding, excluding work directory which is owned by sequencer
from Tidy import Tidy
from Housekeeping import HousekeepingFailure

# interferometer subroutines 
from Interferometer import Interferometer
# mirror rotation and reporting
from Mirror import Mirror
# housekeeping data collection
from Housekeeping import Housekeeping
# autonomic systems - mirror safing et al
from Autonomic import Autonomic
# Gps access
from Gps import Gps

# processing software 
from Science import Science
# display and interactive systems 
from Jada import Jada
# text interface for ingest 
from Status import Status
# Watchdog Timer
from Watchdog import Watchdog

log = logging.getLogger( "Ingest" )


def attach_hardware():
    """return base, housekeeping, mirror, autonomic, interferometer, and GPS objects for a given AERI type
    for the AERI, we use the "original" objects; for the EAERI we instance the Eaeri object and get its subsystem wrappers
    """
    if not Config.IS_EAERI: 
        # start up the housekeeping system; returns when we have good housekeeping, or raises exception
        hsk = Housekeeping()
        # get the mirror programmed, preheated up, checked out 
        mir = Mirror()
        # set up the autonomic systems - mirror safing, hatch control 
        autonom = Autonomic() 
        # hook up the interferometer reading routines
        ifg = Interferometer()
        # grab a GPS instance
        gps = Gps()
        base = None
    else:
        # EAERI hardware support
        from Eaeri import Eaeri
        base = Eaeri()
        hsk = base.subsys_housekeeping()
        mir = base.subsys_mirror()
        ifg = base.subsys_interferometer()
        autonom = base.subsys_autonomic()
        gps = base.subsys_gps()
    return base, hsk, mir, autonom, ifg, gps

class SleepingPuppy:
    """replacement watchdog timer which doesn't so much watch anything"""
    def reset(self):
        pass
    def stop(self):
        pass

class Ingest:

    def __init__(self,tidy_instance,force_newday=False):
        """ Create Ingest instance, which contains proxies for all the subsystems.
        requires a tidy object which has cleared the way for us to run
        """
        if Config.IS_EAERI:
            self.aeri_type = 'EAERI' 
        else:
            self.aeri_type = 'AERI'

        self.tidy = tidy_instance

        log.info( "Starting Watchdog Timer" )
        if not Config.IS_EAERI:
            self.wdt = Watchdog( )
        else:
            self.wdt = SleepingPuppy( )

        # set up the sequencing information in the working directory if necessary
        self.seq = Sequencer(force_newday)
        # initialize the science software directories
        self.sci = Science()
        # now that we have working directories, we can consider a GUI
        self.jada = Jada()
        # text status user interface, either as part of Jada or through a curses interface
        # FUTURE: this may take a jada interface as part of its initializer
        self.doing = Status()

        self.base, self.hsk, self.mir, self.autonom, self.ifg, self.gps = attach_hardware()
        if self.base: # then we have a systems daemon to start for EAERI
            self.base.start()
        
        # pretty-printer for sequence numbers
        self.banner = process( '/usr/bin/banner' )

    def startAll(self):

        spaceNeeded = self.seq.estimateSpaceRequirement()
        log.info("%dM needed on drive" % (spaceNeeded/1024**2))

        # start clearing out old data so that we have enough space for today
        # return when we can at least create configuration files
        log.info( "tidying up to clear space.." )
        tidyTask = self.tidy.startClearSpace( spaceNeeded )

        # set up work directories, preparing for new day if necessary
        resume_day = self.seq.configureDirectories()
        if resume_day:
            log.info("Incomplete ingest for day %s detected on disk" % resume_day)
        else:
            log.info("Configuring science work directories for newday")
            self.sci.configureDirectories()
            log.info("Writing PAR file")
            self.tidy.writeParFile()

        # start ingest messaging server
        log.info( "starting ingest XML message server" )
        self.doing.startMessaging()

        # start qc log task logging warning on fail
        self.doing.startQcLog()

        # program the mirror controller while we're waiting
        log.info("programming mirror controller")
        self.mir.programMirrorController()

        # bring up the housekeeping subsystem so that we can preheat
        # returns only when we can start reading housekeeping
        log.info("starting up housekeeping subsystem")
        self.hskTasks = self.hsk.start()

        # start the mirror preheat
        log.info("preheating mirror motor")
        # preheatTask = self.mir.startMotorPreheat() # FIX ME: implement

        # start up the mirror status daemon to share commands and mirror status between exes
        self.mirTasks = self.mir.start()

        # set Jada up to monitor the science and ingest
        log.info("creating configuration files for Jada")
        self.jada.configureDirectories()
        log.info("bootstrapping Jada user interface")
        self.jadaTask = self.jada.start()

        log.info("cleaning out incomplete uncalibrated data from science work directory")
        # take care of anything with a .FIN 
        self.tidy.cleanFinishedSpcFiles()
        # take care of anything with just a .SIG
        self.tidy.cleanUnusedSpcFiles()

        # wait for the disk clean up to finish off before letting us start science
        # if neccessary
        if tidyTask:
            log.info("waiting for drive cleanup to complete")
            tidyTask.wait()

        # returns when apps are started
        log.info("bootstrapping science pipeline")
        self.sciTasks = self.sci.startPipeline()

        # wait for the mirror preheat to complete
        log.info('waiting for mirror motor preheat to complete')
        #preheatTask.wait() # FIXME
        self.mir.forceReady() # FIXME: this goes away when preheating is implemented

        # start up mirror safing daemon and hatch
        self.autoTasks = self.autonom.start()
        
        # start the java mirror monitor
        log.info('starting the mirror monitor')
        self.doing.startMirrorMon()
        
        # create a CSV file for timings
        fnTimings = Config.calPath("stats.csv")
        tfexisted = os.path.exists(fnTimings)
        self.fpTimings = file(fnTimings,'at')
        if not tfexisted:
            print >>self.fpTimings, 'view,scans,acq96_seconds'

        log.info('AERI ready!')

    def ingestOneSet(self):
        """Main AERI ingest sequence.
        """

        # get the sequence data handle
        setInfo = self.seq.getCurrentSet()
        self.doing.showSet("Starting Set %s" % setInfo)

        # FIXME: implemented, needs testing
        # send ingest XML status message
        try:
            log.info( "Sending XML status message" )
            self.doing.sendMessage( setInfo )
        except IOError, e:
            log.debug( "could not send message: %s", e )

        self.doing.showMirror( "Mirror angle: %s" % setInfo.mirrorAngle() )
        # if the mirror is safed, now is the time to release it if we can
        self.mir.requestFreeFromSafing( )
        # preparation: synchronously rotate the mirror to the indicated sequence position
        self.mir.rotateToPosition( setInfo.mirrorPosition() )

        # preparation: create the .SET file
        #FIXME: this should be done by Interferometer, not Sequencer
        setInfo.createAcqFile()
        self.doing.finished(".SET file for ACQ96 ready")

        try:
            self.banner( setInfo.name() )
        except:
            print "Acquiring data for %s" % setInfo.name()

        # tell the interferometer to get the data, returning immediately
        tacq = time.time()
        ifgTask = self.ifg.startIngestSpectra(setInfo.acq96Pathname())
        log.info("Starting acquisition")

        # start the housekeeping system recording data for this set, syncing to the interferometer 
        # returns when the interferometer is actually reading data
        hskTask = self.hsk.syncAndMarkStartOfSet(ifgTask)
        self.doing.finished("reading housekeeping")

        # wait for the interferometer to complete
        ifgTask.wait()
        tacq = time.time() - tacq
        self.doing.finished("acquire complete")
        print >> self.fpTimings, "%s,%d,%f" % (setInfo.name(),setInfo.interferometerScans(),tacq)

        # catch additional housekeeping and record the pending housekeeping to disk
        self.mir.recordPosition(Config.MIR_POS)
        
        # get GPS position and time sync
        self.gps.writeAeriPos()
        
        # collect and write housekeeping
        hskTask.writeHousekeeping(setInfo.housekeepingPathname())
        self.doing.finished("writing housekeeping")

        # create signal files and take care of other bundling issues
        setInfo.makeSetBundle()

        # tell the sequencer to move on to the next view
        newSet = self.seq.incrementSet()
        self.doing.finished("incremented to %s set" % newSet)

        # transfers the set over to the science software
        self.tidy.moveSpcFilesToScience()
        self.doing.finished("data for %s transferred to science data processing" % setInfo.name() )

        # moves calibarted spc files to zip file
        self.tidy.cleanFinishedSpcFiles()
        
        self.fpTimings.flush()


    def ingestLoop( self, sequences_count = None ):
        """Ingest a series of spectra and associated housekeeping
        """
        self.wdt.reset() # reset the watchdog before beginning
        while None==sequences_count or sequences_count>0:
            try:
                self.ingestOneSet()
            except HousekeepingFailure:
                if Config.IS_EAERI:
                    log.warning("housekeeping failure, will attempt to re-ingest current set", exc_info=True)
                else:
                    log.warning("attempting housekeeping restart and re-ingest of current set", exc_info=True)
                    self.hskTasks.kill()
                    self.autoTasks.kill()
                    self.hskTasks = None
                    self.autoTasks = None
                
                    log.debug("initializing new housekeeping subsystem")
                    self.hsk = Housekeeping()
                    self.hskTasks = self.hsk.start()
                    log.info("housekeeping restarted")
                    self.autonom = Autonomic()
                    self.autoTasks = self.autonom.start()                
                    log.info("mirror safing and hatch restarted")
                
                # attempt to re-ingest the same set, since we raised before incrementSet
                continue
            # reset the Watchdog timer if we got this far
            self.wdt.reset()
            if sequences_count: sequences_count -= 1
        # stop the WDT if we ran to completion
        log.info("done!")
        self.wdt.stop()

def test_newday(args):
    """Initialize the working directories for Sequencer, Science, Jada
    """
    T = Tidy()
    T.configureDirectories()

    S = Sequencer()
    # set up work directories, preparing for the new day if necessary
    S.configureDirectories()

    C = Science()
    # create the FTP and science workign directories
    C.configureDirectories()


def test_cycles(args, Sfactory = Sequencer, Hfactory = Housekeeping, Ifactory = Interferometer, Mfactory = Mirror):
    """minimally run a partial ingest sequence using Sequencer, Housekeeping, Mirror, Interferometer
    takes one argument, number of cycles to run.
    Data is collected as SPC files in Config.workDir()
    """
    ncycles = len(args)==0 and 1 or int(args[0])

    S = Sfactory()
    H = Hfactory()
    I = Ifactory()
    M = Mfactory()

    # bootstrap the housekeeping and reset the mirror
    hskDaemons = H.start()
    M.forceReady()

    while ncycles>0:
        ncycles -= 1
        sinf = S.getCurrentSet()
        print '-'*40
        print sinf
        print '>> ROTATE'
        M.rotateToPosition(sinf.mirrorPosition())
        time.sleep(5) # FIXME shouldn't need this
        print '>> CREATE .SET'
        sinf.createAcqFile()
        print '>> START ACQ96'
        ifgTask = I.startIngestSpectra(sinf.acq96Pathname())
        print '>> MARK HOUSEKEEPING'
        hskTask = H.syncAndMarkStartOfSet(ifgTask)
        print '>> AWAIT ACQ96 OUTPUT'
        ifgTask.wait()
        print '>> MIRROR POSITION REPORT'
        M.recordPosition(Config.MIR_POS)
        print '>> WRITE HOUSEKEEPING FILE'
        hskTask.writeHousekeeping(sinf.housekeepingPathname())
        print '>> TIE OFF SET WITH SIG FILE'
        sinf.makeSetBundle()
        print '>> SEQUENCE TO NEXT SET'
        S.incrementSet()
    print '-'*40
    hskDaemons.kill()
    print 'done.'

def test_clean(args):
    """minimally run a partial ingest sequence using Sequencer,
        Housekeeping, Mirror, Interferometer, Tidy, Science
    takes one argument, number of cycles.
    """

    #FIX ME die if there is only one argument
    ncycles = len(args)==0 and 1 or int(args[0])

    S = Sequencer( )
    H = Housekeeping()
    I = Interferometer()
    M = Mirror()
    T = Tidy( )
    C = Science( )

    T.configureDirectories()

    # start science pipeline
    C.startPipeline()

    # bootstrap the housekeeping and reset the mirror
    hskDaemons = H.start()
    M.forceReady()

    while ncycles>0:
        ncycles -= 1
        sinf = S.getCurrentSet()
        print '-'*40
        print sinf
        print '>> ROTATE'
        M.rotateToPosition(sinf.mirrorPosition())
        print '>> CREATE .SET'
        sinf.createAcqFile()
        print '>> START ACQ96'
        ifgTask = I.startIngestSpectra(sinf.acq96Pathname())
        print '>> MARK HOUSEKEEPING'
        hskTask = H.syncAndMarkStartOfSet(ifgTask)
        print '>> AWAIT ACQ96 OUTPUT'
        ifgTask.wait()
        print '>> MIRROR POSITION REPORT'
        M.recordPosition(Config.MIR_POS)
        print '>> WRITE HOUSEKEEPING FILE'
        hskTask.writeHousekeeping(sinf.housekeepingPathname())
        print '>> TIE OFF SET WITH SIG FILE'
        sinf.makeSetBundle()
        print '>> SEQUENCE TO NEXT SET'
        S.incrementSet()
        T.moveSpcFilesToScience()
        T.cleanFinishedSpcFiles()

    print '-'*40
    hskDaemons.kill()
    print 'done.'

def goodbye(signum,frame):
    """handle SIGTERM gracefully, killing off children and exiting
    """
    log.warning( "received TERM signal - exiting (gracefully)!" )
    sys.exit()


def go():
    # set up logging - phase 1
    # This is where the root logger level is set
    # only drag really nasty stuff into the root log
    rootLogFilename = Config.ERROR_LOG

    logformat = '[%(name)s %(asctime)s %(levelname)s] %(message)s'

    # the error log needed to have it's own filehandler. 
    rootBucket = logging.FileHandler(rootLogFilename,mode='at')
    rootFormatter = logging.Formatter(logformat,'%y%m%d %H:%M:%S')
    rootBucket.setFormatter(rootFormatter)
    rootBucket.setLevel(logging.ERROR)
    logging.getLogger().addHandler(rootBucket)

    # set up SIGTERM handler to allow kill to close ingest
    signal.signal( signal.SIGTERM, goodbye )

    tidy = Tidy()

    # set up SIGHUP to cause us to clean up and exit at end-of-day
    def hastalavistababy( signum,frame,T=tidy,lofi=log ):
        lofi.info("received HUP signal; doing end-of-day shutdown")
        T.rebuildOutgoingDirectory()
        lofi.info("AERI shutting down")
        sys.exit()
    signal.signal( signal.SIGHUP, hastalavistababy )

    # set up work directories, preparing for new day if necessary
    # FIXME: This is a lame thing to do, but is necessary because the 
    # log files for the daemons go into the work directories.
    # Find a better way to handle this, e.g. not creating log files in initializers?
    log.info( "clearing enough space to make directories" )
    task = tidy.startClearSpace( 0 )
    if task:
        log.error( "Tidy shouldn't have to start a task to clear initial space!" )
        task.wait()

    # doublecheck that we're wearing pants
    tidy.sanityCheckBaseDirectories()

    # write ingest pid to a file in the work directory
    file(win32.path_join(Config.workDir(),'Ingest.pid'),'wt').write( '%d\n' % os.getpid() )

    log.info( "creating work directories" )
    created_dirs = tidy.configureDirectories()

    # set up logging - phase 2, now that we have space and directories
    # This is where the root logger level is set
    logFileName = win32.path_join(Config.calDir(Config.YYMMDD),'ingest.log')
    bucket = logging.FileHandler(logFileName,mode='at')
    formatter = logging.Formatter(logformat,'%H:%M:%S')
    bucket.setFormatter(formatter)
    bucket.setLevel(logging.DEBUG)
    logging.getLogger().addHandler(bucket)

    # Set the logging level at the top of the hierarchy to DEBUG
    # allowing all log entries to be logged. The file handlers 
    # created above have the option of additional filtering
    # above the root DEBUG level. 
    logging.getLogger().setLevel( logging.DEBUG )

    # make sure all exceptions that get to this point are logged
    log.info("AERI STARTUP")
    try:
        AERI = Ingest(tidy,force_newday = created_dirs)
        AERI.startAll()
        AERI.ingestLoop()
    except SystemExit, e1:
        # No need to log system exits
        log.debug("System Exit")
    except Exception, e:
        log.error( e, exc_info=True )

if __name__=="__main__":
    from sys import argv
    doc="""
[YYMMDD=<yymmdd>] python Ingest.py newday - Initialize the working directories for 
                            Sequencer, Science, Jada.
[YYMMDD=<yymmdd>] python Ingest.py cycles <#ofcycles> - Minimally run a partial ingest 
                            sequence using Sequencer, Housekeeping, Mirror, Interferometer.
                            SPC files are written to %s
[YYMMDD=<yymmdd>] python Ingest.py clean <#ofcycles> - minimally run a partial ingest 
                            sequence using Sequencer, Housekeeping, Mirror, Interferometer, 
                            Tidy, Science.
""" % Config.workDir()

    tests = { 'newday': test_newday,
              'cycles': test_cycles,
              'clean': test_clean
              }
                  
    if len(argv)==1: # normal run
        go()

    elif (len(argv)==2 and argv[1]=='help') or argv[1] not in tests.keys():
        print doc

    else: # subsystem test
        logging.basicConfig()
        log.setLevel(logging.DEBUG)

        tests[argv[1]](argv[2:])

