#!/usr/bin/env python
"""
$Id: Jada.py,v 1.53 2006/08/29 06:16:17 brucef Exp $
"""
import shutil, scr2xml, logging, signal, os, sys
import string, win32, Verify, Config
from process import process
from os.path import exists, isfile

log = logging.getLogger("Jada")

class Jada:

    def __init__( self, logLevel = 'debug' ):
        """
        logLevel - the log level of the JADA application logging.
        """
        self._logLevel = logLevel
        
        # template files used to write config
        self._loggerCfgTemplate = win32.path_join(Config.configDir(), Config.JADA_LOG_CFG)
    
        # actual config files after processing templates
        self._instsFile = win32.path_join(Config.configDir(), Config.JADA_INSTS_FILE)
        self._loggerCfgFile = win32.path_join(
                                   Config.calDir(Config.YYMMDD),
                                   Config.JADA_LOG_CFG[0:Config.JADA_LOG_CFG.rfind('.template')])
        self._instCfgFile = win32.path_join( Config.outputDir(), '%s.xml' %Config.instrumentName())
        self._jadaLogFile = win32.path_join( Config.calDir( Config.YYMMDD ), Config.JADA_LOG_FILE )

    def configureDirectories(self, dryrun = None ):
        """Setup directory structure for JADA. Config files are copied from Config.configDir() to
        their appropriate location. Config template files are written with values provied by
        Config.py JADA_* properties. The instrument XML configuration file is generated from 
        SCR files in Config.configDir() and written to Config.outputDir(). Errors which occur
        durring copy or file write operations are logged as warnings. Errors that occur durring
        file reads cause termination since the system is missing config files.
        """
        scr_files = [ Config.AESITTER_SCR, Config.RADIANCE_SCR, Config.SUMARY_SCR ]

        # Verify the AEYYMMDD dirs exist and can be written to
        caldir = Config.calDir(Config.YYMMDD)
        if not dryrun and not exists( caldir ) and not os.access( caldir, os.W_OK ):
            log.error("cal dir (%s) does not exist" %caldir)
            raise EnvironmentError("cal dir (%s) does not exist" %caldir)
        outdir = Config.outputDir(Config.YYMMDD)
        if not dryrun and not exists( outdir ) and not os.access( outdir, os.W_OK ):
            log.error("Output dir (%s) does not exist" %outdir)
            raise EnvironmentError("Output dir (%s) does not exist" %outdir)

        # copy files from config directory to output dir
        for filename in scr_files:
            dest_file = win32.path_join( Config.outputDir(Config.YYMMDD), 
                                         win32.path_split(filename)[1] )
                                         
            log.info("Copying %s to %s" %( filename, dest_file ) )
            if not dryrun:
                shutil.copyfile( filename, dest_file )

        log.info('Generating %s.xml insrument config in %s' 
                %(Config.instrumentName(), Config.outputDir()) )
        if not dryrun:
            # write xml config file to the Config.outputDir() directory
            scr2xml.writeXml( name = Config.instrumentName(),
                              out_dir = win32.path_split( self._instCfgFile )[0],
                              scrFiles = scr_files,
                              local='_LOCAL_' )

            # replace flag value to set whether this aeri is local
            scrfile = open( self._instCfgFile, 'rt').read()
            scrfile = string.replace( scrfile, '_LOCAL_', 'true' )
            open( self._instCfgFile, 'w').write(scrfile)

        # replace flags for log file location and debug level
        # write logger.xml to the science working directory
        log.info("Writing %s config template to %s" 
                             %(self._loggerCfgTemplate, self._loggerCfgFile))
        if not dryrun:
            logXml = open( self._loggerCfgTemplate, 'rt').read()
            logXml = string.replace(logXml,'_LOG_FILE_', self._jadaLogFile )
            logXml = string.replace(logXml,'_LOG_LEVEL_', self._logLevel )
            # FIXME This has to be done so the JVM will be able to locate 
            # the files specified under cygwin
            logXml = string.replace(logXml,'\\','\\\\')
            
            # trim '.template' from filename
            open(self._loggerCfgFile,'wt').write( logXml )

        # copy xml logfile dtd to science working directory
        log.info("Copying log4j.dtd to %s" %Config.calDir( Config.YYMMDD ))
        if not dryrun:
            shutil.copyfile( win32.path_join(Config.configDir(), 'log4j.dtd'),
                            win32.path_join(Config.calDir( Config.YYMMDD ), 'log4j.dtd') )

    def start(self, jvm_args=[]):
        """ Create process for JVM that runs JADA. The library path must be set to the 
        location of the DMV JNI libraries. The classpath must be set to include all the
        library dependancies and jada.jar. The Xm* options set the starting and max
        heap sizes to compensate for the large amount of memory needed. 
        All java processes should log to jvm.out with the application name as a prefix.
        
        jvm_args - additional arguments to pass to the Java JVM
        
        Returns a started process
        """
        # instruments file
        inst_cfg = Config.configPath( Config.JADA_INSTS_FILE ) 
        
        jvm_args += ['-Djava.library.path=%s' %Config.JADA_LIB,    # JVM
                      '-Xms25m', '-Xmx128m', '-jar', 
                      win32.path_join(Config.JADA_APP_ROOT,'jada.jar'), # JVM
                      '--log_cfg', Config.calPath('logger.xml'),   # JADA
                      '--inst_cfg', inst_cfg,                      # JADA
                      '--yymmdd', Config.YYMMDD ]                  # JADA
        
        self.jvm = process(Config.JAVA_EXE,
                           arguments=jvm_args,
                           work_dir=Config.JADA_APP_ROOT,
                           output_stream = Config.logStream('jvm.out'),
                           output_prefix = 'jada')
                           
        log.debug('starting JVM: args=%s work_dir=%s' %(' '.join(jvm_args), Config.JADA_APP_ROOT))
        self.jvm.start()
        log.info("JADA Started")
        return self.jvm

    def __repr__(self):
        ret = '[JADA name: %s, day: %s, data_dir: %s logger_cfg:%s inst_cfg:%s ]'
        return  ret % (Config.instrumentName(),
                       Config.YYMMDD,
                       Config.outputDir( Config.YYMMDD ),
                       self._loggerCfgFile,
                       self._instCfgFile)

    def kill(self):
        """Kills the JADA JVM process.
        """
        self.jvm.kill()
        log.info("JADA killed")

def test_datasource(args):
    """ Starts a JMV process that watches for SUM file updates. When data is updated it's
    printed to stdout. Info lines are written with a leading '#' so data may be written to 
    file for later parsing.
    
    args - a list of valid DMV ids to watch
    """

    # add ids to jvm arguments
    jvm_args = ['-Djava.library.path=%s' %Config.JADA_LIB,
                '-cp', Config.JADA_CLASSPATH,
                'edu.wisc.ssec.jada.AdsDiagnose',
                Config.instrumentName(),
                Config.outputDir(),
                Config.YYMMDD] + args

    jvm = process( exe = Config.JAVA_EXE,
                   arguments = jvm_args,
                   work_dir=Config.JADA_APP_ROOT)

    jvm()
    log.info("Datasource test started")

def test_config_dirs(args):
    """Calls configureDirectories()
    """
    j = Jada()
    j.configureDirectories(dryrun=True)

def verifyInstall(args = None):
    """Verify install for the JADA application.

Checks for Config.JAVA_EXE, uses Version.class to verify java libraries are present
on the classpath; a JVM error indicates a dependency problem. Verifies Config.configDir()
exists and the template config files and logger dtd files are present. Checks for the java
libraries are in the correct location specified in Config.JADA_LIB verifies the DMV
JNI library is executable.
    """
    configDir = Config.configDir()
    install_ok = False

    # check for java exe and make double sure it's not a dir
    if not exists( Config.JAVA_EXE ) and isfile( Config.JAVA_EXE ):
        log.error("No Java executable in %s" %Config.JAVA_EXE)
        install_ok = True
    else:
        # Version.class perfoms and import from each JADA dependency to 
        # make sure they are available and also prints Java and Java3D version
        # info. If an error occurs there is a dependency problem.
        jvm = process( exe = Config.JAVA_EXE,
                       arguments = [ '-cp', Config.JADA_CLASSPATH,
                                     'edu.wisc.ssec.jada.Version',
                                     '-version:%s' %Config.JAVA_VERSION ],
                       work_dir = Config.JADA_APP_ROOT)

        #start and wait
        jvm()

    # Verify Config dir exists and template cfg files and dtd files are present
    if not exists( configDir ):
        log.error("Config dir (%s) does not exist" %configDir)
        install_ok = False
    # does the log4j dtd file exist
    if not exists( win32.path_join(configDir, 'log4j.dtd') ):
        log.error("Log4J config dtd not in %s" %configDir)
        install_ok = False
    # does the instruments config file exist
    if not exists( win32.path_join(configDir, 'instruments.cfg' ) ):
        log.error("Instruments config file not in %s" %configDir)
        install_ok = False
    # does the logger config file exist
    if not exists( win32.path_join(configDir, 'logger.xml.template' ) ):
        log.error("Logger config template file not in %s" %configDir)
        install_ok = False

    # is the dmv library dll executable??
    if not (os.stat( win32.path_join(Config.JADA_LIB,'dmvJNI.dll') )[0] & 0100) == 0100:
        log.error( 'DMV Windows library is not executable' )
        install_ok = False
    
    Verify.checkNoNoChars([ Config.AESITTER_SCR, Config.RADIANCE_SCR, Config.SUMARY_SCR ])
    
    return install_ok

def goodbye(signum,frame):
    """handle SIGTERM gracefully, killing off children and exiting
    """
    log.warning( "received TERM signal - exiting (gracefully)!" )
    sys.exit()

def go( jvm_args=[] ):
    # set up SIGTERM handler to allow kill to close ingest
    signal.signal( signal.SIGTERM, goodbye )
    log.info('Starting JADA')
    j = Jada()
    j.configureDirectories()
    jada = j.start(jvm_args)
    jada.wait()

if __name__ == "__main__":
    doc = """
[YYMMDD=<yymmdd>] python Jada.py [-a JVM_ARG]* - start JADA for given day with 
                                                 the specified JVM arguments.
[YYMMDD=<yymmdd>] python Jada.py --verify_install -- test for files n such
[YYMMDD=<yymmdd>] python Jada.py --config_dirs -- test dir configuration
[YYMMDD=<yymmdd>] python Jada.py --data_src dmv_id* -- test the Java data source
"""

    from optparse import OptionParser
    opts = OptionParser(doc)
    
    opts.add_option('-a','', dest='jvm_args', action='append', default=[],
                    help="""Argument for the Java JVM. You must specify one -a
for each JVM argument.""")
    opts.add_option('','--verify_install', dest='verify_install',
                    action='store_true', default=False,
                    help='Verify the Jada installtion. See the doc comment')
    opts.add_option('','--config_dirs', dest='config_dirs', action='store_true',
                    default=False, help='Test the configureDirectories routine')
    opts.add_option('','--test_data_src', dest='data_src', action='store_true',
                    default=False, help="""Runs the DMV datasource as a standalone app
for for the provied dmv identifiers. One or more identifier must be provided""") 
                    
    options, args = opts.parse_args()
    
    tests = { 'verify_install': verifyInstall,
              'config_dirs':test_config_dirs,
              'data_src':test_datasource
              }
    
    if options.verify_install:
        logging.basicConfig()
        verifyInstall(args)
    elif options.config_dirs:
        logging.basicConfig()
        test_config_dirs(args)
    elif options.data_src:
        logging.basicConfig()
        test_datasource(args)
    else:
        go( options.jvm_args )

