#!/usr/bin/env python
"""
$Id: MessageServer.py,v 1.9 2005/08/24 02:41:18 rayg Exp $
Copyright 2005, University of Wisconsin Space Science & Engineering Center

    This file is part of the AERI ingest software.

    The AERI ingest software is free software; you can redistribute it and/or 
    modify it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA



"""

import os, time
from cPickle import dumps, loads
from struct import pack,unpack,calcsize
from pprint import pformat,pprint
from string import replace

import Config
import logging

log = logging.getLogger("MessageServer")

SIZE_FMT = '>I'
SPACE = '       '

MSG_HEADER = r"""<?xml version="1.0"?>
<!DOCTYPE ingest [
<!ELEMENT ingest ( mirror? )>
<!-- Contents should conform to a subset of ISO 8601 
  (in particular, YYYY-MM-DD HH:MM:SS Z.  
  Smaller units may be omitted with a loss of precision) -->
<!ATTLIST ingest date CDATA #REQUIRED>

<!-- MIRROR MESSAGE -->
<!ELEMENT mirror ( angle? | status? | step? | position? )><!-- Indicates some status change of the mirror -->
<!ELEMENT angle ( #PCDATA )><!-- A floating point value representing the angle of the AERI mirror -->
<!ELEMENT status ( #PCDATA )><!-- Current status of the mirror as a string -->
<!ELEMENT step ( #PCDATA )><!-- Current step of the mirror as an integer -->
<!ELEMENT position ( #PCDATA )><!-- Position string: sky, clean, safe... -->
]>
<ingest date="_DATE_" >
"""

MSG_FOOTER = "\n</ingest>"
MSG_PROP = "<%s>%s</%s>"

def mirror( msg ):
    """Returns XML for a mirror message according to key/values in msg dict.
    """
    retMsg = "\n"
    for key in msg:
        retMsg += SPACE + MSG_PROP %( key, msg[key], key) + '\n'
    return retMsg

def makeMessage( msgDict ):
    """Outputs an XML message formated for the XmlMessageClient in the JADA package.
    Supported message types should go in type dict and have conrrespo   nding procedure
    to return XML.
    msgDict - { messagetype: { prop:value, ... } ... }
    
    FIXME: date returned is dependent on the os timezone
    """
    types = { 'mirror': mirror }
    date = time.strftime( '%Y-%m-%d %H:%M:%S %Z' )
    retMsg = replace( MSG_HEADER, '_DATE_', date ) + '\n'
    for key in msgDict:
        retMsg += MSG_PROP %( key, types[key]( msgDict[key] ), key ) + '\n'
    retMsg += MSG_FOOTER
    return retMsg

def readFeed(pd):
    """
    Reads data from the feeder and turns it back into object.
    """
    str_size = pd.read( calcsize( SIZE_FMT ) )
    S = pd.read( unpack( SIZE_FMT, str_size)[0] )
    if not S: return '--EOF--'
    return loads(S)        

def serverMain( pd ):
    """Main server method. Data is read from a pipe which delivers pickled
    dictionaries of ingest message data. The dictionaries are unpickled and
    formated in to XML Ingest messages(see AERI/Jada/config/aeriStatus.dtd).
    These messages are served to a single client when it arrives via the pipe.
    If a connection attempt is made while a client is already connected the 
    current client connection is dropped in lieu of the new connection. If an
    error occurs durring sending the client is dropped.
    """
    log.info('server starting')
    from select import select
    import socket
    srvskt = socket.socket( socket.AF_INET, socket.SOCK_STREAM )
    srvskt.bind( ( Config.MSG_HOST, int(Config.MSG_PORT) ) )
    srvskt.setblocking(0)
    srvskt.listen( Config.MSG_MAX_CONN )
    cliskt = None
    
    while True:
        readlist = [ srvskt, pd ]
        # if cliskt: readlist.append( cliskt ) -- don't really need this unless client talks to us
        readable, writable, errored = select( readlist, [], [] )
        
        if srvskt in readable: # then connect to a new client, dumping the old one
            log.info( "server connecting to new client" )
            if cliskt: 
                log.info( "closing old client connection first" )
                cliskt = None
            (cliskt,addr) = srvskt.accept()
            log.info( "connected to %s on port %s" % cliskt.getsockname() )
            
        if pd in readable: 
            # get picked dictionary
            D = readFeed(pd)
            if cliskt: # then format and send on to the client
                try:
                    # get xml message
                    msg = makeMessage( D )
                    # send header
                    cliskt.sendall( pack( SIZE_FMT, len(msg) ) )
                    # send xml message
                    cliskt.sendall( msg )
                    # send EOF
                    cliskt.sendall( pack( SIZE_FMT, 0 ) ) 
                except Exception, e:
                    log.error(e, exc_info=True)
                    log.warn( "unable to send message, closing socket" )
                    cliskt = None
            else:
                log.debug( "server received but could not forward: %s" % D )
                
                
if __name__=='__main__':
    doc="""[MSG_PORT=port] python MessageServer.py fd
    """
    
    from sys import argv
    if len(argv) == 2: # then we're to run as the server, with the pipe fd being provided
        pd = os.fdopen( int(argv[1]), 'r' )
        serverMain(pd)
    else:
        print doc
