#!/usr/bin/env python
"""
$Id: Mirror.py,v 1.20.2.1 2008/08/26 19:05:42 dennyh Exp $
Copyright 2005, University of Wisconsin Space Science & Engineering Center

    This file is part of the AERI ingest software.

    The AERI ingest software is free software; you can redistribute it and/or 
    modify it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA



"""

import os, logging
import Config
from process import process

from mirscd import mirror_schedule

MIR_READY = 0
MIR_SAFED = 1
MIR_CLEANING = 2
MIR_PREHEATING = 3

log = logging.getLogger("Mirror")



class Mirror:
    def __init__(self):
        # initialize the mirror using aemirprg
        pass        
        
    def start(self):
        """Start up mirror daemon - this shared memory server holds the state of the mirror system
        """
        self.server = process(  Config.AEMIRCMD_EXE, arguments = ['server'], 
                                output_stream=Config.logStream('aemircmd.out'),
                                output_prefix="aemircmd" )
        self.server.start()
        return self # as a task, effectively
        
    def kill(self):
        self.server.kill()
        self.server = None
        
    def isReady(self):
        """return boolean, whether or not the mirror is available to rotate
        """
        return True #FIXME

    def getState(self):
        """ return MIR_* constant indicating the state of the mirror
        """
        pass
        
    def programMirrorController(self):
        p = process(Config.AEMIRPRG_EXE, 
                    output_stream=Config.logStream('mirror.out'),
                    output_prefix='aemirprg' )
        p( Config.MIRPRGM_MOT )
        p = process(Config.AERIMIRP_EXE, 
                    output_stream=Config.logStream('mirror.out'),
                    output_prefix='aerimirp' )
        # disable the fault manager
        p( '1XFK0', '-', 'nul' )
        # power up the coils by running sequence 1
        p( '1Z', '-', 'nul' )
       
        
    def rotateToPosition(self,mirror_position,schedule_pathname=None):
        """rotate the mirror go a given scene using the default mirror schedule file in the workdir
        Note that we're not using aerimirc.exe here, which itself is part mirror / part sequencer. 
        This means that the sequencer takes the on the responsibility (and rightfully so) of 
        sequencing the mirror schedule file.
        """
        p = process(Config.AEMIRGO_EXE, 
                    environ = {"MIRROR_SCD": schedule_pathname or Config.MIRROR_SCD},
                    output_stream=Config.logStream('mirror.out'),
                    output_prefix='aemirgo' )
        log.info("rotating to position %s" % mirror_position)
        p(mirror_position[0])
        #FIXME: determine whether it actually made it there, or stalled out
        
    def requestFreeFromSafing(self):
        """ Signal the mirror system that now is a good time to release the mirror from safing.
        In the case that the mirror isn't safed, no-op.
        If the mirror cannot be released from safe, no-op.
        """
        p = process(Config.AEREQMIR_EXE, 
                    output_stream=Config.logStream('mirror.out'),
                    output_prefix='aereqmir' )
        p()
    
    def forceReady(self):
        """Force the mirror to a homed and ready state.
        """
        p = process(Config.AERIMIRP_EXE, 
                    output_stream=Config.logStream('mirror.out'),
                    output_prefix='aerimirp' )
        p('1XR2','-','nul')
    
    def recordPosition(self,pos_pathname):
        """ write a mirror position file to the given pathname.
        """
        if os.path.exists( pos_pathname ):
            os.unlink(pos_pathname)
        p = process(Config.AERIMIRP_EXE, 
                    output_stream=Config.logStream('mirror.out'),
                    output_prefix='aerimirp' )
        p('-','-',pos_pathname)
        # return the encoder position, or None if no information was obtained
        if not os.path.exists( pos_pathname ):            
            return None
        try:        
            return int( file(pos_pathname,'rt').read()[1:] )
        except:
            log.warning("no position report from mirror")
            return None

def test_program(args):
    """Initialize the mirror controller with its default program.
    This will leave the mirror unable to move (preheat state).
    """
    mir = Mirror()
    log.info("Programming mirror controller")
    mir.programMirrorController()        

def test_ready(args):
    """Force the mirror to a home position, making it ready to rotate.
    """
    mir = Mirror()
    log.info("Forcing mirror to home position")
    mir.forceReady()

def test_rotate(args, Mfactory=Mirror):
    """Rotate the mirror to a series of positions using the existing mirror schedule
        Currently overwrites MIRROR_SCD with MIRROR_BEG
    """
    mir = Mfactory()
    import shutil
    shutil.copyfile(Config.MIRROR_BEG, Config.MIRROR_SCD)
    if not args: args = [ 'H', 'A', 'S', 'A', 'H', 'S' ]

    for each in args:
        mir.rotateToPosition(each)

def test_go_angle(args):
    """ Rotate the mirror to a given angle in degrees, using MIRROR.BEG as reference
    """
    mp = mirror_schedule()
    mp.read( file(Config.MIRROR_BEG,'rt') )
    mp.computeOffset()
    print "offset is %d" % mp.offset_count
    p = process( Config.AERIMIRP_EXE, output_stream = Config.logStream('mirror.out'), output_prefix='go_angle' )
    for each in args:
        ang = int(each)
        steps = mp.stepsForAngle(ang)
        print "rotating to step %d for angle %d" % (steps,ang)
        p( '1D%d' % steps, '-', 'nul' )
        p( '1G', '-', 'nul' )
       
# _WORK_DIR=/tmp AERIMIRP_EXE=/bin/echo MIRROR_BEG=$PWD/mirror.beg python Mirror.py go_angle 60
 

if __name__=='__main__':
    doc = """
python Mirror.py program
python Mirror.py ready
python Mirror.py rotate H A S A H S 
python Mirror.py go_angle ### -- rotate the mirror to a given angle, using MIRROR)BEG
        """
    from sys import argv
    logging.basicConfig()
    log.setLevel( logging.INFO )
    tests = {   'program': test_program,
                'ready': test_ready,
                'rotate': test_rotate,
                "go_angle": test_go_angle
                }

    if len(argv) > 1 and argv[1] in tests.keys():
        tests[argv[1]](argv[2:])
    else:
        print doc
