#!/usr/bin/env python
"""
$Id: Science.py,v 1.27 2005/08/24 21:57:48 brucef Exp $
Copyright 2005, University of Wisconsin Space Science & Engineering Center

    This file is part of the AERI ingest software.

    The AERI ingest software is free software; you can redistribute it and/or 
    modify it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA



"""

import Config, os, sys, logging, time, shutil
from win32 import unix_path, path_join, path_split, path_explode, SEP
from process import process

log = logging.getLogger('Science')

class path_pattern(object):
    """A pattern for a path, to be resolved against a settings dictionary.
    """
    def __init__(self,*args):
        self._val = args
        
    def __call__(self,D):
        """ Evaluate the path components with the dictionary; return a unix-compatible full pathname.
        """
        val = []
        for arg in self._val:
            val += path_explode(arg % D)
        return unix_path(SEP.join(val))
        
    def __repr__(self):
        val = []
        for arg in self._val:
            val += path_explode(arg)
        return 'path%s' % (tuple(val),)
    
        

class localfile(path_pattern):
    """a file in the current working directory
    """
    pass
    
class directory(path_pattern):
    """an absolute directory
    """
    def __init__(self,pname):
        path_pattern.__init__(self,pname,'')
    
class calFile(path_pattern):
    """a configuration file which has been copied to the cal directory 
    """
    def __init__(self,pname):
        path_pattern.__init__(self,Config.runningCfgPath(pname))
    
class configFile(path_pattern):
    """a configuration file in the read-only \config directory
    """
    def __init__(self,filename):
        path_pattern.__init__(self,Config.configPath(filename))
        
class outputFile(path_pattern):
    """a configuration file in the \ftp directory
    """
    def __init__(self,filename):
        path_pattern.__init__(self,Config.outputPath(filename))
    
def expand(lines,D):
    """expand an input line or expression against a variable dictionary
    """
    res = []
    for each in lines:
        if callable(each): res.append(each(D))
        else: res.append( each % D )
    return res
    

#def expand(lines,locvars):
#    return [callable(each) and each(locvars) or each % locvars for each in lines]


def rwaspc_input(D):
    """1. DISABLE REAL-TIME MODE? [Y/N]
2. DATE (YYMMDD)
3. Enter input path (directory)
4. Enter output path (directory)
5. File of "STATIC" parameters"""
    # Paths changed to unix type paths to fix ailing restarts
    RWASPC_INPUT = [
        'N',
        '%(yymmdd)s',
        directory('%(input_dir)s'),
        directory('%(output_dir)s'), 
        calFile( Config.RWASPC_SIP ),
        calFile( Config.MIRROR_BEG )
        ]
    return expand(RWASPC_INPUT,D)

def nlapp_input(D):
    """1. ENTER  Name of Complex Input File (CXS)
2. DISABLE REAL-TIME MODE? [Y/N]
3. ENTER  Name of Complex Output File (CXV)
4. ENTER  File name of "STATIC" Input Parameters"""
    NLAPP_INPUT = [
        outputFile('%(yymmdd)s%(direction)s1.CXS'),
        'N',
        calFile('%(yymmdd)s%(direction)s1.CXV'),
        calFile( Config.NLAPP_SIP )
        ]
    return expand(NLAPP_INPUT,D)

def aericalv_input(D):
    """1. ENTER Input File Name
2. ENTER path or file of orig. complex sp.
3. ENTER FIRST 8 CHARS. FOR OUTPUT FILE NAMES
4. DISABLE REAL-TIME MODE? [Y/N]
5. ENTER NAME OF "VARIANCE" INPUT FILE
6. ENTER NAME OF "STATIC-PARS." INPUT FILE
7. ENTER NAME OF SCENE-MIRROR-POSITION FILE
8. ENTER NAME OF BB PAINT EMISSIVITY FILE
9. ENTER NAME OF PREVIOUS-PROGRAM STATUS FILE"""
    if int(D['band']) ==1:
        cxsv = calFile('%(yymmdd)s%(direction)s%(band)s.CXV')
        stat = calFile('%(yymmdd)s%(direction)s%(band)s.SNL')
    else:
        cxsv = outputFile('%(yymmdd)s%(direction)s%(band)s.CXS')
        stat = calFile('RWASPC.STA')

    AERICALV_INPUT = [
        cxsv,
        outputFile('%(yymmdd)s%(direction)s%(band)s.CXS'),
        '%(yymmdd)s%(direction)s%(band)s',
        'N',
        outputFile('%(yymmdd)s%(direction)s%(band)s.UVS'),
        calFile( Config.AERICALV_SIP ),
        calFile( Config.MIRROR_BEG ),
        calFile( Config.AERIEMIS_ASC ),
        stat 
        ]
    return expand(AERICALV_INPUT,D)

def scndirav_input(D):
    """1. ENTER NAME FOR OUTPUT FILE (AVG. SPECTRA)
2. ENTER  LOG File Name
3. DISABLE  REAL-TIME MODE? [Y/N]
4. INPUT FILE (SCAN DIR. 1)
5. FIRST, LAST INPUT RECORDS
6. INPUT FILE (SCAN DIR. 2)"""

    SCNDIRAV_INPUT = [
        calFile('%(yymmdd)sC%(band)s.RLC'),
        calFile('SCNDIRAV.LC%(band)s'),
        'N',
        calFile('%(yymmdd)sB%(band)s.RLC'),
        '1 99999',
        calFile('%(yymmdd)sF%(band)s.RLC'),
        ]
    return expand(SCNDIRAV_INPUT,D)

def ffovcmr_input(D):
    """1. INPUT SPECTRUM
2. Disable REAL-TIME MODE? [Y/N]
3. OUTPUT FILE (CORRECTED SPECTRUM)
4. FILE OF STATIC PARAMETERS"""

    FFOVCMR_INPUT = [
        calFile('%(yymmdd)sC%(band)s.RLC'),
        'N',
        calFile('%(yymmdd)sC%(band)s.RFC'),
        calFile('ffovcmr.si%(band)s')
        ]
    return expand(FFOVCMR_INPUT,D)

def zfli_input(D):
    """1. WANT TO DISABLE REAL-TIME MODE? [Y/N]
2. ENTER NAME OF INPUT FILE
3. ENTER NAME OF OUTPUT FILE
4. FILE OF STATIC PARAMETERS
5. ENTER NAME OF SCENE-MIRROR-POSITION FILE"""

    ZFLI_INPUT = [
        'N',
        '%(yymmdd)sC%(band)s.RFC',
        outputFile('%(yymmdd)sC%(band)s.RNC'),
        calFile('zfli.si%(band)s'),
        calFile( Config.MIRROR_BEG )
        ]
    return expand(ZFLI_INPUT,D)

def aerisum_input(D):
    """ 1. INPUT FILE (MEAN RAD., "*.RNC", CH.1)
 2. INPUT FILE (MEAN RAD., "*.RNC", CH.2)
 3. INPUT FILE (SKY VAR., "*.CSV", CH.1)
 4. INPUT FILE (SKY-VAR., "*.CSV", CH.2)
 5. INPUT FILE OF VARIANCE("*.CVS", CH.1)
 6. INPUT FILE OF VARIANCE("*.CVS", CH.2)
 7. INPUT FILE (CH.1 RESPONSIVITY, "*F1.RES")
 8. INPUT FILE (CH.2 RESPONSIVITY, "*F2.RES")
 9. INPUT FILE (CH.1 DIAG. SPECTRA, "*F1.IMC")
10. INPUT FILE (CH.2 DIAG. SPECTRA, "*F2.IMC")
11. INPUT FILE (CH.1 COMPLEX SPECTRA, "*F1.CXS")
12. INPUT FILE (CH.2 COMPLEX SPECTRA, "*F2.CXS")
13. DISABLE REAL-TIME MODE? [Y/N]
14. ENTER OUTPUT FILE NAME
15. ENTER STATIC-PAR. INPUT FILE NAME
16. ENTER NAME OF SCENE-MIRROR-POSITION FILE"""

    AERISUM_INPUT = [
        outputFile('%(yymmdd)sC1.RNC'),
        outputFile('%(yymmdd)sC2.RNC'),
        outputFile('%(yymmdd)sF1.CSV'),
        outputFile('%(yymmdd)sF2.CSV'),
        localfile('%(yymmdd)sF1.CVS'),
        localfile('%(yymmdd)sF2.CVS'),
        localfile('%(yymmdd)sF1.RES'),
        localfile('%(yymmdd)sF2.RES'),
        localfile('%(yymmdd)sF1.IMC'),
        localfile('%(yymmdd)sF2.IMC'),
        outputFile('%(yymmdd)sF1.CXS'),
        outputFile('%(yymmdd)sF2.CXS'),
        'N',
        outputFile('%(yymmdd)s.SUM'),
        localfile( Config.AERISUM_SIP ),
        localfile( Config.MIRROR_BEG )
        ]
    return expand(AERISUM_INPUT,D)

def rwaspc_processes( day ):
    yymmdd = day
    caldir = Config.calDir(yymmdd)
    input_dir = Config.scienceWorkDir( yymmdd )
    output_dir = Config.outputDir( yymmdd )
    rwaspc = [ process( exe=Config.RWASPC_EXE,
                      input_lines= rwaspc_input(locals()),
                      output_stream=Config.logStream('rwaspc.out'),
                      output_prefix='rwaspc',
                      tag='rwaspc',
                      work_dir = caldir ) ]
    return rwaspc

def nlapp_processes( day ):
    yymmdd = day
    caldir = Config.calDir(yymmdd)
    nlapp = [ process( exe=Config.NLAPP_EXE,
                       input_lines= nlapp_input(locals()),
                       output_stream=Config.logStream("nlapp%(direction)s1.out" % locals()),
                       output_prefix='nlapp',
                       tag='nlapp%s1' % direction,
                       work_dir = caldir ) for direction in [ 'B', 'F' ]]
    return nlapp    

def aericalv_processes( day ):
    yymmdd = day
    caldir = Config.calDir(yymmdd)
    aericalv = [ process( exe=Config.AERICALV_EXE,
                          input_lines= aericalv_input(locals()),
                          output_stream=Config.logStream("aericalv%(direction)s%(band)s.out" % locals()),
                          output_prefix='aericalv',
                          tag = 'aericalv%s%s' % (direction,band),
                          work_dir = caldir )
                          for direction in [ 'B', 'F' ] for band in [ '1', '2' ]]
    return aericalv

def scndirav_processes( day ):
    yymmdd = day
    caldir = Config.calDir(yymmdd)
    scndirav = [ process( exe=Config.SCNDIRAV_EXE,
                          input_lines= scndirav_input(locals()),
                          output_stream=Config.logStream("scndiravC%(band)s.out" % locals()),
                          output_prefix='scndirav',
                          tag = 'scndiravC%s' % band,
                          work_dir = caldir ) for band in [ '1', '2' ] ] 
    return scndirav

def ffovcmr_processes( day ):
    yymmdd = day
    caldir = Config.calDir(yymmdd)
    ffovcmr = [ process( exe=Config.FFOVCMR_EXE,
                         input_lines= ffovcmr_input(locals()),
                         output_stream=Config.logStream("ffovcmrC%(band)s.out" % locals() ),
                         output_prefix='ffovcmr',
                         tag = 'ffovcmrC%s' % band,
                         work_dir = caldir ) for band in [ '1', '2' ] ] 
    return ffovcmr

def zfli_processes( day ):
    yymmdd = day
    caldir = Config.calDir(yymmdd)
    zfli = [ process( exe=Config.ZFLI_EXE,
                      input_lines= zfli_input(locals()),
                      output_stream=Config.logStream("zfliC%(band)s.out" % locals()),
                      output_prefix='zfli',
                      tag = 'zfliC%s' % band,
                      work_dir = caldir ) for band in [ '1', '2' ] ]           
    return zfli

def aerisum_processes( day ):
    yymmdd = day
    caldir = Config.calDir(yymmdd)
    aerisum = [ process( exe=Config.AERISUM_EXE,
                         input_lines= aerisum_input(locals()),
                         output_stream=Config.logStream("aerisum.out"),
                         output_prefix='aerisum',
                         tag = 'aerisum',
                         work_dir = caldir) ]
    return aerisum

class Science:
    """Class to start up calibration subroutines.
    """

    def __init__(self):
        """Create a science proxy object for a given operational day.
        """
        self._day = Config.YYMMDD

    def configureDirectories(self, dryRun = None):
        """Copies config files to use for day to the Config.calDir().
        dryRun - if True only prints out a log statment
        """
        # set up science application configuration files

        # copy running config to Config.calDir()
        cfgFiles = [ Config.configPath( path_split(f)[1] ) for f in Config.PAR_FILES ]
        for src, dest in zip( cfgFiles, Config.PAR_FILES ):
            if src == dest:
                log.info("%s and %s are the same" %( src, dest) )
                continue
            if not os.path.exists( src ):
                log.error("File %s does not exist, cannot copy to the cal directory" %src )
                continue
            #FIXME are all the files in Config.PAR_FILES necessary?
            # what to do when they are not present
            log.info("Copy %s to %s" %( src, dest ) )
            if not dryRun:
                shutil.copyfile( src, dest )

    def startPipeline(self):
        """Bootstrap the science calibration software.
        """
        self._create_processes()
        self._start_processes()
        return self

    def _create_processes(self):
        """Create Science process objects.
        """
        self.rwaspc = rwaspc_processes( self._day )
        self.nlapp = nlapp_processes( self._day)
        self.aericalv = aericalv_processes( self._day )
        self.scndirav = scndirav_processes( self._day )
        self.ffovcmr = ffovcmr_processes( self._day )
        self.zfli = zfli_processes( self._day )
        self.aerisum = aerisum_processes( self._day )
        self.all = self.rwaspc + self.nlapp + self.aericalv + self.scndirav + self.ffovcmr + self.zfli + self.aerisum
        return self

    def _start_processes(self):
        """Start science calibration software.
        """        
        log.info("starting rwaspc process")
        [obj.start() for obj in self.rwaspc]
        time.sleep(2)

        log.info("starting nlapp processes")
        [obj.start() for obj in self.nlapp]
        time.sleep(2)

        log.info("starting aericalv processes")
        [obj.start() for obj in self.aericalv]
        time.sleep(2)        

        log.info("starting scndirav processes")
        [obj.start() for obj in self.scndirav]
        time.sleep(2)

        log.info("starting ffovcmr proceses")
        [obj.start() for obj in self.ffovcmr]
        time.sleep(2)

        log.info("starting zfli processees")
        [obj.start() for obj in self.zfli]
        time.sleep(2)         

        log.info("starting aerisum process")
        [obj.start() for obj in self.aerisum]
        time.sleep(2)
        
        log.info("verifying that processes are running")
        for each in self.all:
            alive,rc = each.check()
            if not alive: 
                log.error( '%s failed on startup!' % each.tag )
                log.warning( "attempting restart of %s" % each.tag )
                each.start()
                time.sleep(1)
        return self
        
    def check(self):
        """Check whether processes are still running
        returns True,None or False,None
        """
        result = True
        for each in self.all:
            alive,rc = each.check()
            if not alive:
                log.error("%s has died" % each.tag)
                result = False
        return result,None
                
    def wait(self):
        """Wait for the tasks to finish
        """
        [obj.wait() for obj in self.all]
        
        return self 
       
    def kill(self):
        """Kill the tasks we have running.
        """
        for each in self.all:
            log.info( 'killing %s' % each.tag )
            each.kill()

        return self 

def test_start_all(args):
    """Starts rwaspc in realtime mode 
    """
    day = Config.YYMMDD
        
    rwaspc = rwaspc_processes( day )
    [obj.start() for obj in rwaspc]
    time.sleep(2)

    nlapp = nlapp_processes( day )
    [obj.start() for obj in nlapp]
    time.sleep(2)

    aericalv = aericalv_processes( day )
    [obj.start() for obj in aericalv]
    time.sleep(2)

    scndirav = scndirav_processes( day )
    [obj.start() for obj in scndirav]
    time.sleep(2)
    
    ffovcmr = ffovcmr_processes( day )
    [obj.start() for obj in ffovcmr]  
    time.sleep(2)  
    
    zfli = zfli_processes( day )
    [obj.start() for obj in zfli]
    time.sleep(2) 

    aerisum = aerisum_processes( day )
    [obj.start() for obj in aerisum]
    time.sleep(2)
   
    [obj.wait() for obj in rwaspc]
    [obj.wait() for obj in nlapp]
    [obj.wait() for obj in aericalv]
    [obj.wait() for obj in scndirav]
    [obj.wait() for obj in ffovcmr]
    [obj.wait() for obj in zfli]
    [obj.wait() for obj in aerisum]

    print "calibration for %s finished"  % day 
    return
    
def test_config_dirs(args):
    """Test the directory configuration routine.
    """    
    s = Science()
    s.configureDirectories(dryRun=True)

def test_reprocessing(args):
    """ Run reprocessing on date
    """
    # Need to make a rwaspc.sta file
    # Start nlapp, aericalv, scndir, ffovc, zfli, aerisum
    # Have directory with static parameter files
    print "Starting nlapp, aericalv, scndir, ffovc, zfli, aerisum"
    print "FIX ME"
    return

def test_read_write(args):
    """ Run read write on some data from a days directory
    """
    day = Config.YYMMDD
    
    # Unzip the zip file
    # os.system( "unzip SPC_AERI_%s.zip" % day )
    # Only use so many records, remove the rest
    # If there aren't that many records say use a different day

    # Remove the fin files
    #os.system( "rm *.fin" )
    # zip up the file you want to use for restart

    # start read write
    rwaspc = rwaspc_processes( day )
    [obj.start() for obj in rwaspc]
    # make sleep time dependent on how many records we are using
    #time.sleep(60)

    # kill read write
    #rwaspc.kill()    

    # add the new data and start read write again
    # add aeri.ctl file
    # what other files are needed

def test_print_inputs( args ):    
    """Prints all the inputs that is to be used for the science processes.    
    """    
    from string import split
    yymmdd = Config.YYMMDD
    cal_dir = Config.calDir( yymmdd )
    input_dir = Config.scienceWorkDir( yymmdd )
    output_dir = Config.outputDir( yymmdd )
    config = Config.configDir()
    
    band = "1"
    direction = "F"
    
    print "RWASPC:"    
    for s,t in zip(rwaspc_input(locals()), split(rwaspc_input.__doc__,'\n')):
        print "\t%s: %s" %(t,s)
    
    for direction in ['B','F']:
        print "NLAPP (%s):" % direction
        for s,t in zip(nlapp_input(locals()), split(nlapp_input.__doc__,'\n')):
            print "\t%s: %s" %(t,s)
            
    for band in ['1','2']:
        for direction in ['B','F']:        
            print "AERICALV (%s%s):" %(direction, band)
            for s,t in zip(aericalv_input(locals()), split(aericalv_input.__doc__,'\n')):
                print "\t%s: %s" %(t,s)
    
    for band in [ '1', '2' ]:    
        print "SCNDIRAV (%s):" % band
        for s,t in zip(scndirav_input(locals()), split(scndirav_input.__doc__,'\n')):
            print "\t%s: %s" %(t,s)

    for band in [ '1', '2' ]:    
        print "FFOVCMR (%s):" % band
        for s,t in zip(ffovcmr_input(locals()), split(ffovcmr_input.__doc__,'\n')):
            print "\t%s: %s" %(t,s)
        
    for band in [ '1', '2' ]:    
        print "ZFLI (%s):" % band
        for s,t in zip(zfli_input(locals()), split(zfli_input.__doc__,'\n')):
            print "\t%s: %s" %(t,s)
        
    print "AERISUM:"
    for s,t in zip(aerisum_input(locals()), split(aerisum_input.__doc__,'\n')):
        print "\t%s: %s" %(t,s)
        

if __name__=='__main__':
    doc="""
python Science.py start_all -- start all for current day
[YYMMDD=<yymmdd>] python Science.py start_all -- start all for given day
[YYMMDD=<yymmdd>] python Science.py reprocess 
[YYMMDD=<yymmdd>] python Science.py read_write
[YYMMDD=<yymmdd>] python Science.py print_inputs
[YYMMDD=<yymmdd>] python Science.py config_dirs - no files are actually copied
"""

    argv = sys.argv
    logging.basicConfig()
    log.setLevel( logging.DEBUG )

    tests = { 'start_all': test_start_all,
              'reprocess' : test_reprocessing,
              'read_write': test_read_write,
              'print_inputs':test_print_inputs,
              'config_dirs':test_config_dirs
              }
              
    if len(argv) > 1 and argv[1] in tests.keys():
        tests[argv[1]](argv[2:])
    else:
        print doc
