#!/usr/bin/env python
"""
$Id: Sequencer.py,v 1.34 2005/08/24 02:41:18 rayg Exp $
Copyright 2005, University of Wisconsin Space Science & Engineering Center

    This file is part of the AERI ingest software.

    The AERI ingest software is free software; you can redistribute it and/or 
    modify it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA



"""

import os, shutil, re, logging
import Config, win32
from process import process

from mirscd import mirror_schedule

log = logging.getLogger("Sequencer")

aeri_set_info_repr_format = """[AERI set: name %(_set_name)s, %(_scan_count)d scans, %(_mirror_angle)d angle, %(_mirror_command)s command]"""

class aeri_set_info:
    """a class representing the information needed in order to set up the AERI subsystems for a data acquire.
    """

    def __init__(self,set_name,scan_count,mirror_angle,mirror_command):
        self._set_name = set_name   # e.g. "H1"
        self._scan_count = scan_count # integer
        self._mirror_angle = mirror_angle # float
        self._mirror_command = mirror_command # string
        
        # derived
        self._hk_filename = set_name # + 'HK.SPC' appended by AERIHK
        self._sig_filename = set_name + '.SIG'
        self._set_filename = set_name + '.SET'
        
    def __repr__(self):
        return aeri_set_info_repr_format % self.__dict__
        
    def name(self):
        return self._set_name
        
    def mirrorAngle(self):  
        """Return the mirror angle to be used by this acquire.
        """
        return self._mirror_angle #FIXME
        
    def mirrorCommand(self):
        return self._mirror_command
        
    def mirrorPosition(self):
        """Return the mirror position name to be used by this acquire.
        """
        return self._set_name[0]
        
    def interferometerScans(self):
        """return the number of dwells scheduled for the acquire of this set.
        """
        return self._scan_count
        
    def acq96Pathname(self):
        """return the filename which is needed by ACQ96 to perform the spectral acquisition.
        """
        return win32.path_join( Config.workDir(), self._set_filename ) 
        
    def housekeepingPathname(self):
        """return the filename which the housekeeping system should place HK data for this acquisition into.
        """
        return win32.path_join( Config.workDir(), self._hk_filename )
        
    def createAcqFile(self):
        """create a temporary .SET file properly configuring the interferometer ACQ96 ingest
        """
        acqfn = self.acq96Pathname()
        txt = file( Config.OP_SET, 'rt' ).readlines()
        # replace SMNSCA parameter inside the template file
        opset = [ (x[0:6] == 'SMNSCA') and ('SMNSCA = %d\n' % self._scan_count) or x for x in txt ]
        file(acqfn, 'wt').write(''.join(opset))

    def makeSetBundle(self):
        """Make a bundle out of the collected files for the set.
           Mainly just write a .sig file.
        """
        os.mknod(win32.path_join( Config.workDir(), self._set_name + '.SIG'), 0666)
        
    def filenamesForSet(self):  
        """Return a tuple of filenames which belong to this set.
        These will reside in the workspace directory upon completion of the acquire.
        The sigmove module will be responsible for moving them to the working directory used by the
        processing.
        """
        sn = self._set_name
        return [ '%sM%X.SPC' % (sn,n) for n in range(0,16) ] + [ self._hk_filename, self._sig_filename ]
        


re_seq = re.compile(r'(\d+)')

class sequence_labels:
    """Sequence label file structure with read/write
    """
    def read(self, fp):
        l = fp.readline()
        self.seq = [ int(x) for x in re_seq.findall( l ) ]
        
    def write(self, fp):
        print >>fp, ' '.join( [ `x` for x in self.seq ] )
                
    def clear(self):
        self.seq = [0] * 26

    def __init__(self,fp = None):
        if fp: self.read(fp)
        


class Sequencer:
    """A representation of the instrument sequencing logic. 
    This logic provides a series of "set" objects to the Ingest sequencer, indicating what it should do
    with all the subsystems. 
    This code is responsible for maintaining a variety of ingest status files and ensuring that they're
    intact and coherent across a variety of restart and error conditions.
    """

    def __init__(self,force_newday=False):
        """Construct a sequencing configuration for the working day, or ready a pre-existing 
           set of sequence files for resumption during the day. 
        """
        self._day = Config.YYMMDD
        self._currentset = None
        self._force_newday = force_newday
        
    def configureDirectories(self):
        """Set up working directory for sequence ingest.
        
        Returns yymmdd of day to be resumed, None if newday.
        """    
        if self._day=='TEST': return None
        diskday = None
        
        # check the current working directory to see what day we were last on 
        exists = os.path.exists
        if not self._force_newday and os.path.exists(Config.DISKDAY):
            diskday = file(Config.DISKDAY,'rt').read().strip()
            diskdayDirsExist = exists(Config.calDir(diskday)) and exists(Config.scienceWorkDir(diskday))  and exists(Config.outputDir(diskday))
            # NOTE: these directories are created by Tidy in its configureDirectories. 
            if not diskdayDirsExist: 
                log.warning( 'Directories for %s do not exist. Can not resume.' % diskday )
                diskday = None
                                
        # if it matches, we're on a restart/resume
        if diskday == self._day:
            stateExists = os.path.exists(Config.MIRROR_SCD) and os.path.exists(Config.SEQ_LAB)
            if not stateExists:
                # given that directories exist but state is lost -- this is an error
                log.error("Can not resume without %s and %s!" % (Config.MIRROR_SCD,Config.SEQ_LAB))
            else:
                self._prepareToResume()
                return self._day
        
        # otherwise, we're on a new day
        log.info("NEWDAY %s" % self._day)
        log.info("writing %s to state file %s" % (self._day,Config.DISKDAY))
        file(Config.DISKDAY,'wt').write('%s\n' % self._day)
        self._prepareNewDay()
        return None


    def _prepareToResume(self):
        """private: Fix any inconsistencies in the state files. Reload the state information.
        Ready the system for a logical next sequence (i.e. starting from the top of the mirror schedule).
        """
        log.info("preparing to resume ingest sequence")
        # verify that SEQ_LAB and MIRROR_BEG exist
        # still have to start from the top of the mirror sequence
        ms = mirror_schedule( file(Config.MIRROR_SCD,'rt') )
        ms.iview = 1
        ms.write( file(Config.MIRROR_SCD,'wt') )
        sl = sequence_labels(file(Config.SEQ_LAB,'rt'))
        # start at the top of the next sequence by marking everything <max as max
        log.info( 'Sequence labels: %s' % `sl.seq` )

        last_seq = max(sl.seq)
        # if we've completed an even number of views, continue with next odd
        #     S3 -> H4
        # if we've completed an odd number of views, skip to the top
        #     S2 -> H3
        # by fiat: the first half of the mirror program are odd-numbered scenes (H1, A1, S1), 
        #          second half is even-numbered (H2, A2, S2)
        #          starting at the top of the mirror sequence means going odd
        if (last_seq % 2) == 1: 
            last_seq += 1 
        sl.seq = [ last_seq ] * len(sl.seq)
        
        sl.write( file(Config.SEQ_LAB,'wt') )
        log.info("Resuming ingest at sequence number %d" % (last_seq+1))
        
        # create set information, so that we can get the set name and write AERI.CTL
        newset = self._generateSetFromState(ms,sl)
        log.info("Starting set is %s" % newset)
        
        log.info("writing %s to end of %s" % (newset.name(),Config.AERI_CTL))
        file(Config.AERI_CTL,'ab').write( newset.name() + '\n' )
            
    def _cleanWorkDir(self):
        #FIXME - this should be implemented by Tidy
        pass        
        
    def _prepareNewDay(self):
        """private: Reinitialize all the sequence control files to their new-day condition.
        Reset the mirror sequence, reset the set counters. Flush out any pending data that may not have been
        processed or turned over.
        """
        log.info("preparing for new day")
        # get rid of buildup and leftovers 
        self._cleanWorkDir()
        # initialize mirror sequence for the day
        self._initSequenceFiles()

    def _initSequenceFiles(self):
        """initialize sequence label files to start-of-day, copy starting mirror sequence to state
        """
        # copy the base mirror sequence 
        log.info("copying mirror sequence from %s to %s" % (Config.MIRROR_BEG,Config.MIRROR_SCD))
        shutil.copyfile( Config.MIRROR_BEG, Config.MIRROR_SCD )
        # initialize seq.lab file
        sl = sequence_labels()
        sl.clear()
        sl.write( file( Config.SEQ_LAB, 'wt' ) )
        
    def estimateSpaceRequirement(self):
        """estimate the number of bytes of disk needed to make it through the end of the day, conservatively
        """
        return 1000000000L #FIXME this is best estimated based on the mirror sequence

    def _generateSetFromState(self,mirscd,seqlab):
        """For a given mirror schedule and sequencel label state, return a aeri_set_info structure
        """
        view = mirscd.views[mirscd.iview - 1]
        viewpos = ord(view.name[0]) - ord('A')        
        # seq.lab contains the number we've COMPLETED of this view (start-at-0)
        # the OS/2 AERIs seq.lab contain the next sequence number (start-at-1)
        viewNumber = seqlab.seq[viewpos] + 1
        
        return aeri_set_info(set_name = '%s%d' % (view.name[0],viewNumber), scan_count = view.scans, 
                                         mirror_angle = view.angle, mirror_command = view.command)
         
    def _readCurrentSet(self):
        """ extract information from the state files in the working directory to figure out what the 
        next set of data that should be ingested should be
        """
        # read the sequence label file
        self.seqlab = sequence_labels( file( Config.SEQ_LAB, 'rt' ) )
        
        # read the mirror schedule file
        self.mirscd = mirror_schedule( file( Config.MIRROR_SCD, 'rt' ) )

        self._currentset = self._generateSetFromState(self.mirscd,self.seqlab)
        return self._currentset
        
    def getCurrentSet(self):
        """return an object representing the current set to be acquired from the instrument.
        """
        if self._currentset: return self._currentset
        self._currentset = self._readCurrentSet()
        return self._currentset        
        
    def incrementSet(self):
        """Move the instrument state on to the next set in the series.
           return an object representing the new set concisely.
           """
        # read the mirror schedule file and the label sequence file
        # upseq = process(exe = Config.aeriRunExe('a_upseq'))
        # upseq()
        
        self.mirscd = mirror_schedule( file(Config.MIRROR_SCD, 'rt') )
        self.seqlab = sequence_labels( file(Config.SEQ_LAB, 'rt') )
        log.info( "loaded mirror schedule and sequence label files" )

        # increment the sequence label
        view = self.mirscd.views[self.mirscd.iview - 1]
        viewpos = ord(view.name[0]) - ord('A')        
        # seq.lab contains how many of these view we've completed; increment
        self.seqlab.seq[viewpos] += 1
        log.info( "incremented %c scene to %d completed views" % (view.name[0],self.seqlab.seq[viewpos]) )
        
        # and move on to the next entry in the schedule
        self.mirscd.iview += 1
        if self.mirscd.iview > self.mirscd.nviews: 
            self.mirscd.iview = 1
        log.info( 'incremented mirror sequence on to view %d = %s' % (self.mirscd.iview, self.mirscd.views[self.mirscd.iview - 1] ) )
            
        # write back the files
        #FIXME: this should write to scratch files and then rename into place
        self.mirscd.write( file(Config.MIRROR_SCD, 'wt') )
        self.seqlab.write( file(Config.SEQ_LAB, 'wt') )
        log.info('wrote %s and %s' % (Config.MIRROR_SCD, Config.SEQ_LAB) )
        
        self._currentset = None
        
        return self._readCurrentSet()
        

def _test_set_config( testdir='/tmp' ):
    """Monkeypatch Config for test use"""
    opj = os.path.join
    Config.OP_SET = opj(testdir,'op.set')
    Config.YYMMDD='TEST'
    Config.MIRROR_SCD = opj(testdir,'mirror.scd')
    Config.SEQ_LAB = opj(testdir,'seq.lab')
    def wdir(): 
        return testdir
    Config.workDir = wdir

def test_mirscd( args ):
    """read a mirror schedule file and write it back to a different filename
    """
    print "--mirror schedule read/write test--"
    ms = mirror_schedule( file(args[0],'rt') )
    print ms
    fpo = file(args[1],'wt')
    ms.write(fpo)
    
def test_seq1( args ):
    """generate a .SET file for the current sequence using config files forced into /tmp
    """
    print "--simple sequencer test 1--"
    _test_set_config()
    S = Sequencer()
    cs = S.getCurrentSet()
    print cs
    print "creating %s for %s set with %d scans at angle %d" % (cs.acq96Pathname(), cs.name(), cs.interferometerScans(), cs.mirrorAngle())
    cs.createAcqFile()
    
def test_incrseq( args ):
    """ Increment sequence label and mirror schedule file on /tmp
    """
    _test_set_config()
    S = Sequencer()
    cs = S.getCurrentSet()
    print cs
    S.incrementSet()
    cs = S.getCurrentSet()
    print cs
    
def test_config(args):
    """ Set up configuration files for a given day.
    """
    S = Sequencer()
    S.configureDirectories()
    

# Example invocation for local testing
# DISKDAY=/tmp/disk.day SEQ_LAB=/tmp/seq.lab _WORK_DIR=/tmp MIRROR_BEG=$HOME/Workspace/aeri/config/mirror.beg MIRROR_SCD=/tmp/mirror.scd python Sequencer.py config 050313

if __name__=='__main__':
    doc = """
[YYMMDD=<yymmdd>] python Sequencer.py setup  -- configure working directory for a given YYMMDD
python Sequencer.py mirscd /tmp/mirror.beg /tmp/mirror.cpy -- test mirror schedule i/o
python Sequencer.py seq1  -- create .SET file from /tmp/mirror.scd, /tmp/seq.lab, /tmp/op.set
python Sequencer.py seq2  -- increment /tmp/mirror.scd and /tmp/seq.lab
    """
    from sys import argv
    logging.basicConfig()
    log.setLevel(logging.DEBUG)
    if len(argv)<2: print doc
    
    tests = { 'mirscd': test_mirscd,
              'setup': test_config,
              'seq1': test_seq1,
              'seq2': test_incrseq
              }

    if len(argv) > 1 and argv[1] in tests.keys():
        tests[argv[1]](argv[2:])
    else:
        print doc
    
