#!/usr/bin/env python
"""
$Id: Status.py,v 1.24 2005/11/28 14:25:30 brucef Exp $
Copyright 2005, University of Wisconsin Space Science & Engineering Center

    This file is part of the AERI ingest software.

    The AERI ingest software is free software; you can redistribute it and/or 
    modify it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA



"""

from Sequencer import aeri_set_info
from process import process
import Config, win32, logging, os
from cPickle import dumps, loads
from struct import pack,unpack,calcsize
from pprint import pformat,pprint

log = logging.getLogger( "Status" )

SIZE_FMT = '>I'

class Status:
    """textual status interface, e.g. using Curses
    """

    def __init__(self):
        self._pipe = None

    def showSet(self, setname):
        log.info(setname)

    def showMirror(self, mirPos):
        log.info(mirPos)

    def finished(self, complete):
        log.info(complete)

    def startMirrorMon(self):
        """Start a local mirror monitor java process.
        """
        self._mirMon = process(exe = Config.JAVA_EXE,
                               arguments=['-cp', Config.JADA_CLASSPATH,
                                          'edu.wisc.ssec.jada.AeriMirrorMonitor',
                                          '--port', Config.MSG_PORT
                                          ],
                               output_stream = Config.logStream("jvm.out"),
                               output_prefix = "mirror_mon",
                               work_dir=Config.JADA_APP_ROOT )
        self._mirMon.start()
        return self._mirMon

    def startQcLog( self ):
        """Creates and starts the process for creatinsg QC logs.
        returns the pid of the started process.
        """
        sumFile = win32.path_join( Config.outputDir(Config.YYMMDD), '%s.SUM' % Config.YYMMDD )
        qcLog = win32.path_join( Config.outputDir(Config.YYMMDD), Config.AERIQC_LOG )
        args = [ '-DFILE=%s' % sumFile, Config.AESITTER_SCR ]
        self._qcTask = process( exe = Config.AERIQC_EXE,
                                  arguments = args,
                                  environ = { 'AERIQC_LOG': qcLog },
                                  output_stream = Config.logStream('aeriqc.out'),
                                  output_prefix = 'aeriqc' )
        self._qcTask.start()
        return self._qcTask

    def startMessaging( self ):
        """Starts ingest messaging. There are 2 processes, the parent is the feeder
        of messages and the child is the reader of the messages. The child python process
        Config.MSG_SVR is expected to read messages from the pipe and deliver them
        via a localhost port specified in Config.py.
        """
        rfd, wfd = os.pipe()   # would use a second pipe to convey output, if desired

        log.debug( "forking message pipes" )
        pid = os.fork()
        if pid!=0: # then I am the parent == the feeder of ingest messages
            # close the end we're not using
            os.fdopen( rfd, 'r' ).close()
            # run the feeder main 
            self._pipe = os.fdopen( wfd,'wb' )
        else:
            # close the end we're not using
            os.fdopen(wfd,'w').close()
            # execute process for serving messages
            os.execvp( 'python', ['python', Config.MSG_SVR, '%d' % rfd] )

    def sendMessage(self, aeriSet ):
        """Creates a dictionary of values from a Sequencer.aeri_set_info object, pickles
        it, and sends it over a pipe to a process responsible for formating and sending
        the information to message listeners.
        
        aeriSet - a Sequencer.aeri_set_info object
        """
        aeriDict = {
            "mirror": {"angle": str( aeriSet.mirrorAngle() ),
                       "position": aeriSet.mirrorPosition(),
                       "step": aeriSet.mirrorCommand() }
            }

        # pickle dictionary to a string
        S = dumps( aeriDict )
        # write string length as a header
        self._pipe.write( pack( SIZE_FMT, len(S) ) )
        # write string dictionary
        self._pipe.write( S )
        self._pipe.flush()

    def shutdownMessaging(self):
        if self._pipe:
            log.info("Sutting down Ingest XML Messaging")
            self._pipe.close()
            self._pipe = None
        else:
            log.info("Messaging not running")

    def kill( self ):
        log.info("killing QC log task")
        self._qcTask.kill()
        log.info("killing Mirror Monitor")
        self._mirMon.kill()
        self.shutdownMessaging()

def test_mirrorMon(args = None):
    s = Status()
    #test_messaging()
    mir = s.startMirrorMon()
    mir.wait()

def test_messaging(args = None):

    msgCount = 100
    if args and len(args) > 0: msgCount = int(args[0])

    import time,string
    from random import random

    stat = Status()
    stat.startMessaging()

    # just capital letters
    p = string.ascii_letters[26:]

    # generate random msgs
    for i in range( msgCount ):
        r = random()
        asi = aeri_set_info( '%s'%( p[i%26] ), 0, r * 360, 'D%s' %( int(r * 1000) ))
        stat.sendMessage( asi )
        log.info("Sent :%s" %asi)
        time.sleep(5)
    stat.shutdownMessaging()

def test_qclog(args):
    s = Status()
    qc = s.startQcLog()
    qc.wait()

if __name__=='__main__':
    doc = """
python Status.py messaging [<msg_count>] - Send specifed number of generated messages
python Status.py mirrormon - Start java process for mirror monitor
[YYMMDD=<yymmdd>] python Status.py qclog - Start aeriqc.exe process
"""
    from sys import argv
    logging.basicConfig()
    log.setLevel(logging.DEBUG)

    tests = {   'messaging': test_messaging,
                'qclog': test_qclog,
                'mirrormon': test_mirrorMon
                }

    if len(argv) > 1 and argv[1] in tests.keys():
        tests[argv[1]](argv[2:])
    else:
        print doc

