#!/usr/bin/env python
"""
$Id: Tidy.py,v 1.33.2.2 2008/09/03 03:19:27 dennyh Exp $
Copyright 2005, University of Wisconsin Space Science & Engineering Center

    This file is part of the AERI ingest software.

    The AERI ingest software is free software; you can redistribute it and/or 
    modify it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA



"""

import re, operator, logging, os, shutil, sys, time
from glob import glob
import Config, win32, sigmove
from process import process

log = logging.getLogger('Tidy')

# regular expression parsing a line of output from 'df'
RE_DF = re.compile( r'(\S+)\s+(\d+)\s+(\d+)\s+(\d+)\s+(\d+)+%\s+(\S+)' )
# regular expression parsing a line of output from 'du'
RE_DU = re.compile( r'(\d+)\s+(\S+)' )

RE_AEYYMMDD = re.compile( r'^AE\d+$' )


def bytesDiskAvailable(directory='.'):
    """ Return the approximate bytes of disk space available for the volume holding the given working directory.
    """
    # barf, let's do some screen scraping
    text = os.popen2( [Config.DF_EXE, '-k', directory], mode='r' )[1].readlines()[-1]
    tup = RE_DF.findall(text)[0]
    return int(tup[3])*1024L

def bytesConsumed(directory='.'):
    """ Return the approximate number of bytes consumed by a given directory.
    """
    # barf yet again, let's do some more screen scraping
    text = os.popen2( [Config.DU_EXE,'-sk',directory], mode='r' )[1].readlines()[-1]
    tup = RE_DU.findall(text)[0]
    return int(tup[0])*1024L
    
def yesterYYMMDD( yymmdd, days=1 ):
    """given integer YYMMDD, rewind by N days
    """
    if 'UTC' not in time.tzname:
        log.error("time zone is not UTC!")
        #os.environ['TZ']='UTC'
        #time.tzset()
        time.timezone = 0
        
    yymmdd = int(yymmdd) # in case it's a string
    yyyy = yymmdd/10000 + 2000 # FIXME: Y2K and all that
    mm = (yymmdd/100) % 100
    dd = yymmdd % 100
    ttup = ( yyyy,mm,dd, 12,00,00, 0,0,0 )
    old = time.mktime( ttup ) - days*24*3600
    ttup = time.gmtime( old )
    yy = ttup[0] % 100
    mm = ttup[1] 
    dd = ttup[2]
    return '%02d%02d%02d' % (yy,mm,dd)
    

def findPurgeGroups( managed_dirs = None, glob_pattern = 'AE*', glob_regex = RE_AEYYMMDD ):
    """ 
    return a dictionary of grouped directories we can consider for deletion
    
    By default, list of managed directories is the FTP directory and the science work directory.
    """
    # glob does not like \\ dirs from win32
    # glob all the AE directories in the work areas
    if not managed_dirs:
        #managed_dirs = [ '/cygdrive/c/ftp', '/cygdrive/c/' ]
        managed_dirs = [ win32.unix_path(x) for x in [ Config.outputDir(), Config.scienceWorkDir() ] ]
    
    #globs = reduce( operator.add, [ glob(win32.path_join(mdn, glob_pattern)) for mdn in managed_dirs ], [] )
    import os.path 
    globs = reduce( operator.add, [ glob( os.path.join(mdn, glob_pattern)) for mdn in managed_dirs ], [] )
    # @@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@

    log.debug( 'available directories for purge: %s' % globs )
    
    # sort them into a dictionary e.g. { 'AE050311': [ '\\ftp\\AE050311', '\\AE030511' ] }
    all = {}
    for path in globs:
        if not os.path.isdir(path):
            log.warning( '%s isn\'t a directory - ignoring' % path )
            continue
        stem = win32.path_split( path )[1]
        if not glob_regex.match( stem ):
            log.debug('%s doesn\'t match regex filter - ignoring' % path)
            continue
        v = all.get( stem, [] )
        v.append( path )
        all[stem] = v
    return all


def findPurgeCandidates( bytes_to_free, managed_dirs = None, glob_pattern = 'AE*', glob_regex = RE_AEYYMMDD ):
    """ Find an ordered list of tuples naming 
    directory groups to be deleted in order to
    get back a specified number of bytes of space.
    
    Returns success(boolean), list-of-directory-paths
    
    By default, list of managed directories is the FTP directory and the science work directory.
    """

    all = findPurgeGroups( managed_dirs, glob_pattern, glob_regex )
    log.debug( 'sorted available directories for purge: %s' % all )
        
    # scan candidate directories in key order until we've found enough space
    
    stems = [ k for k in all.keys() ]
    stems.sort() 
    
    bytes_would_free = 0
    kill_list = []
    
    log.debug('scanning directories to find enough space...') 
    for stem in stems:
        dirs = all[stem]
        bytes = reduce( operator.add, [ bytesConsumed(x) for x in dirs ], 0 )
        log.debug( '%d bytes total in %s' % (bytes,dirs) )
        kill_list += dirs 
        bytes_would_free += bytes
        if bytes_would_free > bytes_to_free: break
        
    if bytes_would_free < bytes_to_free:
        log.error( 'Unable to free %d bytes of space! Only %d can be reclaimed!' % (bytes_to_free, bytes_would_free) )
        return False, kill_list
        
    return True, kill_list

def purgeDirectories( kill_list, glob_regex = RE_AEYYMMDD ):
    """Given a list of directory paths, purge them and all their subdirectories. DANGEROUS!
    """
    log.debug("kill list is %s" % kill_list)
    for item in kill_list:
        stem = win32.path_split(item)[1]
        if not glob_regex.match( stem ):
            log.error("%s does not match a purge-able pattern! ignoring." % item)
        else:
            log.info( 'recursively deleting %s...' % item )
            shutil.rmtree( item )
    log.debug('purgeDirectories complete')        


class Tidy: 
    """Class which manages disk space for AERI work. 
    """
    
    def __init__(self):
        self._yymmdd = Config.YYMMDD

    def sanityCheckBaseDirectories(self,*more_directories):
        """ Confirm that the impossible has not occurred.
        If it has, fix it, log it, and get on with life.
        """
        omgwtf = [ Config.configDir(), Config.scienceWorkDir(), Config.workDir(), Config._OUTGOING_DIR ] + list(more_directories)
        
        for dirname in omgwtf: 
            if not os.path.isdir(dirname): 
                log.critical('%s does not exist! creating it.' % dirname)
                if os.path.exists(dirname):
                    log.error('%s has a file in the way! removing it.' % dirname)
                    os.unlink(dirname)
                os.makedirs( dirname, mode=0777 )
                log.warning('unexpectedly forced to create %s (success)' % dirname)
    
        
    def configureDirectories(self):
        """Create work directories for today's date.
        Includes FTP directory, science work directory.
        Set up configuration files (YYMMDD.par) in the science work directory.
        Returns boolean whether directories were created or not.
        Create \FTP\current and \current directories softlinked to \aeYYMMDD et al
        """
        workspace = Config.workDir()
        ftpbase = Config.outputDir()
        sciwork = Config.scienceWorkDir(self._yymmdd)        
        outwork = Config.outputDir(self._yymmdd)
        cal = Config.calDir(self._yymmdd)
        created = False
        for each in [ workspace, ftpbase, sciwork, outwork, cal ]: 
            if not os.path.isdir(each): 
                log.info("creating %s" % each)
                os.makedirs(each)
                created = True
            else:
                log.debug("%s already exists" % each)
        scicur = Config.scienceWorkDir('YYMMDD')
        outcur = Config.outputDir('YYMMDD')
        for lfrom,lto in [ (sciwork,scicur), (outwork,outcur) ]:
            # GOTCHA - XP does not return True os.path.exists() 
            # correctly for soft links
            if os.path.islink(lto): os.unlink(lto)
            log.info('softlinking %s to %s' % (lfrom,lto))
            os.symlink(lfrom,lto)
        return created

    def startClearSpace(self, space_needed):
        """Create a subprocess which clears out free space for the day's work.
        Returns when there's at least enough space (~1MB) for ingest to start.
        """
        bda = bytesDiskAvailable(Config.workDir())
        
        ispace = Config.INITIAL_SPACE
        if bda < ispace:
            ispace -= bda
            log.info("clearing %d bytes of ingest space" % ispace)
            rc,kill_list = findPurgeCandidates(ispace)
            if not rc:
                log.error("unable to find initial space! exiting..")
                assert(0)
            purgeDirectories(kill_list)
            bda = Config.INITIAL_SPACE
        
        # now, fork a child process which clears enough space for the day
        if bda < space_needed:
            space_needed -= bda
            log.info("clearing %d bytes for today\'s data" % space_needed)
            
            pid = os.fork()            
            if pid==0: # then I am the child
                # subprocess does this.. 
                rc,kill_list = findPurgeCandidates(space_needed)
                if not rc:
                    log.error("insufficient free space for the day!")
                    sys.exit(2)
                purgeDirectories(kill_list)
                log.info("disk space cleared")
                sys.exit(0)
                
            else: # we're the parent process
                return process(pid=pid)
        else: # nothing to do!
            return None

    def writeParFile(self):
        """ write YYMMDD.PAR into the cal directory.
        inst.id startaps.lst aerihk.cfg mandb rbridge.res numsamps.avg hres.cfg abcvals.num sumary.scr aesitter.scr aeri.pos mirror.beg op.set aeriemis.asc rwaspc.sip nlapp.sip aericalv.sip ffovcmr.si1 ffovcmr.si2 zfli.si1 zfli.si2 aerisum.sip channel.map aerigps.cfg BBCTRL.TMP radiance.scr motheat.cfg ctl_swts outgoing.lis aemirsaf.cfg disk_min.cfg
    
        """
        fp = file( win32.path_join(Config.outputDir(Config.YYMMDD),'%s.PAR' %Config.YYMMDD), 'wt' )
        for pathname in Config.PAR_FILES:
            if not os.path.exists( pathname ):
                log.warn("Could not add %s to par file, file does not exist" %pathname )
                continue
            fname = win32.path_split( pathname )[1]
            print >>fp, '@@@@@ BEGIN SECTION - %s @@@@@' % fname
            fp.write( '\n'.join( [ x.strip() for x in file(pathname,'rt').readlines() ] ) )
            print >>fp, '@@@@@ END OF SECTION - %s @@@@@' % fname
    

           
    def moveSpcFilesToScience(self):
        """Move SPC files from ingest work directory to science work directory
        """
        scidir = Config.scienceWorkDir(self._yymmdd)
        jlist = sigmove.findJobs( Config.workDir() )
        if not jlist: log.warning("no spc files to move to science work directory!")
        else: sigmove.runJobs( scidir, jlist, 'move' )
                
    def cleanFinishedSpcFiles(self, job_type = 'zip', is_fin=True ):
        """Clean SPC files out of the science work directory
        """
        scidir = Config.scienceWorkDir(self._yymmdd)
        zip_pathname = win32.path_join( scidir, 'SPC_%s_%s.zip' % (Config.instrumentName(), self._yymmdd) )
        log.info( "moving SPC files to zipfile %s" % zip_pathname)
        jlist = sigmove.findJobs( scidir, is_fin )
        #FIXME: verify that this shouldn't be a warning
        if not jlist: log.debug("no spc files to purge from science work dir!")
        else: sigmove.runJobs( zip_pathname, jlist, job_type )
        
    def cleanUnusedSpcFiles(self, job_type = 'zip' ):
        self.cleanFinishedSpcFiles(is_fin=False)   
        jlist = sigmove.findJobs( Config.workDir(), False )
        if not jlist: log.debug( "no pending SIGs in ingest workspace" )
        else: 
            log.info('removing workspace files: %s' % jlist)
            sigmove.runJobs( None, jlist, 'purge' )
        
    def readOutgoingGlobList(self):
        """Read the OUTGOING.LIS file to get glob patterns which go into outgoing directory
        """
        lines = [ x.strip() for x in file(Config.OUTGOING_LIS,'rt').readlines() ]
        patterns = [ x for x in lines if (len(x)>0) and x[0] not in [ '#', ';' ] ]
        return patterns
        
    def importToOutgoing( self, srcdir, dstdir, globlist = None, move = False, dryrun = False ):
        """create hard links to dstdir of files in srcdir which match globlist.
        """
        if not globlist: globlist = self.readOutgoingGlobList()         
        log.info('importing data from %s to %s for patterns %s' % (srcdir,dstdir,globlist) )
        for pat in globlist:
            all = glob( win32.unix_path(win32.path_join( srcdir, pat )) )
            log.info( "for %s found %s" % (pat,all) )
            for each in all: 
                try:
                    fn = win32.path_split( each )[1]
                    fn_dst = win32.path_join( dstdir, fn )
                    if os.path.isfile(fn_dst): 
                        log.debug("file %d already exists, is presumably already hard-linked" % fn_dst)
                        continue
                    if dryrun: 
                        log.info("copy %s to %s" % (each,fn_dst) )
                    else:
                        if move: 
                            shutil.move( each, fn_dst )
                        else:
                            os.link( each, fn_dst )
                            #shutil.copyfile( each, fn_dst )
                except:
                    log.error("could not copy %s to %s" % (each,fn_dst))
        
    def rebuildOutgoingDirectory( self, yymmdd = None, globlist = None, dryrun = False ):
        """ copy yesterday, and day before yesterday's data into new outgoing. rename as outgoing
            advisable to do this in a forked process
        """
        log.info("rebuilding outgoing directory")
        # days which we're transferring data from
        if not yymmdd: yymmdd = Config.YYMMDD
        yester = [ yesterYYMMDD(int(yymmdd),1), yesterYYMMDD(int(yymmdd),2) ]
        
        if not globlist: globlist = self.readOutgoingGlobList()
        
        tmpdir = win32.path_join( Config.outputDir( ), 'OUTGOING.NEW' )
        dstdir = Config._OUTGOING_DIR
        
        # check and make sure the temporary directory doesn't exist
        if os.path.exists( tmpdir ):
            log.warning("%s already exists! removing.." % tmpdir)
            try:
                shutil.rmtree( tmpdir )
            except:
                log.error("could not remove %s" % tmpdir)
                
        # make a temp directory which we'll replace the outgoing directory with
        try:
            os.mkdir( tmpdir )
        except:
            log.error( "could not create temporary directory %s" % tmpdir )
            log.warning( "working around by rerouting data straight to outgoing" )
            tmpdir = dstdir
                
        # move un-recovered data from old outgoing to new outgoing
        for day in yester:
            pat = '*%s*' % day
            log.info( 'moving %s from old outgoing to new' % pat )
            self.importToOutgoing( srcdir = dstdir, dstdir = tmpdir, globlist = [ pat ], move=True, dryrun=dryrun )
        
        # copy additional files from the provided day into the new outgoing
        self.importToOutgoing( srcdir=Config.outputDir(yymmdd), dstdir=tmpdir, globlist=globlist )

        # move the new outgoing into place
        if tmpdir!=dstdir:    
            try:
                shutil.rmtree( dstdir )   # dangerous!
            except:
                log.error( 'failed to remove %s' % dstdir )
            try:
                shutil.move( tmpdir, dstdir )
            except:
                log.error( "unable to move %s to %s!" % (tmpdir,dstdir) )
        log.info('done rebuilding outgoing directory')
        


def test_create_data_directory( pathname, data_size=100000000 ):
    """Create a fake directory with a few files of approximate size
    """
    log.debug("creating directory %s" % pathname)
    os.mkdir( pathname )
    
    def mkfile(pathname, size=data_size):
        log.debug('creating file %s of size %d' % (pathname,size))
        f = file( pathname, 'wb' )
        f.seek( size-1 )
        f.write( '!' )
        f.close()    
    [ mkfile( win32.path_join(pathname,'xxxxxx%s') % sfx ) for sfx in [ 'c1.rnc', 'c2.rnc', '.sum' ] ]

    
def test_mkdirs( args ):    
    """Create test data which fills up space.
    """
    ndirs = int(args[0])
    for base in [ Config.scienceWorkDir(), Config.outputDir() ]:
        for n in range(0,ndirs):
            test_create_data_directory( win32.path_join( base, 'AE%06d' % n ) )
    # also create an AERI directory set in the scienceWorkDir path for complete testing
    try:
        os.mkdir( win32.path_join( Config.scienceWorkDir(), 'AERI_RUN' ) )
    except:
        log.info("AERI_RUN directory exists")
    try:
        os.mkdir( win32.path_join( Config.scienceWorkDir(), 'AERI' ) )
    except:
        log.info("AERI directory exists")

def test_move_sig( args ):
    """ Create sample files with .SIG file in work directory
    Migrate them to test science work directory AE000000.
    """
    day = '000000'
    T = Tidy()
    swd = Config.scienceWorkDir(day)
    try:
        os.mkdir(swd)
        didcreate = True
    except: 
        didcreate = False
    log.info("creating test files in %s" % swd)
    sigmove.test_create_spc_files( Config.workDir() )
    log.info("testing SPC move to science work directory")
    T.moveSpcFilesToScience()
    log.info("done, cleaning up")
    if didcreate: shutil.rmtree( swd )
    
def test_move_fin( args ):
    """Create samples files in science work dir AE000000, with .SIG and .FIN
    Migrate files to output zipfile SPC_instname_000000.zip
    """
    day = '000000'
    T = Tidy()
    swd = Config.scienceWorkDir(day)    
    try:
        os.mkdir(swd)
        didcreate = True
    except: 
        didcreate = False
    log.info("creating test files in %s" % swd)
    sigmove.test_create_spc_files(swd, suffii = ['HK.SPC', '.SIG', '.FIN'])
    log.info("testing SPC archival")
    T.cleanFinishedSpcFiles()
    log.info("done, cleaning up")
    if didcreate: shutil.rmtree(swd)    

def test_sigmove(args):
    """Run through a sigmove cycle
    """
    T = Tidy()
    T.moveSpcFilesToScience()
    
def test_delspcs(args):
    """Run through a sigmove cycle
    """
    T = Tidy()
    print "cleaning FIN-marked files"
    T.cleanFinishedSpcFiles()
    print "cleaning SIG-marked files"
    T.cleanUnusedSpcFiles()

def test_need(args):    
    """List the directories that would be removed in order to free up N bytes.
    """
    from pprint import pprint
    pprint( findPurgeCandidates( int(args[0]) ))

def test_free(args):
    """Tell how much free space there is for the work directory.
    """
    print "free space for %s: %d" % (Config.workDir(),bytesDiskAvailable(Config.workDir()))

def test_outgoing(args):
    """hardlink files to outgoing directory
    """
    T = Tidy()
    T.rebuildOutgoingDirectory(dryrun=True)
    
def test_sanity(args):
    log.debug("Checking sanity...")
    T = Tidy()
    T.sanityCheckBaseDirectories()
    log.debug("Done")

def test_configDirs(args):
    T = Tidy()
    T.configureDirectories()

def test_makepar(args):
    T = Tidy()
    T.writeParFile()

def test_groups(args):
    from pprint import pprint
    pprint(findPurgeGroups())

def test_del_a_dir(args):
    log.setLevel(logging.INFO)
    pg = findPurgeGroups()
    kl = pg.keys()
    kl.sort()
    n = 1
    npg = {}
    for k in kl:
        lis = pg[k]
        ndirs = len(lis)
        print "%d: %s (%d director%s)" % (n,k,ndirs,ndirs==1 and 'y' or 'ies')
        npg[n] = k,lis
        n += 1
    choice = None
    while choice != 0:
        choice = int(raw_input( 'Which to purge (0 to exit)? ' ))
        if choice!=0:
            name,todel = npg[choice]
            print "Checking disk space.."
            bytes = reduce( operator.add, [ bytesConsumed(x) for x in todel ], 0 )
            ndirs = len(todel)
            print '- '*32
            print ', '.join([win32.windows_path(x) for x in todel])
            yn = raw_input( "Confirm Y/N: Shall I delete %s with %dMB in %d director%s? " % (name,bytes/1024**2,ndirs,ndirs==1 and 'y' or 'ies') )
            if yn in 'Yy': 
                purgeDirectories(todel)
                print "Purged."
            else:
                print "Not purged."
    print "Done."

def test_purge(args):
    #from pprint import pprint
    rc,kill_list = findPurgeCandidates( int(args[0]) )
    if rc: 
        print "looks good! purging.."
        purgeDirectories(kill_list)
    else:
        print "no purge; try a smaller size"
        
    print "testing purge avoidance - this should print two errors re AERI_RUN and AERI"
    try: 
        purgeDirectories( [ win32.path_join(Config.scienceWorkDir(),x) for x in [ 'AERI', 'AERI_RUN' ] ] + [ 'AE999999' ] )
    except OSError:
        print "correctly received exception due to invalid pathname - OK"
        


if __name__=='__main__':
    doc="""$Id: Tidy.py,v 1.33.2.2 2008/09/03 03:19:27 dennyh Exp $
python Tidy.py makedirs #directories -- make N test directories of form AE000001, 2, 3, ...
python Tidy.py available -- show available disk space for the work directory
python Tidy.py need number-of-bytes -- show which directories would get deleted to free N bytes
python Tidy.py purge number-of-bytes -- delete directories to free N bytes
python Tidy.py sanity -- ensure that base directories exist
python Tidy.py groups -- list groups of directories containing deletable data
python Tidy.py del_a_dir -- list directories available for purging, prompt to purge one
YYMMDD=[yymmdd] python Tidy.py setup -- create default directories for YYMMDD
YYMMDD=[yymmdd] python Tidy.py makepar -- write par file to FTP directory for the day
    The config files used to create the PAR file are assumed to reside in the cal dir
    as if Science.configureDirectories()
YYMMDD=[yymmdd] python Tidy.py simple_sig -- test sigmove (workdir -> scienceworkdir) function with false data
YYMMDD=[yymmdd] python Tidy.py simple_fin -- test delspcs (scienceworkdir -> zipfile) function with false data
YYMMDD=[yymmdd] python Tidy.py sigmove -- test sigmove functionality with real data
YYMMDD=[yymmdd] python Tidy.py delspcs -- test delspcs functionality with real data
YYMMDD=[yymmdd] python Tidy.py outgoing -- update outgoing directory for YYMMDD and the two preceding days
"""
    argv = sys.argv
    logging.basicConfig()
    log.setLevel(logging.DEBUG)
        
    tests = { 'makedirs': test_mkdirs,
              'need': test_need,
              'available': test_free,
              'purge': test_purge,
              "simple_sig": test_move_sig,
              "simple_fin": test_move_fin,
              'sigmove': test_sigmove,
              'delspcs': test_delspcs,
              'outgoing': test_outgoing,
              'sanity': test_sanity,
              'makepar': test_makepar,
              'groups': test_groups,
              'setup': test_configDirs,
              'del_a_dir': test_del_a_dir
              }

    if len(argv) > 1 and argv[1] in tests.keys():
        tests[argv[1]](argv[2:])
    else:
        print doc
