#!/usr/bin/env python 
"""
$Id: process.py,v 1.26 2008/05/22 22:19:21 rayg Exp $
Copyright 2005, University of Wisconsin Space Science & Engineering Center

    This file is part of the AERI ingest software.

    The AERI ingest software is free software; you can redistribute it and/or 
    modify it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA



"""

import os,sys,signal,logging #,exceptions


#FUTURE: 
# Setting up signal (SIGCHLD) so that the completion of a child process 

log = logging.getLogger('process')

#def handle_sigchld( signumber, stackframe ):
#    # print ">>>child process exited"
#    log.info("child process ended (SIGCHLD)")
#
#signal.signal( signal.SIGCHLD, handle_sigchld )
#FIXME: signals interrupt things like waitpid, which means trouble.
# handling this requires that any syscall handle "interrupted" exceptions (OSError, IOError)
# which isn't currently worth the gains it would provide us in knowing which processes are alive.

class process:
    """A simple process wrapper. 
    example showing all parameters: 
    p = process( exe='/bin/ls', 
                 input_lines=[ 'line1', 'line2' ],   # alternately input_stream = file-like object obeying .fileno()
                 arguments=[ '-l', '-a' ],
                 environ={"IT": '7'},     # additional environment
                 output_stream=file("dump.txt",'w'),
                 work_dir = '/tmp' )
                 # optional output_prefix='program-name' in presence of output_stream directs stdout/err through awk
    p.start()
    p.wait()
    """

    def __init__(self,exe=None,**kwargs):
        self.pid = None
        self.arguments = []
        self.environ = {}
        
        # die misserably if exe does not exist
        if not os.path.exists(exe):
            raise EnvironmentError("%s does not exist" %exe)
            
        self.exe = exe
        self.__dict__.update(kwargs)
     
    def __repr__(self):
        if self.pid: txt = "run as pid %d" % self.pid
        else: txt = ''
        return "[ executable '%s' %s ]" % (self.exe, txt)
     
    def __del__(self):
        if self.pid:
            log.warning("killing process %d due to process object destructor" % self.pid)
            try: 
                os.kill(self.pid,signal.SIGKILL)
                log.warning("killed process %d during process.__del__" % self.pid)
            except:
                pass
            try:
                log.debug( "try wait pid %d" % self.pid )
                result = os.waitpid(self.pid,0)
                log.debug( "succeed wait pid %d" % self.pid)
                log.warning("result of kill: %s" % `result`)
            except: 
                pass
        
    #def is_running(self):
    #    if not self.pid: return False
    #    try: #FIXME: find a more elegant way to do this test
    #        os.waitpid(self.pid, os.NOWAIT)
    #    except OSError:
    #        return False
    #    return True

    def start( self, *args, **kwargs ):
        """ Start the executable as a separate process.
            Optionally, a list of command-line arguments (e.g. arguments = [ '-fr', '/' ]) 
            and a list of input lines ( input_lines = [ 'my', 'dog', 'has', 'fleas' ] ) can be given
            returns process id
        """
            
        if self.pid: return -1 # is already running; FIXME have this throw something
        if args: self.arguments = list(args)
        self.__dict__.update( kwargs )  # this is small laziness + large flexibility (possibly too much)
        
        # set up input file descriptor from a buffer or stdin
        if hasattr(self,'input_stream'): 
            input_stream = self.input_stream
        elif hasattr(self, 'input_lines'):
            input_stream = os.tmpfile()   # disk buffer not attached to a filename
            print >>input_stream, '\n'.join(self.input_lines)
            input_stream.seek(0)
        else:
            input_stream = sys.stdin
        
        self.exedirname, self.exefilename = os.path.split( self.exe )
        cmd = [ self.exe ] + self.arguments
        log.debug('about to start %s' % str(cmd))

        self.pid = os.fork()
	
        if self.pid: # then I am the parent
            # write the inputs to the pipe
            log.debug( "fork %s as %d in dir %s" % (self.exe, self.pid, getattr(self,'work_dir',os.getcwd())) )
            return self.pid
            
        # we are the child process, let's run away
        else:
            self.pid = os.getpid()
            if hasattr(self,'work_dir'):
                os.chdir(self.work_dir)
            #FIXME: the fact that os.environ does not appear to work properly with .update may be a Python bug
            if self.environ: 
                for k,v in self.environ.items():
                    os.environ[k] = v
            os.dup2( input_stream.fileno(), 0 ) # change stdin
            
            if hasattr(self,'output_stream') and hasattr(self,'output_prefix'):
                arfd,awfd = os.pipe()
                awkpid = os.fork()
                if not awkpid: # then I am the child; I am become awk, reformatter of stdout
                    os.dup2( arfd, 0 ) # duplicate readable end of pipe to awk's stdin
                    os.fdopen(awfd,'w').close()
                    os.dup2( self.output_stream.fileno(), 1 )  # set stdout to go to original output_stream
                    os.execvp( 'awk', ['awk', '{ print strftime("[' + self.output_prefix + ' %Y%m%d %H:%M:%S]",systime()), $_; }'] )
                else: # I am to become the actual executable, but I talk to awk
                    #log.debug("awk fork pid %d" % awkpid)
                    os.fdopen(arfd,'r').close()
                    os.dup2( awfd, 1 ) # stdout of parent -> awk 
                    os.dup2( awfd, 2 ) # stderr -> awk
                    
            elif hasattr(self,'output_stream'): 
                os.dup2( self.output_stream.fileno(), 1 ) # change stdout

            os.execv( cmd[0], cmd ) # and finally, start the intended child program
            return 0 # never runs
            
    def check(self):
        """Inquire whether a process is still running.
        Returns True,None if process is running
        Returns False,return-code if process has terminated
            Subsequent queries will have None as return code
        Returns False,None if process was not started
        May return incorrectly if process has died and its PID re-used (unlikely).
        """
        if not self.pid: return False,None
        log.debug( 'signaling pid %d to check liveness' % self.pid )
        status = None
        try:
            # reap the process if it's died
            pid,status = os.waitpid( self.pid, os.P_NOWAIT )
            # this will raise OSError if the process failed
            os.waitpid( self.pid, os.P_NOWAIT )
            return True, None
        except OSError:
            log.debug( "pid %d DOA" % self.pid )
            self.pid = None
            return False, status

    def wait(self):
        """wait for a start()ed executable to exit
           returns result-code
        """
        if not self.pid: return -1
        log.debug( "try wait pid %d" % self.pid )
        rc = os.waitpid( self.pid, 0 )
        log.debug( "succeed wait pid %d" % self.pid )
        self.pid = None
        return rc

    def kill(self,signal_type=signal.SIGTERM):
        """kill process, wait for it to complete
        """
        if not self.pid: return -1
        os.kill(self.pid,signal_type)
        return self.wait()
        
        
    def __call__(self,*args,**kwargs):
        if self.pid: return -1 # FIXME throw an exception
        if args: self.arguments = list(args)
        self.__dict__.update(kwargs)
        self.start()
        return self.wait()
        
    def getpid(self):
        return self.pid
        
        
        
def test_output_prefix( args ):
    import time
    counter = 1
    while counter <= int(args[0]):
        log.info( "Process %s", counter )
        p = process( exe = '/bin/ls', 
                     output_prefix = 'prefix-test',
                     output_stream = file('prefix-test.out','at') )
        p()
        
        time.sleep( int(args[1]) )
        counter += 1
        
def test_process( args ):
    f = process( exe = '/bin/ls', 
                 arguments = [ '-l', '-a' ], 
                 output_stream = sys.stdout )
    f.start()
    log.info( "pid is %d" % f.getpid() )
    f.wait()
    g = process( exe='/usr/bin/env', 
                 environ={ "FOO": "BAR", "BAR": "BAZ" },
                 output_stream = sys.stdout )
    g.start()
    g.wait()
    f('-C',work_dir='/tmp')
    cat = process( exe = '/bin/cat', input_lines=[ 'my', 'dog', 'has', 'fleas' ], 
                  output_stream = sys.stdout )
    cat()
    ls = process( exe = '/bin/ls', 
                 arguments = [ '-l' ], 
                 output_stream = sys.stdout, 
                 output_prefix='ls-prefix-test')
    ls.start()
    import time
    time.sleep(2)
    log.info("checking process non-aliveness..")
    alive,rc = ls.check()
    assert( not alive )
    ess = process('/bin/sleep')
    ess.start('45')
    time.sleep(1)
    log.info( "checking process liveness" )
    alive,rc = ess.check()
    assert( alive )
    log.info( "Killing sleep process %s", ess.kill() )
    log.info( 'done' )
    tee = process('/bin/sleep')
    tee.start('45')
    time.sleep(1)
    # let the process die by way of destructor
    print "<<< forcing destructor kill, 3 warnings should occur"
    del tee
    print "force kill done>>>"
    print "basic test done."
        


if __name__=='__main__':
    doc = """
python process.py output_prefix count wait(s) 
    - count is number of process to run
    - wait in secs is between processes
"""
    import sys
    
    # set up logging 
    hdlr = logging.StreamHandler()
    formatter = logging.Formatter('[%(asctime)s %(levelname)s] %(message)s')
    hdlr.setFormatter(formatter)
    log.addHandler(hdlr) 
    log.setLevel(logging.DEBUG)
    
    tests = { 
            'output_prefix': test_output_prefix,
            'basic': test_process, 
            '-t': test_process,
            }
    
    if len(sys.argv) > 1 and sys.argv[1] in tests.keys():
        tests[sys.argv[1]](sys.argv[2:])
    else:
        print doc
        