# scr2xml.py
# Version $Id: scr2xml.py,v 1.18 2005/07/22 19:58:42 brucef Exp $
# Converts AERI SCR configuration scripts to XML based on the aeriConfig.dtd located 
# at https://cvs.ssec.wisc.edu/cgi-bin/cvsweb.cgi/AERI/dev/htmlmads/util/aeriConfig.dtd.

# NOTE: All tags from the SCR file are processes, but in an attempt to clean the XML
# all processing related to unused tags is commented out with keyword LEGACY

import treeParse # https://cvs.ssec.wisc.edu/cgi-bin/cvsweb.cgi/AERI/dev/htmlmads/treeParse.py
import string
import os.path

SPACE = "    "
NL = "\n"
PROLOG = '<?xml version="1.0"?>'
DOCTYPE = r"""<!DOCTYPE aeri [
<!ELEMENT aeri (radiance | summary | metadata )*>
    <!ATTLIST aeri name CDATA #REQUIRED >
    <!ATTLIST aeri local (true|false) "false" >

<!ENTITY % configObject "( x | y | row | column | low | high | lowX | highX | lowY | highY | width | height | property | graph | cell | string | on | off )" >

<!-- Containers -->
<!--- Configuration for the radiance AeriGraphs and their windows -->
<!ELEMENT radiance (graph)*>

<!--- Configuration for the summary AeriGraphs and their windows -->
<!ELEMENT summary (graph)*>

<!-- Configuration for graphs -->
<!ELEMENT graph (plot | property)*>

<!--- Configuration for the SitterTable -->
<!ELEMENT metadata (property | cell)*>

<!--- Configuration for a SitterCell in the SitterTable -->
<!ELEMENT cell ( property )*>
    <!ATTLIST cell name CDATA #REQUIRED >
    
<!--- Configuration for each plot in an AeriGraph -->
<!ELEMENT plot ( property )*>

<!--- Configuration for a window(JFrame) containing a JADA GUI element -->
<!-- <!ELEMENT window ( property | graph )*> -->

<!--- Types allowed for AERI config elements -->
<!ELEMENT property ( row | column | x | y | low | high | lowX | highX | lowY | highY | width | height |string | on | off )*>
    <!ATTLIST property name CDATA #REQUIRED >

<!--- Data types -->
<!ELEMENT string (#PCDATA)>
<!ELEMENT on EMPTY>           <!-- boolean true -->
<!ELEMENT off EMPTY>          <!-- boolean false -->
<!-- Integer values indicating a position -->
<!ELEMENT row (#PCDATA)>      <!-- int -->
<!ELEMENT column (#PCDATA)>   <!-- int -->
<!ELEMENT low (#PCDATA)>      <!-- double -->
<!ELEMENT high (#PCDATA)>     <!-- double -->
<!-- double data range values -->
<!ELEMENT lowX (#PCDATA)>     <!-- double -->
<!ELEMENT highX (#PCDATA)>    <!-- double -->
<!ELEMENT lowY (#PCDATA)>     <!-- double -->
<!ELEMENT highY (#PCDATA)>    <!-- double -->
<!-- These values should be in decimal percentages of screen size. ie: .5 for 50%-->
<!ELEMENT x (#PCDATA)>        <!-- double -->
<!ELEMENT y (#PCDATA)>        <!-- double -->
<!ELEMENT width (#PCDATA)>    <!-- int -->
<!ELEMENT height (#PCDATA)>   <!-- int -->
]>
"""

METADATA_HEADER = """
<!-- 
The metadata tag contains the configuration for the Sitter and its cells.
The radiance tag contains the configuration for the radiance graphs.
The summary tag contains the configuration of the summary graphs.

The plot tag has an identifier which is of the format [domain id]:[range id]
-->\n
"""
POSITION_HEADER = """
<!--
A Note about the 2 different position types:
    x and y tags: Positions which have a double x and y value in decimal % of screen size.
    row and column tags: The position associated with a cell would 
        have an integer row and column designating the position in the SitterTable.
-->\n
"""

# NOTE: the lists are ordered as they are found in 
#   the SCR files. If you change the order you'll
#   make uncle Brucie very sad.
keys = {'position':['x','y'],
        'table':['column','row'],
        'size':['width', 'height'],
        'yellow':['low','high'],
        'green':['low','high'],
        'range':['lowX','highX','lowY','highY']}

# Names to ignore because they're currently not being used        
LEGACY_NAMES = ['format','help file','title','help id','real-time','bottom-label',
                'common-x','common-y','decimal-hours-x','visible-menu','horizontal-divisions',
                'vertical-divisions']

SCREEN_WIDTH = 1024.0
SCREEN_HEIGHT = 768.0

cachedPos = [0,0]
cachedSize = [-1,-1]

currentFile = [""]

def _indent( i ):
    retStr = ""
    for i in range(0,i):
        retStr += SPACE
    return retStr

def _printValues(name, element, index, cells=False):
    """
    Prints all primative values and their tags based on what datatype they are
    determined to be. 
    """
    
    parts = string.split(element)
    retStr = "" # final return string
    strings = [] # list of primative strings

    # assumes properly formated numbers when split
    if name in keys:
        # make a list of all of the primitive strings
        if cells and name == 'position':
            for key,part in zip(keys['table'],parts):
                strings.append('%s<%s>%s</%s>%s' % (_indent(index+2),key,int(part),key,NL))
                cells = False
        else:
            for key,part in zip(keys[name],parts):
                f = float(part)
                if key =='width':
                    cachedSize[0] = f
                elif key == 'x':
                    cachedPos[0] = f
                # It is assumed that the position comes before the size in the SCR files.
                # So when the size tag is read go ahead and print the cached values for 
                # both the size and position
                elif key == 'height':
                    cachedSize[1] = f
                    cachedPos[1] = SCREEN_HEIGHT - ( cachedPos[1] + cachedSize[1] )

                    if cachedPos[1] < 0: cachedPos[1] = 0

                    strings.append('%s<property name="size"><width>%6.3f</width>' \
                                    %(_indent(index+2),cachedSize[0]/SCREEN_WIDTH) + \
                                   '<height>%6.3f</height></property>%s' \
                                    %(cachedSize[1]/SCREEN_HEIGHT,NL))
                    strings.append('%s<property name="position"><x>%6.3f</x><y>%6.3f</y></property>%s' \
                        %(_indent(index+2),cachedPos[0]/SCREEN_WIDTH,cachedPos[1]/SCREEN_HEIGHT,NL))

                elif key == 'y':
                    cachedPos[1] = f
                    # This prints the position for the sitter window that  !=  contained in a sub-tree
                    # in the SCR file
                    if cachedSize[0] == -1 and cachedSize[1] == -1:
                        strings.append('%s<x>%6.3f</x><y>%6.3f</y>%s' \
                            %(_indent(index+2),cachedPos[0]/SCREEN_WIDTH,cachedPos[1]/SCREEN_HEIGHT,NL))
                        # Clear flags
                        cachedSize[0] = 0
                        cachedSize[1] = 0
                else:
                    strings.append('%s<%s>%6.3f</%s>%s' % (_indent(index+2),key,f,key,NL))
                       
        for s in strings:
            retStr += s
        
        return retStr
    
    # return tags for boolean on off tags
    if element == 'on':
        return '' # LEGACY '%s<on/>%s' % (_indent(index+2),NL)
    elif element == "off":
        return '' # LEGACY '%s<off/>%s' % (_indent(index+2),NL)
    
    # nothing else is true so it must be a string
    retStr = element
    return '%s<string>%s</string>%s' % (_indent(index+2),retStr,NL)

def _tryFloat( str ):
    """
    Helper that determines if a string is a float or not. A number is 
    considered a float if it can be cast to a float without throwing
    a ValueError.
    """
    try:
        float( str )
    except ValueError:
        return False
    return True
    
def _tryInt( str ):
    """
    Helper that determines if a string is an int or not. A string is
    considered an int if it does not contain a '.' and can be cast to 
    an int.
    """
    if string.find( str, '.' ) > 0:
        return False
    try:
        int( str )
    except ValueError:
        return False
    return True

def writeXml(name='aeri', out_dir=None, local='false', scrFiles=['AESITTER.SCR','RADIANCE.SCR','SUMARY.SCR']):
    """Meat and potatoes of the parser. It assumes that the SCR files specifed are present
    in the current directory or the one specifed.
    """
    out_name = name
    if out_dir:
      out_name = os.path.join( out_dir, name )

    out = open(out_name+'.xml', 'w')
    # begin the document
    out.write(PROLOG + NL)
    out.write(DOCTYPE + NL)
    out.write(POSITION_HEADER)
    if local == 'false':
        out.write('<aeri name="%s">%s' % (name, NL))
    else:
        out.write('<aeri name="%s" local="%s">%s' % (name, local, NL)) 

    for f in scrFiles:
        currentFile[0] = f
        _writeConfig(f, out)

    out.write('</aeri>%s' % NL )
    out.close()
    
def _writeConfig(file,out):
    """
    Write the config to the correct tag based on the static filenames
    curently used in the mads system. The AESITTER, RADIANCE, and 
    SUMMARY files are the only necessary ones as far as I can tell
    so everything else is thrown into generic config tags with name
    attributes containing the filename.
    """
    index = 1
    # get tree from tree parser
    tree = treeParse.getTree(file)
    if string.find(file,'AESITTER') >= 0:
        out.write(METADATA_HEADER)
        out.write(SPACE+'<metadata>%s' % NL)
        _parseMetadataTree(tree, out, index)
        out.write(SPACE+'</metadata>%s' % NL)
    elif string.find(file,'SUMARY') >= 0:
        out.write(SPACE+'<summary>%s' % NL)
        _parseWindowTree(tree, out, index)
        out.write(SPACE+'</summary>%s' % NL)
    elif string.find(file,'RADIANCE') >= 0:
        out.write(SPACE+'<radiance>%s' % NL)
        _parseWindowTree(tree, out, index)
        out.write(SPACE+'</radiance>%s' % NL)
    else:
        print "ERROR"

def _parseMetadataTree(subTree, out, index, cells=False):
    """
    Recursive procedure to create XML tree.
    """
    #                             SUB-SUBTREE 
    #                               v-----v
    # [ name, value, [ name, value, [ ... ] ] ]
    # ^              ^------ SUBTREE -------^ ^
    # |_______________ TREE __________________|
    #
    for name, value, links in subTree:
        # if there is a subtree
        if len( links ) > 0:
            # if the element has and associated value or longname
            if value != '':
                cells = True
                out.write( '%s<cell name="%s">%s' % (_indent(index),name,NL))
                out.write( '%s<property name="identifier">%s' %(_indent(index+1),NL))
                out.write( '%s<string>%s</string>%s' %(_indent(index+2),value,NL))
                out.write( '%s</property>%s'%(_indent(index+1),NL))
                
            # increase indent and recurse
            index += 1
            _parseMetadataTree( links, out, index, cells )
            # upon returning decrease indent
            index -= 1
            if value  !=  '':
                out.write( '%s</cell>%s' % (_indent(index),NL) )
            
        # there is no subtree so it must be a key element
        # so print primative values
        elif not LEGACY_NAMES.__contains__( name ):

            out.write('%s<property name="%s">%s'%(_indent(index+1),name,NL))
            out.write( _printValues(name, value, index, cells) )
            out.write('%s</property>%s'%(_indent(index+1),NL))

            
def _parseWindowTree(subTree, out, index):
    """
    Recursive procedure to create XML tree.
    """
    # the string of spaces for indenting
    indent = SPACE
    # add space for every index level
    for i in range( 0, index ):
        indent += SPACE
    
    for name, value, links in subTree:
        # if there is a subtree
        if len( links ) > 0:
            # if the element has subproperties it must be a
            # graph with an identifier
            if value  !=  '':
                out.write( '%s<plot>%s' % (indent,NL))
                out.write( '%s<property name="identifier">%s' %(indent+SPACE,NL))
                if value[:5] == 'FILE1':
                  out.write( '%s<string>waveNumber:radianceChannel1</string>%s' %(indent+SPACE+SPACE,NL))
                elif value[:5] == 'FILE2':
                  out.write( '%s<string>waveNumber:radianceChannel2</string>%s' %(indent+SPACE+SPACE,NL))
                elif value[:9] == 'FILE(meta':
                  out.write( '%s<string>%s</string>%s' %(indent+SPACE+SPACE,value[10:-1],NL))
                else:
                  out.write( '%s<string>%s</string>%s' %(indent+SPACE+SPACE,value,NL))
                out.write( '%s</property>%s'%(indent+SPACE,NL))
            # there is no value, must be a window
            else:
                out.write( '%s<graph>%s' % (indent,NL))
                
            # increase indent and recurse
            index += 1
            _parseWindowTree( links, out, index )
            # upon returning decrease indent
            index -= 1
            # end of plot 
            if value  !=  '':
                out.write( '%s</plot>%s' % (indent,NL) )
            # end of window
            else:
                out.write( '%s</graph>%s' % (indent,NL) )
            
        # there is no subtree so it must be a key element
        # so print primative values
        elif not LEGACY_NAMES.__contains__( name ):
            
            if name == 'position' or name == 'size':
                out.write( _printValues(name, value, index) )
            else:
                out.write('%s<property name="%s">%s'%(_indent(index),name,NL))
                out.write( _printValues(name, value, index))
                out.write('%s</property>%s'%(_indent(index),NL))
            
if __name__ == "__main__":
    doc="""python scr2xml.py <name>
    """
    from sys import argv
    try:
        writeXml(argv[1])
    except:
        print doc
