#!/usr/bin/env python
"""
$Id: sigmove.py,v 1.12 2005/08/17 19:06:05 rayg Exp $

Copyright 2005, University of Wisconsin Space Science & Engineering Center



  process to copy groups of files on signal receipt

    seeks out *.sig in directory indicated by SIGSRC environment variable
    copies files <sigfilerootname>*.* from that directory to directory indicated
          by SIGDEST environment variable
    if files cannot be copied reliably, waits until next cycle and tries again
    else finishes move and destroys source files and signal file

  The revised algorithm:
    fetch list of all *.SIG files in source directory and order them
    for each sig file, read in files matching the root, i.e. H10.sig -> H10*
    Sort the set of filenames by sequence #, sequence character, suffix
      (Vnnn.* always comes after VnnnQQQQ.*)
    Move the ordered files; if a problem occurs, skip to the next sequence
      group.


   Note that it's been set up so that we can use multi-character view names,
     e.g. SC10.SIG. However, all "IFG/SPC/etc" files are required to have
     an alphabetical after their sequence number and before their suffix
     (i.e. 1 or more letters + 1 or more digits + 1 or more letters + suffix)

   It records any problems into SIGMOVE.LOG, which is rewritten every startup
    (including program start time). The log file is placed in the directory the
    program was executed from, and all entries are timestamped.


  based on SIGMOVE, Raymond K. Garcia, 8/18/1992


Author: R.K.Garcia
"""

import glob
import os
import re
import time
import logging
from exceptions import *
from zipfile import ZipFile

IDLE_SECONDS = 30
    
log = logging.getLogger( "sigmove" )

# used for sorting signal files in numerical order
redig = re.compile( '\d+' )


def _filename( fn ): 
    "extract the filename from a path"
    return os.path.split( fn )[1]


def _numbers( pn ):
    "return a tuple of the integers in a pathname for sorting and matching purposes"
    return  tuple( [ int(v) for v in redig.findall( pn ) ] )
    

def _startsWith( a,b ):
    "return whether tuple a starts with tuple b or not"
    # print a,b
    return b==a[0:len(b)]

FILENAME_XFORMS = [ (re.compile(r'(\w\d+)M([a-f])\.[sS][pP][cC]'), lambda seq,hex: '%sM%s.SPC' % (seq,hex.upper()) ) ]

def _targetPath( dst_dir, src_pathname ):
    """ transform a source pathname to a destination pathname
        dst_dir can be None if only a filename transform is needed
    """
    src_filename = _filename(src_pathname)
    dst_filename = src_filename
    for inpat,outfn in FILENAME_XFORMS:
        m = inpat.match( src_filename )
        if m:
            dst_filename = outfn( *m.groups() )
            log.debug('transformed %s -> %s' % (src_filename,dst_filename))
    if dst_dir!=None: return os.path.join( dst_dir, dst_filename )
    return dst_filename


def findJobs( dirname, is_fin = False ):
    """return list of pairs ( signal-file, [ list-of-files ] ) telling what jobs are available in a directory
    optional parameter boolean is_fin indicates whether to look for *.SIG or *.FIN
    """
    pattern = is_fin and '*.[fF][iI][nN]' or '*.[sS][iI][gG]'
    suffix = is_fin and '.fin' or '.sig'
    sigs = glob.glob( os.path.join( dirname, pattern ) )
    numerized = [ ( _numbers(_filename(s)), s ) for s in sigs ]
    numerized.sort()
    ret = []
    for num,sig in numerized:
        stem = os.path.splitext(sig)[0]
        # grab all the files matching the sig*; throw out .sig file itself, and files which 
        # have different numbers so that H1.SIG doesn't match H10M0.SPC
        ret.append( ( sig, [ x for x in glob.glob( stem + '*' ) if x[-4:].lower() != suffix and _startsWith( _numbers( _filename(x) ), num ) ] ) ) 
    return ret

        
def moveJob( target, ( sig, filelist ) ):
    """ target: directory to move files to 
        sig: full pathname of a sig file
        filelist: files belonging to that sigfile
        
        Link all files in file list into the target directory
        unlink file from the source directory/s
        link sig file to target directory
        unlink source sig file
    """
    # copy phase
    #nfo = '%s <- %s:%s' % ( target, _filename( sig ), [ _filename(x) for x in filelist ] )
    #log.info( nfo )
    
    # FIXME: handle the case where the link already exists
    for afile in filelist:
        targfile = _targetPath( target, afile ) 
        try:
            #log.debug("link %s -> %s" % (afile,targfile))
            os.link( afile, targfile )
        except OSError, EX:
            log.exception( 'unable to link %s to %s: %s' % ( afile, targfile, EX ) )
            return None
    # sigfile move
    targsig = _targetPath( target, sig )
    try:
        #log.debug("link %s -> %s" % (sig,targsig))
        os.link( sig, targsig )
    except OSError, EX:
        log.exception( 'unable to link %s to %s: %s' % ( sig, targsig, EX ) )
        return None


    # delete phase
    for afile in filelist:
        try: 
            #log.debug("unlink %s" % afile)
            os.unlink( afile )
        except OSError, EX:
            log.exception( 'unable to unlink %s: %s', ( afile, EX ) )
            return None
    try:
        #log.debug("unlink %s" % sig)
        os.unlink( sig )
    except OSError, EX:
        log.exception( 'unable to unlink %s: %s' % ( sig, EX ) )
        return None
        
    return targsig    

def zipJob( target, (sig, filelist) ):
    """ Move the files in a job into a zipfile 
        target: ZipFile instance
        sig: full pathname of a sig file
        filelist: files belonging to that sigfile
    """
    # copy phase
    for afile in filelist:
        targfile = _targetPath( None, afile )
        try:
            target.write( afile, targfile )
        except:
            log.exception( 'unable to zip %s to %s' % ( afile, targfile ) )
            return None
    # sigfile move
    targsig = _targetPath( None, sig ) 
    try:
        target.write(sig,targsig)
    except:
        log.exception( 'unable to zip signal %s to %s' % ( sig, targsig ) )
        return None

    # delete phase
    for afile in filelist:
        try: 
            os.unlink( afile )
        except OSError, EX:
            log.exception( 'unable to unlink %s: %s', ( afile, EX ) )
            return None
    try:
        os.unlink( sig )
    except OSError, EX:
        log.exception( 'unable to unlink %s: %s' % ( sig, EX ) )
        return None
        
    return targsig    

def purgeJob( (sig, filelist) ):
    """ Purge files
        sig: full pathname of a sig file
        filelist: files belonging to that sigfile
    """
    # delete phase
    for afile in filelist:
        try: 
            os.unlink( afile )
        except OSError, EX:
            log.exception( 'unable to unlink %s: %s', ( afile, EX ) )
            return None
    try:
        os.unlink( sig )
    except OSError, EX:
        log.exception( 'unable to unlink %s: %s' % ( sig, EX ) )
        return None
        
    return None

    

def runJobs( targdir, jobs, job_type="move" ):
    """Run a sequence of jobs returned by findJobs.
    Default job type is 'move', moving files to target directory
    job_type='zip' implies move files into zipfile named as target
    job_type='purge' implies delete files
    """
    log.info( ','.join( [ '%s(%d)' % (_filename(a),len(b)) for a,b in jobs ] ) )
    if job_type=='zip':
        log.info("opening zipfile %s" % targdir)
        if not os.path.exists(targdir): 
            log.info("creating zipfile %s" % targdir)
            target = ZipFile(file(targdir,'wb+'), 'w')
        else:
            log.info("re-opening zipfile %s" % targdir)
            target = ZipFile(file(targdir,'rb+'), 'a')
        for x in jobs: zipJob(target, x)
        target.close()
        log.debug("done zipping")
        
    elif job_type=='purge':
        log.info("purging files %s" % jobs)
        return [purgeJob(x) for x in jobs]
    else:    
        return [moveJob(targdir,x) for x in jobs]
        

def test_create_spc_files( srcdir, suffii = [ '.SIG' ], scenes = [ 'H', 'S', 'E', 'A' ], majors = None, minors = None ):
    """ create a set of filler files of the right shape
    """
    all = []
    if not majors: majors = range(1, 25)
    if not minors: minors = range(0, 16)
    for scene in scenes:
        for majo in majors:
            for mino in minors:
                name = '%c%dM%x.SPC' % ( scene, majo, mino )
                file( os.path.join( srcdir, name ), 'w' ).write( name )
                all.append( name )
            for suffix in suffii: 
                name = '%c%d%s' % ( scene, majo, suffix )
                file( os.path.join( srcdir, name ), 'w' ).write( name )
                all.append( name )
    return all
    


def test1( safe=True, job_type='move' ):
    "self-test creating sample work directories" 
    srcdir = "_testsrc" # os.tmpnam()
    targdir ="_testtgt" # os.tmpnam()
    zipping = (job_type=='zip')
    if zipping: targdir = '_testtgt.zip'
    if not os.path.exists(srcdir): os.mkdir( srcdir )
    if not zipping and not os.path.exists(targdir): os.mkdir( targdir )
    print "%s <- %s" % ( targdir, srcdir )

    test_create_spc_files(srcdir)
        
    jobs = findJobs( srcdir )
    from pprint import pprint
    pprint( jobs )
    
    if safe: # almost anything goes in a test routine
        os.link = lambda x,y: pprint("link %s -> %s" % (x,y))
        os.unlink = lambda x: pprint( "unlink %s" % x )
    
    runJobs( targdir, jobs, job_type )


#FIXME: stress test this with zip files larger than 2GB

def main( args ):
    nfo = "$Id: sigmove.py,v 1.12 2005/08/17 19:06:05 rayg Exp $"
    print nfo
    log.setLevel( logging.INFO )
    log.info( nfo )
    
    sigsrc = None
    sigdst = None
    if len(args)>2: sigdst = args[2]
    if len(args)>1: sigsrc = args[1]
    try:
        if not sigsrc: sigsrc = os.environ[ "SIGSRC" ]
        if not sigdst: sigdst = os.environ[ "SIGDEST" ]
    except: 
        log.error( "source and destination (%s,%s) need to be provided from the command line or the environment variables SIGSRC and SIGDEST" % ( sigsrc, sigdst ) )
        return 9
    
    delay = None
    while True:     
        if delay: time.sleep( delay )
        delay = IDLE_SECONDS
        
        try:
            jobs = findJobs( sigsrc )    
            print '%s: %s' % ( time.asctime(), [ _filename(x[0]) for x in jobs ] )
        except OSError, EX:
            log.error( "unable to find jobs in %s: %s" % ( sigsrc, EX ) )
            continue
        if len(jobs)==0: continue
        
        results = runJobs( sigdst, jobs )
        gotcha = None in results
        if gotcha: log.warning( 'suspicious job results:\njobs =\n%sresults =\n%s\n' % ( jobs, results ) )
        
    return gotcha and 5 or 0 
    
if __name__=='__main__':
    import sys
    
    bucket = logging.FileHandler( 'sigmove.log', mode='at' )
    logformat = '\n+++++ %(levelname)s %(asctime)s [%(relativeCreated)s]\n%(message)s'
    formatter = logging.Formatter( logformat, '%y%m%d %H:%M:%S' )
    bucket.setFormatter( formatter )
    bucket.setLevel( logging.DEBUG )
    logging.getLogger().addHandler( bucket )
    
    sys.exit( main( sys.argv ) )
    
