###############################################################################
#
#   $Id: treeParse.py,v 1.3 2005/09/08 17:11:27 brucef Exp $
#	treeParse.py
#
#	Started On: 6/5/01
#
#	Created By: Greg Quinn
#
#	This library contains functions to ingest AESITTER.SCR file info
#	into a tree of nested tuples. The data is not interpreted in any
#	way so this library is useful by any files of SCR format.
#
#	The format in which the data is stored is a tree-like structure.
#	At the top-level is a list, which contains tuples. Each tuple has
#	three items: a name, a value, and a list containing all tuples
#	with are nested within the node.
#
#	ex. [(a, b, []), (c, d, [(e, f, []), (g, h, [])])
#
#	equals:		a	c
#			       / \
#			      e	  g
#
###############################################################################

from string import split, join

delims = ['is', 'of', 'with']

# this is the function called externally to parse an SCR file into the
# appropriate tree structure

def getTree(filename):

    # make a global object storing all the tokens from the SCR file
    global treeList
    treeList = FileTokenizer(filename)

    tree = []			# tree starts as empty list

    # recursively store all information from file
    while treeList.hasMoreTokens() != 0:
        name = treeList.getNextToken()
        while 1:
            token = treeList.getNextToken()
            if token in delims:		# when 'is' encountered, get links
                break			# and value
            name = name + ' ' + token	
        tree.append(getTuple(name))	
    return tree

# this is a recursive function that returns the name, value, and list of
# links for each node in the tree

def getTuple(name):

    links = []		# initialize variables
    tempStr = ''
    while 1:
        token = treeList.getNextToken()
        if token == ';' or token == 'end;':	# end reached, return tuple
            if tempStr != '':
                if tempStr[0] == '"' and tempStr[-1] == '"':
                    tempStr = tempStr[1:-1]
            return (name, tempStr, links)
        elif token[-1] == ';':		# end reached, append rest of value
            if len(tempStr) == 0:	# and return tuple
                tempStr = token[:-1]
            else:
                tempStr = tempStr + ' ' + token[:-1]
            if tempStr != '':
                if tempStr[0] == '"' and tempStr[-1] == '"':
                    tempStr = tempStr[1:-1]
            return (name, tempStr, links)
        elif token in delims:		# 'is' found, make recursive call
            links.append(getTuple(tempStr))
            tempStr = ''
        else:				# no marker found, continue to
            if len(tempStr) == 0:	# iterate through tokens
                tempStr = token
            else:
                tempStr = tempStr + ' ' + token


# the FileTokenizer class is used to read in an entire text file and store
# it in a list in memory, discarding comments. the list can then be
# iterated over without the user having to keep track of any list indices.

class FileTokenizer:

    # contructor reads file into list of tokens
    # comment argument can be used to discard lines
    # beginning with that character

    def __init__(self, filename, comment='#'):
        file = open(filename, 'r')

        lines = file.readlines()
        for i in range(len(lines)):
            if lines[i][:len(comment)] == comment:
                lines[i] = ''

        self.tokenList = split(join(lines))
        self.nextTokenIndex = 0

    # this function returns the next token as determined by the
    # nextTokenIndex variable	

    def getNextToken(self):
        self.nextTokenIndex = self.nextTokenIndex + 1
        return self.tokenList[self.nextTokenIndex - 1]

    # this function tells whether there are any more tokens or not

    def hasMoreTokens(self):
        return self.nextTokenIndex < len(self.tokenList)
    

if __name__ == '__main__':
    import sys
    print getTree(sys.argv[1])
