#!/usr/bin/env python
# encoding: utf-8
"""
txt2hkspc.py

This converts EAERI format text files containing housekeeping data into DMV-formatted *HK.SPC files.

Created by Keoni on 2008-03-24.
Copyright (c) 2008 University of Wisconsin SSEC. All rights reserved.
"""

import sys, os, re, logging, time, datetime, calendar

# from time_tools import *
import dmv

HISTORY_FMT = """written by $Id: txt2hkspc.py,v 1.1.2.23 2008/09/28 20:28:46 rayg Exp $ from %s"""

LOG = logging.getLogger('txt2hkspc')

# keys into the housekeeping dictionary which are unlabeled in the text file
KEY_YYMMDD = 'YYMMDD'
KEY_HHMMSS = 'HHMMSS'
KEY_JULIANDAY = 'JulianDay'   # derived from YYYYMMDD
KEY_TIME = 'Time'   # derived from HHMMSS
KEY_DATETIME = 'PythonDatetime' # datetime object in python

# nan constant
nan = float('nan')

#
# Summary functions
# reduce the lists of value down to single summary values
#

def summarize_first_value(values):
    return values[0]
    
def summarize_last_value(values):
    return values[-1]
    
def summarize_center_value(values):
    return values[int(len(values)/2)]

def summarize_mean_value(values):
    return sum(values) / float(len(values))
    
def summarize_flag(values):
    "return False for fully off, True for fully on, None for unstable"
    tot = sum([x and 1 or 0 for x in values])
    if tot==len(values): return True
    if tot==0: return False
    return None

def calc_hatch_open(D):
    """assign hatch values given hatchopendetect and hatchcloseddetect similarly to aerihk.cpp:AERIHK::calcHatchOpen
    """
    isOpen = summarize_flag(D['openLimitSwitch'])
    LOG.debug("isOpen: %s" % isOpen)
    isClosed = summarize_flag(D['closeLimitSwitch'])
    LOG.debug("isClosed: %s" % isClosed)
    if isOpen is True and isClosed is True: return -1.
    if isOpen is False and isClosed is True: return 0.
    if isOpen is True and isClosed is False: return 1.
    return -3.
    
def calc_jday(ymdstr,hours):
    """convert a YYYYMMDD string to day-of-year
    """
    yyyy = int(ymdstr[:4])
    mm = int(ymdstr[4:6])
    dd = int(ymdstr[6:8])
    t = time.mktime((yyyy, mm, dd, 0, 0, 0, 0, 0, 0))
    # compute julian day
    jday = time.gmtime(t)[7]
    LOG.debug('julian day is %d' % jday)
    return jday + hours / 24.0
    
def calc_hours(hhmmss):
    """convert a HHMMSS string to decimal hours, returning a float
    """
    hh = int(hhmmss[0:2])
    mm = int(hhmmss[2:4])
    ss = int(hhmmss[4:6])
    T = float(hh) + float(mm)/60. + float(ss)/3600.
    LOG.debug('decimal hours is %f' % T)
    return T

def calc_max_temp_diff( vectors, namelist ):
    """return maximum temperature diff between three sets of measurement lists
    FUTURE: this should use aerihk.cfg to get the list of channels to be used for a given blackbody
    and ignore unlisted channels and nans
    original equation was the max diff between more recent value of enabled channels (typically all: top, apex, bottom)
    this takes the maximum difference within the enabled channels' high-res samples, which is more of a measure of measurement stability 
    whereas the original measurement was more of an indicator of blackbody turbulence
    """
    minsies = [min( vectors[x] ) for x in namelist] 
    maxsies = [max( vectors[x] ) for x in namelist] 
    # return max( [(b-a) for a,b in zip(minsies,maxsies)] )  # new equation - measurement stability oriented
    return max(maxsies) - min(minsies)  # closer to original equation - highest measured minus lowest measured, more turbulence oriented
    # takes more data into account than the original aerihk.cpp implementation, which only took the most recent high-rez sample
    # we instead consider all the high-res samples

class UTC(datetime.tzinfo):
    """UTC"""
    ZERO = datetime.timedelta(0)
    def utcoffset(self, dt):
        return self.ZERO
    def tzname(self, dt):
        return "UTC"
    def dst(self, dt):
        return self.ZERO

def calc_datetime(yyyymmdd_str, hhmmss_str):
    yyyy = int(yyyymmdd_str[:4])
    mm = int(yyyymmdd_str[4:6])
    dd = int(yyyymmdd_str[6:8])
    h = int(hhmmss_str[:2])
    m = int(hhmmss_str[2:4])
    s = int(hhmmss_str[4:6])
    us = float(hhmmss_str[6:] or '0') * 1e6
    return datetime.datetime(yyyy, mm, dd, h, m, s, us, UTC())
    
def epoch_from_datetime(q): # adapted from cvs/devel/Retrieve/uwphysret/py/iasi_tools
    """
    convert datetime object to double precision seconds since unix epoch
    >>> epoch_from_datetime( datetimes_from_string( '20071204:093930-20071204')[0] )

    """
    return float(calendar.timegm( q.utctimetuple() )) + float(q.microsecond)/1e6

def celsiusToKelvin(vecdata):
    "summarize a value list in Celsius as a single Kelvin mean"
    return summarize_mean_value( [ 273.15 + x for x in vecdata ] )

def HX92AV_RH(vecdata):
    "convert HX92AV 0V-1V to 0%-100% RH"
    return 100. * summarize_mean_value( vecdata )

def atmosphericPressureSensor(vecdata):
    "convert EAERI atmosphericPressureSensor to millibars"
    return 120. * summarize_mean_value( vecdata ) + 500.

# FIXME: move this to a mandb-equivalent configuration file
def vancouverEaeriDiodeDetectorTemp(vecdata):
    "convert Vancouver E-AERI D2 Diode readout to detectorTemp, per spreadsheet 'diode data'"
    x = summarize_mean_value( vecdata ) 
    y = -584.92 * (x - 0.6872) + 295.15
    return y
    
def eaeriCoolerCurrent(vecdata):
    "summarize cooler current using voltage across 0.1ohm shunt resistor"
    v = summarize_mean_value( vecdata )
    # good old V = IR -> I = V/R
    return v * 10.

def read_txt(txt):
    """ reads EAERI .hk format file content given as a text string
    returns a dictionary of value sequences for housekeeping text
    read_txt(eaeri-housekeeping-text-string)
        => { 'name': [ floats ] , ... }
    """
    # read the file and extract list of pairs
    p = re.findall(r'(\w+)[\s=]+([.\dnNaA\-]+)', txt) # note assumption that findall is stable
    LOG.debug('found %d pairs in housekeeping text' % len(p))
    # extract date and time
    times = re.findall( r'^(\d{8})_([.\d]+)', txt, re.MULTILINE)    
    LOG.debug("found %d time values in text" % len(times))
    # make a dictionary of name: [list-of-values] in time order
    tab = { }
    for name,val in p:
        tab.setdefault(name,[]).append(float(val))
    tab[KEY_YYMMDD] = [int(x[0]) % 1000000 for x in times]
    tab[KEY_HHMMSS] = [float(x[1]) for x in times]
    hours = [calc_hours(x[1]) for x in times]
    tab[KEY_TIME] = hours
    tab[KEY_JULIANDAY] = [calc_jday(x[0],y) for x,y in zip(times,hours)]
    tab[KEY_DATETIME] = datis = [calc_datetime(x[0],x[1]) for x in times]
    LOG.debug("found keys in file: %s" % str(tab.keys()))
    return tab

# list of shortname, eaeriname, label, longname, converter-function, units, precision for vectors
# default converter function is in 
# see agilent.config for more details on how these measurements are made
# ( 'shortname', 'eaeriname', 'label', 'longname', None, 'units', -4 )
VECTOR_TABLE = [ ('timeHHMMSS',KEY_HHMMSS,'Time in HHMMSS',None,None,None,0),
            ('dateYYMMDD',KEY_YYMMDD,'Date in YYMMDD',None,None,None,0),
            ('Time', KEY_TIME, 'Time of day in decimal hours', None,None,None, -4),
            ('JulianDay', KEY_JULIANDAY, 'Julian day', None,None,None, 0),
            
            ('HBBapexTemp', 'hotBlackbodyApexTemperature', 'HBB Apex Temp', None,celsiusToKelvin,'degC', -3), 
            ('HBBtopTemp', 'hotBlackbodyTopRimTemperature', 'HBB Top Temp', None,celsiusToKelvin,'degC', -3),
            ('HBBbottomTemp', 'hotBlackbodyBottomRimTemperature', 'HBB Bottom Temp', None,celsiusToKelvin,'degC', -3),
            ('ABBapexTemp', 'ambientBlackbodyApexTemperature', 'ABB Apex Temp', None,celsiusToKelvin,'degC', -3), 
            ('ABBtopTemp', 'ambientBlackbodyTopRimTemperature', 'ABB Top Temp', None, celsiusToKelvin,'degC',-3),
            ('ABBbottomTemp', 'ambientBlackbodyBottomRimTemperature', 'ABB Bottom Temp', None, celsiusToKelvin,'degC',-3),

            ('fixed97KohmResistor', 'refResistance3', None, None, None, 'Ohms',-3),  # FIXME: 1, 2, or 3?
            ('fixed2500ohmResistor', 'refResistance1', None, None, None, 'Ohms',-3), # FIXME: 1, 2, or 3?
            ('fixed12KohmResistor', 'refResistance2', None, None, None, 'Ohms',-3),  # FIXME: 1, 2, or 3?

            ('BBsupportStructureTemp', 'blackbodySupportStructureTemperature', None, None, celsiusToKelvin,'degC',-3),
            ('outsideAirTemp', 'outsideAirTemperature', None, None, celsiusToKelvin,'degC',-3),
            ('airNearInterferometerTemp','interferometerAirTemperature','',None, celsiusToKelvin,'degC',-3),
            ('mirrorMotorTemp','mirrorMotorTemperature','',None,celsiusToKelvin,'degC',-3),            
            ('sceneMirrorTemp','sceneMirrorTemperature','',None,celsiusToKelvin,'degC',-3),
            ('interferometerWindowTemp','sceneWindowTemperature','',None,celsiusToKelvin,'degC',-3),
            
            ('coolerExpanderTemp','stirlingCoolerExpenderResistance','',None,celsiusToKelvin,'degC',-3),
            ('coolerExpanderTemp','stirlingCoolerExpenderTemperature','',None,celsiusToKelvin,'degC',-3),  # FIXME misspelling
            ('coolerCompressorTemp','stirlingCoolerCompressorTemperature','',None,celsiusToKelvin,'degC',-3),
            ('detectorTemp','detectorTemperatureSensor','',None,vancouverEaeriDiodeDetectorTemp,'V',-3),   # FIXME doublecheck that this isn't already done in FTSW
            ('opticsCompartmentRelativeHumidity','interferometerRelativeHumidity','',None,HX92AV_RH,'V',-3),

            ('atmosphericRelativeHumidity','atmosphericRelativeHumidity','',None,HX92AV_RH,'V',-3),
            ('airNearBBsTemp','blackbodyAirTemperature','',None,celsiusToKelvin,'degC',-3),
            ('sceneMirrorAngle','pointingMirrorAngle','',None,None,'degrees',-3),
            ('atmosphericPressure','atmosphericPressureSensor','',None,atmosphericPressureSensor,'V',-3), 
            ('','externalSunDetectorSensor','',None,None,None,-3),
            ('rainSensorIntensity','internalRainDetectorSensor','',None,None,None,-3), # boolean form is insideRainSensor, which is 1 on clear and 0 on rainy; also outsideRainSensor / externalRainDetectorSensor
            ('SCEtemp','dataAcquisitionModulesTemperature','',None,celsiusToKelvin,'degC',-3),

            ('','ambientBlackbodyTopRimResistance','',None,None,"Ohms",-3),
            ('','ambientBlackbodyApexResistance','',None,None,"Ohms",-3),
            ('','ambientBlackbodyBottomRimResistance','',None,None,"Ohms",-3),
            ('','hotBlackbodyTopRimResistance','',None,None,"Ohms",-3),
            ('','hotBlackbodyApexResistance','',None,None,"Ohms",-3),
            ('','hotBlackbodyBottomRimResistance','',None,None,"Ohms",-3),
            ('','refTemp1','',None,celsiusToKelvin,"degC",-3),
            ('','dataAcquisitionModulesResistance','',None,None,"Ohms",-3),
            ('','sceneWindowResistance','',None,None,"Ohms",-3),
            ('','blackbodySupportStructureResistance','',None,None,"Ohms",-3),
            ('','mirrorMotorResistance','',None,None,"Ohms",-3),
            ('','powerSupplyResistance24V','',None,None,None,-3),
            ('coolerCurrent','stirlingCoolerCurrent','',None,None,'V',-3), # direct voltage that needs to be converted to a current
            ('','inputMirrorMotorDriveElectronicResistance','',None,None,'Ohms',-3),
            ('','blackbodyAirResistance','',None,None,'Ohms',-3),
            ('','refTemp2','',None,None,None,-3),
            ('','refTemp3','',None,None,None,-3),
            ('','outsideRainSensor','',None,None,None,-3),
            ('','powerSupplyTemperature28V','',None,celsiusToKelvin,'degC',-3),
            ('hatchClosedDetect','closeLimitSwitch','',None,None,None,-3),
            ('','externalRainDetectorSensor','',None,None,None,-3),
            ('','thermalEnclosureResistance','',None,None,'Ohms',-3),
            ('','powerSupplyResistance28V','',None,None,None,-3),
            ('','outsideAirResistance','',None,None,'Ohms',-3),
            ('','protectiveEnclosureResistance','',None,None,'Ohms',-3),
            ('sunSensorIntensity','internalSunDetectorSensor','',None,None,None,-3),
            ('motorDriverTemp','inputMirrorMotorDriveElectronicTemperature','',None,celsiusToKelvin,'degC',-3),
            ('','powerSupplyTemperature24V','',None,celsiusToKelvin,'degC',-3),
            ('','protectiveEnclosureTemperature','',None,celsiusToKelvin,'degC',-3), # encloses the Thermal Enclosure when outdoors
            ('','interferometerAirResistance','',None,None,'Ohms',-3),
            ('','stirlingCoolerCompressorResistance','',None,None,'Ohms',-3),
            ('','sceneMirrorResistance','',None,None,None,-3),
            ('','fanAirflowTemperature','',None,celsiusToKelvin,'degC',-3),
            ('','fanAirflowResistance','',None,None,'Ohms',-3),
            ('','insideRainSensor','',None,None,None,-3),
            ('rackAmbientTemp','thermalEnclosureTemperature','',None,celsiusToKelvin,'degC',-3),  # holds the interferometer and electronics
            ('hatchOpenDetect','openLimitSwitch','',None,None,None,-3),
            ]    

def extract_scalars(vectors, table = VECTOR_TABLE, abb_thermistor_names = None, hbb_thermistor_names = None):
    """ extract scalars from the high-resolution samples
    return dictionary of name -> float
    refer to augmentListsCalc() in aerihk.cpp
    """
    zult = { }
    for vec_sname, vec_ename, _, _, handler, _, _ in table:
        if not vec_sname: continue
        if not handler: handler = summarize_mean_value
        zult[vec_sname] = handler(vectors[vec_ename])
        
    # now extract values with more obscure handling
    zult['hatchOpen'] = calc_hatch_open(vectors)
    # assign encoder position as the pointing mirror angle - until such time as we get a full encoder readout
    zult['sceneMirrorMotorStep'] = zult['sceneMirPosEncoder'] = summarize_mean_value(vectors['pointingMirrorAngle'])
    zult['sceneMirPosEncoderDrift'] = max(vectors['pointingMirrorAngle']) - min(vectors['pointingMirrorAngle']) 
    zult['frontEndFanAirTempDiff'] = nan  
    zult['maxPitch'] = nan
    zult['maxSampleStdDev'] = nan   # FUTURE maximum housekeeping std would be useful for judging electronics stability
    zult['ABBmaxTempDiff'] = calc_max_temp_diff( vectors, abb_thermistor_names or ['ambientBlackbody%sTemperature' % x for x in ['TopRim', 'Apex', 'BottomRim'] ] )
    zult['HBBmaxTempDiff'] = calc_max_temp_diff( vectors, hbb_thermistor_names or ['hotBlackbody%sTemperature' % x for x in ['TopRim', 'Apex', 'BottomRim'] ] )

    # values that should be replaced with sequencer information
    zult['sceneMirrorPosition'] = nan   # this should be added externally prior to writing the file

    zult['sceneViewDuration'] = epoch_from_datetime( vectors[KEY_DATETIME][-1] ) - epoch_from_datetime( vectors[KEY_DATETIME][0] )
    
    # LOG.warning( "Not Implemented: " + "assorted housekeeping value computations not available / impossible" )
    
    #/ zult['maxSampleStdDev'] = FIXME   # maximum standard deviation among selected HK channels, for stability measure
    #/ zult['maxPitch'] = FIXME  # clinometer max pitch, degrees
    #/ zult['sceneMirrorPosition'] ? ? ? FIXME do we need this?
    #/ zult['frontEndFanAirTempDiff'] = FIXME  # difference of heated - unheated front end fan air temps
    # Time and JulianDay are unpacked during file read
    return zult
    
AERI_MODELS = { 'AERI': 1, 
                'MAERI': 2,
                'PAERI': 3,
                'EAERI': 4 
                }
                
def read_inst_id(filename):
    """ return model and serial number tuple given the filename of an inst.id file
    refer aerihk.cpp:AERIexterns::readINST_ID()
    """
    if not os.path.exists(filename):
        LOG.error('INST.ID is missing and should be set to EAERI-###')
        return nan, nan
    txt = file(filename,'rt').read().strip()
    LOG.debug('text read from INST.ID: "%s"' % txt)
    smod,sser = txt.split('-')
    return AERI_MODELS[smod], int(sser)
    
# regular expression which locates the number of coadds (not commented out) in a SET file
RE_COADDS = re.compile(r'^[^#]*smnsca\s*=\s*(\d+)', re.IGNORECASE|re.MULTILINE)

def read_set_coadds(filename):
    """ return coadds from .SET file
    refer aerihk.cpp:AERIexterns::readSetCoadds()
    """
    if not os.path.exists(filename):
        LOG.error('SET file is missing')
        return nan
    text = file(filename,'rt').read()
    # note that this will throw a ValueError exception in the case that more than one valid value is found
    (q,) = RE_COADDS.findall(text)
    return int(q) 

RE_LATLONALT = re.compile(r'(\w\w\w)\s*=\s*([-\.0-9]+)', re.MULTILINE)

def read_aeri_pos(filename):
    """ return Latitude N, Longitude E, Altitude meters values
    LAT = 43.070597
    LON = -89.407239
    ALT = 237.430000
    """
    if not os.path.exists(filename):
        LOG.warning('%s does not exist!' % filename)
        return nan, nan, nan
    D = dict( [(x,float(y)) for x,y in RE_LATLONALT.findall(file(filename,'rt').read())] )
    lat = D.get('LAT',nan)
    lon = D.get('LON',nan)
    alt = D.get('ALT',nan)
    LOG.debug('%s' % str(D))
    return lat, lon, alt

def count_scene_types(sched):
    """count the unique named mirror positions in the schedule
    see AERIexterns::countPositions in aerihk.cpp
    """
    # cheap set alternative
    qq = [ (scene.name[0].upper(),None) for scene in sched.views ]
    est = dict(qq)
    LOG.debug('scene types: %s' % str(list(est.keys())))
    return len(est)
    
def config_scalars(set_filename, inst_id_filename, aeri_pos_filename, mirror_scd_filename):
    """ return scalars which need to be read from configuration files
    set_filename is the filename of the .SET file which was used for this ingest cycle, 
    typically a generated file living in the ingest work directory and not a config file
    FUTURE: the Sequencer can provide the number of coadds to this routine instead
    """
    if inst_id_filename:
        model,serial = read_inst_id(inst_id_filename)
    else:
        LOG.warning("no INST.ID filename to read model and serial number from")
        model,serial = nan,nan
        
    if not set_filename:
        LOG.error("no SET file to read coadds from!")
        coadds = 0
    else:
        coadds = read_set_coadds(set_filename)
    
    if not aeri_pos_filename:
        LOG.warning("no AERI.POS filename provided -> no position information")
        # FUTURE: we should short-cut AERI.POS with a direct Gps.py readout state?? 
        lat, lon, alt = nan, nan, nan
    else:
        lat, lon, alt = read_aeri_pos(aeri_pos_filename)
        LOG.debug('latitude, longitude, altitude = %s, %s, %s' % (lat,lon,alt))
        
    if not mirror_scd_filename:
        LOG.warning('no mirror.scd filename provided -> no sceneMirPosCount or sceneMirrorPosition')
        smpc, smp = nan, nan
    else:
        from mirscd import mirror_schedule
        LOG.debug('reading mirror schedule file')
        emm = mirror_schedule( file(mirror_scd_filename,'rt') )
        smpc = float(count_scene_types(emm))
        smp = float(ord( emm.views[emm.iview-1].name[0] ))
        LOG.info('mirror position count = %s, mirror position = %s' % (smpc,smp))
    
    return {'AERIunitNumber': model, 
            'instrumentUnitNumber': serial, 
            'systemReleaseNumber': 4.0,   # is in AERIHK.CFG which we retire for EAERI
            'coadditionsCount': coadds,  # number of co-adds specified in .SET file
            'Latitude': lat,
            'Longitude': lon,
            'Altitude': alt,
            'sceneMirPosCount': smpc,
            'sceneMirrorPosition': smp,
            'missingDataFlag': 0.0,
            }
            
def check_for_safing(vectors):
    """handle mirror safing
    if insideRainSensor does not average to 1.0, then set 
    sceneMirrorPosition to '?'
    sceneMirPosEncoder to nan
    sceneMirrorAngle gets left alone, reports actual safing position
    sceneMirPosEncoderDrift to nan
    this emulates our inability to read the encoder when the safing program has control
    """
    scalars = { }   # empty dictionary of update values
    irs = vectors['insideRainSensor']
    wet = bool(summarize_mean_value(irs) < 1.0)
    LOG.debug("inside rain sensor: %s (%s)" % (str(irs),wet))
    if wet:
        scalars['sceneMirrorPosition'] = float(ord('?'))
        scalars['sceneMirPosEncoder'] = nan
        scalars['sceneMirPosEncoderDrift'] = nan
        LOG.warning('mirror is believed safed, setting mirror position and encoder value')
    return scalars

def _create_file(filename, max_vector_length, table = VECTOR_TABLE, history=''):
    if os.path.exists(filename):
        raise EnvironmentError, 'file %s already exists' % filename
    fp = dmv.dmv()
    # reserve storage space for scalar housekeeping values
    fp.addHousekeepingSubheader()
    # set file metadata
    fp.setApplication('txt2hkspc')
    fp.setVersion('$Id: txt2hkspc.py,v 1.1.2.23 2008/09/28 20:28:46 rayg Exp $')
    fp.setHistory(history)
    # reserve storage space for high-resolution housekeeping
    name2vid = {}
    vid = 1
    for vec_sname, vec_ename, vec_label, vec_lname, handler, vec_units, vec_precision in table:
        # print vec_ename
        LOG.debug("creating %s (%s) (%s, %s)" % (vec_ename, vec_sname or '', vec_label, vec_units)) 
        fp.newVector(vid, 1.0, float(max_vector_length), max_vector_length)
        fp.setVectorShortname(vec_ename)
        fp.setVectorLongname(vec_lname or vec_sname or vec_ename)
        fp.setVectorLabel(vec_label or vec_sname or vec_ename)
        fp.setVectorUnits(vec_units or 'unknown')
        fp.setVectorPrecision(0,vec_precision or -3)
        fp.commitNewVector()
        name2vid[vec_ename] = vid
        vid += 1        
    rc = fp.createFile(filename)
    assert(rc==0)
    return fp, name2vid

def _write_record(fp, scalars, vectors, max_vector_length, name2vid, record = 1):
    # get ids for scalars
    scalar_tmp = [(fp.findMetaID(key),val) for (key,val) in scalars.items()]
    scalar_ids = dmv.ints([x[0] for x in scalar_tmp])
    scalar_vals = dmv.floats([x[1] for x in scalar_tmp])
    # create a new record
    rc = fp.newRecord(record)
    assert(rc==0)
    # add the high resolution housekeeping
    for name,vod in name2vid.items():
        value = vectors[name]
        if len(value)<max_vector_length: # not technically a problem; pad with NaN
            LOG.debug("%s has only %d values!" % (name,len(value)))
        if len(value) > max_vector_length:
            value = value[-max_vector_length:]
            LOG.warning("%s has too many values! truncating to last %d" % (name,len(value)))
        # note that there may be fewer values in 
        rc = fp.addVector(name2vid[name], dmv.floats(value))
        assert(rc==0)
    # add the scalars
    rc = fp.addMeta(scalar_ids, scalar_vals)
    assert(rc==0)
    rc = fp.commitNewRecord()
    assert(rc==0)

            
def write_file(filename, scalars, vectors, max_vector_length, table = VECTOR_TABLE, history='', fp=None, name2vid=None, record=1):
    """Write a file using the DMV OHWHIO module, containing the housekeeping scalars and vectors given.
    Set the file attributes correctly and informatively.
    """
    if not fp or not name2vid:
        fp, name2vid = _create_file(filename, max_vector_length, table, history)
    _write_record(fp, scalars, vectors, max_vector_length, name2vid, record)
    return fp, name2vid

def test_write_file():
    # table of vectors to build
    vexation = [ ('HBBapexTemp', 'HBBapexTemp', 'HBB Apex Temp', None, celsiusToKelvin,'degC', -4), 
                ('HBBtopTemp', 'HBBtopTemp', 'HBB Top Temp', None, None, None, 0),
                ('HBBbottomTemp', 'HBBbottomTemp', 'HBB Bottom Temp', None, None, None, 0)
                ]
    scatest = { 'HBBapexTemp': -333., 'HBBtopTemp': 111., 'HBBbottomTemp': 222. }
    vectest = { 'HBBapexTemp': range(12), 
                'HBBtopTemp': range(12),
                'HBBbottomTemp': range(12)
                }
    TESTFN='/tmp/testhk.spc'
    if os.path.exists(TESTFN):
        os.unlink(TESTFN)
    write_file(TESTFN, scatest, vectest, table = vexation, max_vector_length=32)

def txt2hkspc(output_filename, input_filenames, set_filename=None, inst_id_filename=None, aeri_pos_filename=None, mirror_scd_filename=None, max_vector_length=None):
    """ write a *HK.SPC file from an EAERI .hk file
    Scalars are extracted from "high resolution" input data using a variety of functions.
    set_filename is the path of the current .SET file used by acqeaeri.
    """ 
    fp = None
    name2vid = None
    record = 1
    for input_filename in input_filenames:
        # read the input file into a dictionary of arrays
        LOG.debug('reading %s' % input_filename)
        txt = file(input_filename).read()
        hires = read_txt(txt)
        # extract the scalars
        scalars = extract_scalars(hires)
        LOG.debug('extracted %d scalars from hi-res data' % len(scalars))
        # read assorted configuration and output file values
        scalars.update( config_scalars(set_filename,inst_id_filename,aeri_pos_filename,mirror_scd_filename) )  
        if not mirror_scd_filename: 
            smp = ord(os.path.split(input_filename)[1][0])
            LOG.info('guessing scene type "%s" from filename' % chr(smp))
            scalars['sceneMirrorPosition'] = smp
        LOG.debug('%s' % str(scalars))
        for k in scalars.keys():
            if k in os.environ:
                scalars[k] = float(os.environ[k])
                LOG.info( 'overriding %s as %f' % (k,scalars[k]))
        smpc = scalars['sceneMirPosCount']
        if smpc!=smpc:
            LOG.error('no sceneMirPosCount! set in environment if no mirror.scd available')
        scalars.update(check_for_safing(hires))
        # FUTURE: use GPS coords if available in hires
        # determine the max vector length
        if max_vector_length is None:
           mvl = max( [len(q) for q in hires.values()] )
        else:
           mvl = max_vector_length
        LOG.debug("max vector length is %d" % mvl)
        LOG.debug('%s' % dict( [ (name,len(values)) for name,values in hires.items() if len(values)==mvl ] ))
        LOG.debug('%s' % dict( [ (name,len(values)) for name,values in hires.items() if len(values)!=mvl ] ))
        LOG.debug('writing the output file %s' % output_filename)    
        if not fp:
            fp, name2vid = _create_file(output_filename, mvl, history = HISTORY_FMT % str(input_filenames))
        _write_record(fp, scalars, hires, mvl, name2vid, record)            
        # fp, name2vid = write_file(output_filename, scalars, hires, mvl, history = HISTORY_FMT % input_filename, fp = fp, name2vid = name2vid, record = record)
        record += 1
    LOG.debug('done')
    return hires

def test_write_glob(filename, globname):
    """write a series of text files to a single CXS-like file of housekeeping
    sorts files by their creation time
    """
    from glob import glob
    lal = sorted( [ (os.stat(fn).st_ctime, fn) for fn in glob(globname) if '.' not in fn ] ) # A1, H2, A2, C2, etc
    input_filenames = [fn for _,fn in lal]
    txt2hkspc(filename, input_filenames, max_vector_length = 64)
    

if __name__ == '__main__':
    logging.basicConfig()
    if 'DEBUG' in os.environ:
        LOG.setLevel(logging.DEBUG)
    else:
        LOG.setLevel(logging.INFO)
    if len(sys.argv)==1: 
        LOG.info('running self-test') 
        test_write_file()
        print "usage: txt2hkspc outputhk.spc /path/to/hkdata.hk {/path/to/hkdata.hk ...}" 
    else:
        max_vector_length = None
        if 'max_vector_length' in os.environ:
            max_vector_length = int(os.environ['max_vector_length'])
            LOG.info('max_vector_length is %d' % max_vector_length)
        D = txt2hkspc(sys.argv[1], sys.argv[2:], max_vector_length = max_vector_length)
